/**
 * Game of Life
 *
 */
package gridlife.color;



/**
 * Imports
 * 
 */
import gridlife.core.GridLifeConstants;
import java.awt.Color        ;
import java.util.HashMap     ;
import java.util.Collection  ;
import java.util.Iterator    ;
import java.util.Arrays      ;
import java.io.File          ;




/**
 * This class stores a complete list of color/age records and supports loading and storing them.
 * It also keeps local copies of the currently active sets, its name and its sorted in[] array
 * to speed up drawing.
 * 
 * @author Ulf Wagemann
 * 
 */
public class GridLifeColorManager 
{
 private HashMap<String, GridLifeColorSet> m_tColorSets;
 
 private String           m_sActiveSetName;
 private int[]            m_tActiveSetAges;
 private GridLifeColorSet m_tActiveSet    ;
  
  
 /**
  * Constructor
  * 
  */
 public GridLifeColorManager()
 {
  m_tColorSets     = new HashMap<String, GridLifeColorSet>(); 
  m_sActiveSetName = "";
  m_tActiveSet     = null;
  
  loadColorSets();
 }
 
 
 
 /**
  * We do not want to eat up memory, since we store own datastructures we need to remove them from the
  * hashmap and mark then as gone for the GC.
  * 
  * dispose() should be called when this shapeset is not needed anymore.
  * 
  */
 public final void dispose()
 {
  if (null != m_tColorSets) {m_tColorSets.clear();}
 }
 

 
 /**
  * Returns the color value for the given age
  * 
  * @return  -1 in case of error, color value otherwise
  */
 public final GridLifeColorSet getColorSet(String pa_sName)
 {
  if (null != m_tColorSets)
  {
   return m_tColorSets.get(pa_sName);
  }
  return null;
 }
 
 
  
 /**
  * Loads the color sets
  * 
  */
 public final void loadColorSets()
 {
  File   lc_tFile             = null;
  File   lc_tColorDirectory   = null;
  File[] lc_tDirectoryEntries = null;
  
  int lc_iAmount        = 0;
  int lc_iI             = 0;
 
  GridLifeColorSet           lc_tColorSet = null;
  GridLifeColorSetFileFilter lc_tFilter   = new GridLifeColorSetFileFilter();
  
  if (null != (lc_tColorDirectory = new File(GridLifeConstants.GOL_FILE_COLOR_DIR)))
  {
   if (true == lc_tColorDirectory.exists() && true == lc_tColorDirectory.isDirectory())
   {
    if (null != (lc_tDirectoryEntries = lc_tColorDirectory.listFiles(lc_tFilter)))
    {
     if (0 < (lc_iAmount = lc_tDirectoryEntries.length))
     {
      if (1 < lc_iAmount)
      {
       Arrays.sort(lc_tDirectoryEntries);        
      }
      for (lc_iI = 0; lc_iI < lc_iAmount; lc_iI++)
      {
       if (null != (lc_tFile = lc_tDirectoryEntries[lc_iI])) 
       {
        if (null != (lc_tColorSet = new GridLifeColorSet(lc_tFile)))
        {
         m_tColorSets.put(lc_tColorSet.getSetName(), lc_tColorSet); 
        }
       }
      }
     }
    }
   }    
  }
   
 }
 
 
 
 /**
  * Retrieves the names of all of the colors belonging to the current set
  * 
  * @return   Names of all shapes as a set, null in case of error
  */
 public final String[] getColorSetNames()
 {
  GridLifeColorSet             lc_tSet        = null; 
  Collection<GridLifeColorSet> lc_tSetEntries = null;
  Iterator<GridLifeColorSet>   lc_tIterator   = null;
  String[]                     lc_tResult     = null;
  
  int lc_iPos = 0;
  
  if (null != m_tColorSets)
  {
   if (0 < m_tColorSets.size())
   {
    if (null != (lc_tSetEntries = m_tColorSets.values()))
    {
     if (null != (lc_tResult = new String[m_tColorSets.size()]))
     {
      if (null != (lc_tIterator = lc_tSetEntries.iterator()))
      {
       while (lc_tIterator.hasNext()) 
       {
        if (null != (lc_tSet = (GridLifeColorSet) lc_tIterator.next()))
        {
         lc_tResult[lc_iPos++] = lc_tSet.getSetName();
        }
       }
       Arrays.sort(lc_tResult);
       return lc_tResult;   
      }
     }
    }
   }
  }
  return null;
 } 
 
 
 
 
 /**
  * Returns an array of color records for the given set name
  * 
  * @param pa_sName   set name
  * @return  null in case of error, color record array otherwise
  */
 public final GridLifeColorRecord[] getColorSetRecords(String pa_sName)
 {
  GridLifeColorSet  lc_tColorSet = null;
 
  if (null != m_tColorSets)
  {
   if (null != (lc_tColorSet = m_tColorSets.get(pa_sName)))
   {
    return lc_tColorSet.getSet(); 
   }
  }
  return null; 
 }



 /**
  * Returns the color for the fiven age according to the given color set
  * 
  * @param pa_sName   color set name
  * @param pa_iAge    age
  * @return   Color.LIGHT_GRAY in case of error, matching color otherwise
  */
 public final Color getColorForSet(String pa_sName, int pa_iAge)
 {
  GridLifeColorSet  lc_tColorSet = null;
 
  if (null != m_tColorSets)
  {
   if (null != (lc_tColorSet = m_tColorSets.get(pa_sName)))
   {
    return lc_tColorSet.getColor(pa_iAge); 
   }
  }   
  return Color.LIGHT_GRAY; 
 }



 /**
  * Returns the amount of color sets contained
  * 
  * @return   -1 in case of error, otherwise amount of sets (might be zero)
  */
 public final int getColorSetAmount() {return (null != m_tColorSets ? m_tColorSets.size() : -1);}

 
 
 /**
  * Returns the name of the active color set
  * 
  * @return "" if no set is active, set name otherwise
  */
 public final String getActiveSetName() {return m_sActiveSetName;}
 
 
 
 /**
  * REturns the color for the given age accordint to the currently active set
  * 
  * @param pa_iAge
  * @return 
  */
 public final Color getActiveColor(int pa_iAge) {return (null != m_tActiveSet ? m_tActiveSet.getColor(pa_iAge) : Color.LIGHT_GRAY);}
         
 
 /**
  * Sets the name of the active color set to the given value
  * 
  * @param pa_sName   new name
  */
 public final void setActiveSetName(String pa_sName)
 {
  m_sActiveSetName = pa_sName;  
  m_tActiveSetAges = getAgesForSet(pa_sName);
  m_tActiveSet     = getColorSet(pa_sName);
 }

 
 
 /**
  * Returns a sorted array of ages for the the given color set
  * 
  * @param pa_sName   color set name
  * @return   null in case of error, sorted age array otherwise
  */
 public final int[] getAgesForSet(String pa_sName)
 {
  GridLifeColorSet  lc_tColorSet = null;
 
  if (null != m_tColorSets)
  {
   if (null != (lc_tColorSet = m_tColorSets.get(pa_sName)))
   {
    return lc_tColorSet.getAges(); 
   }
  }   
  return null; 
 }
 
 
 
 /**
  * Retrieves the age class for a given cell age according to the given colorset
  * 
  * @param pa_tAge           given age   
  * @return 
  */
 public final Integer getAgeClass(String pa_sName, int pa_iAge, int[] pa_tAgeKeys)
 {
  int     lc_iResult = -1;
  int     lc_iLength = 0;
  int     lc_iPos    = 0;
  boolean lc_bFound = false;
  
  int[] lc_tKeys = (null != pa_tAgeKeys ? pa_tAgeKeys : m_tActiveSetAges); // work with references for max speed! 
  
  //
  // we should NEVER get into here, but for saftey...
  //
  if (null == lc_tKeys)
  {
   lc_tKeys = getAgesForSet(pa_sName); 
  }
  
  if (0 < pa_iAge && Integer.MAX_VALUE > pa_iAge && null != lc_tKeys)
  {
   if (0 < (lc_iLength = lc_tKeys.length)) 
   {
    //
    // iterate over keys
    //
    while (false == lc_bFound && lc_iPos < lc_iLength) 
    {
     if (lc_tKeys[lc_iPos] >= pa_iAge)
     {
      lc_iResult = lc_tKeys[lc_iPos]; 
      lc_bFound  = true;
     }
     else
     {
      lc_iPos++;
     }
    } // while
    
    //
    // add to last key, if no key is greater then the search key
    //
    if (false == lc_bFound)
    {
     lc_iResult = lc_tKeys[lc_iLength - 1]; 
    }
    return new Integer(lc_iResult);
   }
  }
  return null;
 }

 
 /**
  * Returns the currently active ages
  * 
  * @return 
  */
 public final int[] getActiveAges() {return m_tActiveSetAges;}
 
 
 
 /**
  * Returns the currently active colorset
  * 
  * @return   colorset
  */
 public final GridLifeColorSet getActiveSet() {return m_tActiveSet;}
}
