/**
 * Game of Life
 * 
 * 1.04: added zooming via mouse wheel
 *       fixed wrong color bug for two color popups in the settings window
 *       added new setting "screenshot cell size" in UI and properties, modified screenshot routine accordingly
 *       minor code and UI display/settings fixes 
 *       screenshot method checks the available memory using runtime now
 *       revised UI, introduced icon buttons instead of text buttons
 *       fixed a nasty bug in the computation section, now taking care of rectangular (not neccessarily squared!) cell matrix.
 * 
 * Nore: the current set of properties is stored in the cell matrix object m_tCellMatrix, because the values of the properties windows may be subject to change at any time!
 * 
 */
package gridlife.core;



/**
 * Tons of imports :-)
 * 
 */
import gridlife.color.*;
import gridlife.ui.*   ;
import gridlife.shape.*; 

import java.awt.event.WindowListener     ;
import java.awt.event.MouseListener      ;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelListener ;
import java.awt.event.MouseWheelEvent    ;
import java.awt.event.AdjustmentEvent    ;
import java.awt.event.AdjustmentListener ;
import java.awt.event.WindowEvent        ;
import java.awt.event.MouseEvent         ;
import java.awt.event.ItemEvent          ;
import java.awt.image.BufferedImage      ;

import java.awt.Component;
import java.awt.Point    ;
import java.awt.Color    ;
import java.awt.Rectangle;
import java.awt.Dimension;
import java.awt.Toolkit  ;

import java.io.File                    ;
import java.io.FilenameFilter          ;
import java.util.ResourceBundle        ;
import java.util.Arrays                ;
import java.beans.PropertyVetoException;

import javax.swing.JOptionPane   ; 
import javax.swing.JPanel        ;
import javax.swing.JFrame        ;
import javax.swing.JInternalFrame;   
import javax.swing.JScrollBar    ;
import javax.swing.JFileChooser  ;
import javax.swing.SwingUtilities;
import javax.swing.JColorChooser ;

import javax.swing.event.DocumentEvent        ;
import javax.swing.event.DocumentListener     ;
import javax.swing.event.InternalFrameEvent   ;
import javax.swing.event.InternalFrameListener;
import javax.swing.event.ListSelectionEvent   ;
import javax.swing.event.ChangeEvent          ;

import org.jdesktop.application.Application           ;
import org.jdesktop.application.SingleFrameApplication;



/**
 * The main class of the application. Yes, I know it is huge. Maybe too huge. But this is the tradeoff caused
 * by the need of having one instance storing all the cell data and reacting on gui events which impact the cell
 * data in terms of drawing and computation. Any further splitting into smaller pieces adds no value and is 
 * neither easier to maintain nor to understand.
 * 
 * @author  Ulf Wagemann
 */
public class GridLifeApp extends    SingleFrameApplication 
                         implements GridLifeGuiConnector  ,
                                    WindowListener        ,
                                    InternalFrameListener ,
                                    MouseListener         ,
                                    MouseMotionListener   ,
                                    MouseWheelListener    ,
                                    AdjustmentListener    ,
                                    DocumentListener 
{
 private GridLifeMainWindow           m_tMainWindow         ;
 private GridLifePropertiesWindow     m_tPropertiesWindow   ;
 private GridLifeStatisticsWindow     m_tStatisticsWindow   ;
 private GridLifeShapePaletteWindow   m_tShapePaletteWindow ;
 private GridLifeCellMatrixWindow     m_tCellMatrixWindow   ;
 private GridLifeColorWindow          m_tColorWindow        ;
 private GridLifePaletteEditorWindow  m_tPaletteEditorWindow;
 private GridLifeQuickSlotDialog      m_tQuickSlotDialog    ;
        
 private GridLifeStatisticsPaintPanel   m_tStatisticsPaintPanel ;
 private GridLifeCellMatrixPaintPanel   m_tCellPaintPanel       ;
 private GridLifeAgeHistogramPaintPanel m_tHistogramPaintPanel  ;
 private Color                          m_tCellAgeFillColor     ;
 private Point                          m_tLastMouseCellPosition;
 
 private GridLifeWindowSettings      m_tWindowSettings     ;
 private GridLifeShapeSetManager     m_tShapeSetManager    ;
 private GridLifeColorManager        m_tColorManager       ;
 private GridLifeStatisticsManager   m_tStatisticsManager  ;
 private GridLifeAgeHistogramManager m_tAgeHistogramManager;

 private GridLifeCellMatrix         m_tCellMatrix       ;
 private GridLifeAnimationThread    m_tAnimationThread  ;
 private GridLifeLookAndFeelMenu    m_tLFMenu           ;
 private ResourceBundle             m_tLocalizedMessages;
 
 private int m_iZoomFactor        ;
 private int m_iFillRatio         ;
 private int m_iCellSetAge        ;
 private int m_iIOThreadsRunning  ;
 private int m_iSceenshotCellSize ;
 
 private boolean m_bLifeInProgress  ;
 private boolean m_bMatrixModified  ;     
 private boolean m_bInvertMouseWheel;
 
 
 /**
  * Constructor
  * 
  */
 public GridLifeApp()
 {
  super();
  
  m_tMainWindow           = null;
  m_tPropertiesWindow     = null;
  m_tStatisticsWindow     = null;
  m_tShapePaletteWindow   = null;
  m_tCellMatrixWindow     = null;
  m_tColorWindow          = null;
  m_tPaletteEditorWindow  = null;
  m_tQuickSlotDialog      = null;
  m_tLFMenu               = null;
  
  m_tCellPaintPanel       = null;
  m_tStatisticsPaintPanel = null;
  m_tHistogramPaintPanel  = null;
  m_tCellAgeFillColor     = null;
  
  m_tWindowSettings       = null;
  m_tShapeSetManager      = null;
  m_tColorManager         = null;
  m_tCellMatrix           = null;
  m_tStatisticsManager    = null;
  m_tAnimationThread      = null;
  m_tAgeHistogramManager  = null;
  
  m_iZoomFactor         = 0 ;
  m_iCellSetAge         = 1 ;
  m_iFillRatio          = 10;
  m_iIOThreadsRunning   = 0 ;
  m_iSceenshotCellSize  = 3 ;
  
  m_bLifeInProgress        = false;
  m_bMatrixModified        = false;
  m_bInvertMouseWheel      = false;
  m_tLastMouseCellPosition = new Point(0,0);
  
  m_tLocalizedMessages = ResourceBundle.getBundle(GridLifeConstants.GOL_FILE_LOCALIZED_MESSAGES);
 }
 
 
 
 /**
  * At startup create and show the main frame of the application.
  * Also, internal frames are set up and opened.
  * 
  */
 @Override protected void startup() 
 {
  m_tColorManager        = new GridLifeColorManager()                      ; // global color manager
  m_tShapeSetManager     = new GridLifeShapeSetManager()                   ; // global shapeset manager
  m_tStatisticsManager   = new GridLifeStatisticsManager()                 ; // global statistics manager
  m_tAgeHistogramManager = new GridLifeAgeHistogramManager(m_tColorManager); // global histogram manager
  
  //
  // Main window allocation and opening
  //
  if (null != (m_tMainWindow = new GridLifeMainWindow((GridLifeGuiConnector) this)))
  {
   m_tMainWindow.addWindowListener((WindowListener) this); 
   show(m_tMainWindow);
  
   //
   // prepare windows settings object
   //
   if (null != (m_tWindowSettings = new GridLifeWindowSettings()))
   {
    //
    // Palette editor window
    //
    if (null != (m_tPaletteEditorWindow = new GridLifePaletteEditorWindow((GridLifeGuiConnector) this)))
    {
     m_tMainWindow.add(m_tPaletteEditorWindow);
     m_tWindowSettings.addFrame(m_tPaletteEditorWindow, GridLifeConstants.GOL_WINDOW_PALETTE_EDITOR);
     m_tPaletteEditorWindow.addInternalFrameListener((InternalFrameListener) this);
    }
     
    //
    // Cell matrix window
    //
    if (null != (m_tCellMatrixWindow = new GridLifeCellMatrixWindow((GridLifeGuiConnector) this,
                                                                    (DocumentListener    ) this)))
    {
     m_tMainWindow.add(m_tCellMatrixWindow);
     m_tWindowSettings.addFrame(m_tCellMatrixWindow, GridLifeConstants.GOL_WINDOW_CELL_MATRIX);
     m_tCellMatrixWindow.addInternalFrameListener((InternalFrameListener) this);
     m_tCellMatrixWindow.setVisible(true);
    }
   
    //
    // Properties window
    //
    if (null != (m_tPropertiesWindow = new GridLifePropertiesWindow((GridLifeGuiConnector) this)))   
    {
     m_tMainWindow.add(m_tPropertiesWindow);
     m_tWindowSettings.addFrame(m_tPropertiesWindow, GridLifeConstants.GOL_WINDOW_PROPERTIES);
     m_tPropertiesWindow.addInternalFrameListener((InternalFrameListener) this);
    }
   
    //
    // Statistics window
    //
    if (null != (m_tStatisticsWindow  = new GridLifeStatisticsWindow()))   
    {
     m_tMainWindow.add(m_tStatisticsWindow);
     m_tWindowSettings.addFrame(m_tStatisticsWindow, GridLifeConstants.GOL_WINDOW_STATISTICS);
     m_tStatisticsWindow.addInternalFrameListener ((InternalFrameListener) this);
    }
   
    //
    // Shape palette window
    //
    if (null != (m_tShapePaletteWindow = new GridLifeShapePaletteWindow((GridLifeGuiConnector) this)))
    {
     m_tMainWindow.add(m_tShapePaletteWindow);
     m_tWindowSettings.addFrame(m_tShapePaletteWindow, GridLifeConstants.GOL_WINDOW_SHAPE_PALETTE);
     m_tShapePaletteWindow.addInternalFrameListener((InternalFrameListener) this);
    }    
    
    //
    // Color window
    //
    if (null != (m_tColorWindow = new GridLifeColorWindow((GridLifeGuiConnector) this)))
    {
     m_tMainWindow.add(m_tColorWindow);
     m_tWindowSettings.addFrame(m_tColorWindow, GridLifeConstants.GOL_WINDOW_COLOR);
     m_tColorWindow.addInternalFrameListener((InternalFrameListener) this);
    }        
    
    //
    // Quickslot dialog
    //
    m_tQuickSlotDialog = new GridLifeQuickSlotDialog(m_tMainWindow, true, (GridLifeGuiConnector) this);
   }
  }
 }

  
  
 /**
  * This method is to initialize the specified window by injecting resources.
  * Windows shown in our application come fully initialized from the GUI
  * builder, so this additional configuration is not needed.
  */
 @Override 
 protected void configureWindow(java.awt.Window root){}

  
  
  /**
   * A convenient static getter for the application instance.
   * @return the instance of GridLifeApp
   */
  public static GridLifeApp getApplication() {return Application.getInstance(GridLifeApp.class);}

   
   
  /**
   * Main method launching the application.
   */
  public static void main(String[] args) 
  {
   //
   // launch application
   //
   launch(GridLifeApp.class, args);
  }

  
 
/****************************************** WindowListener methods ***********************************************************************/ 
/**
 * WindowListener.windowClosing()
 * 
 * @param pa_tEvent  Event
 */ 
 public void windowClosing(WindowEvent pa_tEvent) 
 {
  Object lc_tWindowObject = null;
 
  if (null != (lc_tWindowObject = pa_tEvent.getSource()))
  {
   if (lc_tWindowObject instanceof JFrame)
   {
    if ((JFrame) lc_tWindowObject == m_tMainWindow)   
    {
     handleProgramExit();
    }
   }
  }
 }

 
 
/**
 * WindowListener.windowClosed()
 * 
 * @param pa_tEvent  Event
 */ 
 public void windowClosed(WindowEvent pa_tEvent) 
 {}

 
 
/**
 * WindowListener.windowOpened()
 *
 * As soon as the main window is open, do all the intialization stuff
 *
 * @param pa_tEvent  Event
 */ 
public void windowOpened(WindowEvent pa_tEvent) 
{
 Object             lc_tWindowObject = null;
 GridLifeProperties lc_tProperties   = null;
 
 if (null != (lc_tWindowObject = pa_tEvent.getSource()))
 {
  //
  // check whether the main window was opened
  //
  if ((JFrame) lc_tWindowObject == m_tMainWindow)   
  {
   //
   // now, adjust windows to previous settings
   //
   if (null != m_tWindowSettings)
   {
    if (true == m_tWindowSettings.isValid())
    {
     //
     // check whether current settings fit the current screen settings
     //
     if (true == checkWindowSettings()) 
     {
      m_tWindowSettings.adjustFrames();
     }
     else
     {
      createDefaultWindowSetup(true, true);       
     }
    }
    else
    {
     createDefaultWindowSetup(true, false);
    }
   }
   else
   {
    createDefaultWindowSetup(true, false); 
   }
   
   
   //
   // set menu checkboxes according to window states,
   // install look and feel menu and switch to the
   // ui defined in window settings
   //
   updateMainMenu();
   installLFMenu();
   
   //
   // update palette window
   //
   if (null != m_tShapePaletteWindow && null != m_tShapeSetManager)
   {
    m_tShapePaletteWindow.updatePalettes(m_tShapeSetManager, (null != m_tWindowSettings ? m_tWindowSettings.getShapePaletteName() : null));
    if (null != m_tShapeSetManager.getShapeSet(m_tShapePaletteWindow.getSelectedPaletteName()))
    {
     m_tShapePaletteWindow.initList(m_tShapeSetManager.getShapeSetShapes(m_tShapePaletteWindow.getSelectedPaletteName()), (null != m_tWindowSettings ? m_tWindowSettings.getShapePaletteEntryName() : null));
    }
   }
   
   //
   // update color window, check if a set is active and if so, propagate this information to the ColorManager.
   // The call of initList() also modifies the ColorManager, it preselects the first element of the combobox
   // as current color set
   //
   if (null != m_tColorManager && null != m_tColorWindow)
   {
    if (null != m_tWindowSettings) m_tColorManager.setActiveSetName(m_tWindowSettings.getColorPaletteName());

    m_tColorWindow.updateColorSets(m_tColorManager, (null != m_tWindowSettings ? m_tWindowSettings.getColorPaletteName() : null));    
    m_tColorWindow.initList(m_tColorManager.getColorSetRecords(m_tColorManager.getActiveSetName()),(null != m_tWindowSettings ? m_tWindowSettings.getColorPaletteEntryName() : null));
   }
   
   //
   // initialize properties, cell matrix and cell grid window
   //
   if (null != m_tCellMatrixWindow)
   {
    if (null != m_tWindowSettings)
    {
     m_iZoomFactor = m_tWindowSettings.getZoom();
     m_iFillRatio  = m_tWindowSettings.getFill();
     m_tCellMatrixWindow.setPaintMode(m_tWindowSettings.getPaintMode());
     m_tCellMatrixWindow.setPaintType(m_tWindowSettings.getPaintType());
    }
    m_tCellMatrixWindow.setZoomValue (m_iZoomFactor);
    m_tCellMatrixWindow.setFillRatio (m_iFillRatio );
    m_tCellMatrixWindow.setAdjustmentListener((AdjustmentListener) this);
    m_tCellMatrixWindow.setSlot(null != m_tWindowSettings ? m_tWindowSettings.getQuickSaveSlot() : null);
   }
   
   //
   // try to load properties, if this fails, use default properties
   //
   if (null != (lc_tProperties = loadProperties(GridLifeConstants.GOL_FILE_PROPERTIES_NAME)))
   {
    m_iSceenshotCellSize = lc_tProperties.getScreenshotCellSize(); 
    m_bInvertMouseWheel  = (1 == lc_tProperties.getInvertMouseWheel()); 
            
    m_tCellMatrix        = new GridLifeCellMatrix(lc_tProperties, m_tColorManager, m_tAgeHistogramManager);  
    if (null != m_tPropertiesWindow) m_tPropertiesWindow.setProperties(m_tCellMatrix.getProperties());
    
   }
   else
   {
    m_tCellMatrix = new GridLifeCellMatrix(m_tPropertiesWindow.getProperties(), m_tColorManager, m_tAgeHistogramManager); 
   }
   
   //
   // head on with cell paint panel
   //
   if (null != m_tCellMatrix)
   {
    if (null != (m_tCellPaintPanel = new GridLifeCellMatrixPaintPanel(m_tCellMatrix.getCells(), m_tCellMatrixWindow.getCellScrollPane(), m_tCellMatrix.getProperties(), m_tColorManager, 
                                                                       m_iZoomFactor + GridLifeConstants.GOL_CELL_MIN_SIZE)))
    {
     //
     // attach special paint panel to cell matrix window, poke mouse listener into it
     //
     m_tCellPaintPanel.setPaletteShape(m_tShapePaletteWindow.getSelectedShape());
     m_tCellPaintPanel.addMouseListener((MouseListener) this);
     m_tCellPaintPanel.addMouseMotionListener((MouseMotionListener) this); 
     m_tCellPaintPanel.addMouseWheelListener ((MouseWheelListener ) this); 
     
     m_tCellMatrixWindow.setCellPaintPanel(m_tCellPaintPanel);
     m_tCellPaintPanel.changeCellSize(GridLifeConstants.GOL_CELL_MIN_SIZE + m_iZoomFactor);
     m_tCellPaintPanel.setShapeXorPaintMode(m_tCellMatrixWindow.isShapePaintXorModeActive());
    }
   }
            
   //
   // update statistics window
   //
   if (null != m_tStatisticsWindow)
   {
    if (null != (m_tStatisticsPaintPanel = new GridLifeStatisticsPaintPanel(m_tStatisticsWindow.getStatisticsScrollPane(), m_tCellMatrix.getProperties(), m_tStatisticsManager)))      
    {
     //
     // attach graph paint panel to cell matrix window
     //
     m_tStatisticsWindow.setStatisticsPaintPanel(m_tStatisticsPaintPanel);         
    }
    if (null != (m_tHistogramPaintPanel = new GridLifeAgeHistogramPaintPanel(m_tStatisticsWindow.getHistogramScrollPane(), m_tCellMatrix.getProperties(), m_tAgeHistogramManager, m_tColorManager)))
    { 
     //
     // attach histogram paint panel to cell matrix window
     //
     m_tStatisticsWindow.setHistogramPaintPanel(m_tHistogramPaintPanel);    
     if (0 < m_tHistogramPaintPanel.checkForEnlargement(m_tColorManager.getActiveSet())) m_tStatisticsWindow.revalidate();     
    }
    m_tStatisticsWindow.setMaxCells(m_tCellMatrix.getColumns() * m_tCellMatrix.getRows());
    m_tStatisticsWindow.updateStatistics(getStatisticsRecord()); 
   }
   
   //
   // initialize cell fill color according to color sets and update cell paint panel with it
   //
   if (null != m_tColorWindow && null != m_tCellPaintPanel)
   {
    m_iCellSetAge       = m_tColorWindow.getColorSelectionAge();   
    m_tCellAgeFillColor = m_tColorManager.getColorForSet(m_tColorWindow.getSelectedColorSetName(), m_iCellSetAge) ;
    m_tCellPaintPanel.setCellFillColor(m_tCellAgeFillColor);    
   }
   
   //
   // update button states according to new data
   //
   checkButtonStates();  
   checkButtonStatesQuickSave(true);
   checkButtonStatesPropertiesWindow();
   checkButtonStatesPaletteEditorWindow();
   if (null != m_tCellPaintPanel) m_tCellPaintPanel.changeCellSize(m_iZoomFactor + GridLifeConstants.GOL_CELL_MIN_SIZE);   
   
   
   //
   // finally, select L&F
   //
   initLF();
   
   //
   // check for successful creation of all needed resources, exit, if any is missing
   //
   if (null == m_tCellMatrix         || null == m_tMainWindow         || null == m_tCellMatrixWindow    ||
       null == m_tShapeSetManager    || null == m_tShapePaletteWindow || null == m_tCellPaintPanel      ||
       null == m_tColorManager       || null == m_tColorWindow        || null == m_tPropertiesWindow    ||
       null == m_tStatisticsManager  || null == m_tStatisticsWindow   || null == m_tHistogramPaintPanel ||
       null == m_tAgeHistogramManager)
   {
    System.exit(0); 
   }
  }
 }
}



/**
 * WindowListener.windowIconified()
 * 
 * @param pa_tEvent  Event
 */ 
public void windowIconified(WindowEvent pa_tEvent) 
{}



/**
 * WindowListener.windowDeiconified()
 * 
 * @param pa_tEvent  Event
 */ 
public void windowDeiconified(WindowEvent pa_tEvent) 
{}



/**
 * WindowListener.windowActivated()
 * 
 * @param pa_tEvent  Event
 */ 
public void windowActivated(WindowEvent pa_tEvent) 
{}



/**
 * WindowListener.windowDeactivated()
 * 
 * @param pa_tEvent  Event
 */ 
public void windowDeactivated(WindowEvent pa_tEvent) 
{}



/****************************************** InternalFrameListener methods ***********************************************************************/
public void internalFrameDeactivated(InternalFrameEvent pa_tEvent)
{}

public void internalFrameActivated(InternalFrameEvent pa_tEvent)
{}

public void internalFrameIconified(InternalFrameEvent pa_tEvent)
{}

public void internalFrameDeiconified(InternalFrameEvent pa_tEvent)
{}

public void internalFrameOpened(InternalFrameEvent pa_tEvent)
{}

public void internalFrameClosed(InternalFrameEvent pa_tEvent)
{}



/**
 * JInternalFrameListener.internalFrameClosing
 * 
 * @param pa_tEvent   Event
 */
public void internalFrameClosing(InternalFrameEvent pa_tEvent)
{
 Object lc_tWindowObject = null;
 
 if (null != (lc_tWindowObject = pa_tEvent.getSource()))
 {
  if (lc_tWindowObject instanceof JInternalFrame)
  {
   if ((JInternalFrame) lc_tWindowObject == m_tPropertiesWindow)   
   {
    m_tPropertiesWindow.setVisible(false);
   }
   if ((JInternalFrame) lc_tWindowObject == m_tStatisticsWindow)   
   {
    m_tStatisticsWindow.setVisible(false);
   }
   if ((JInternalFrame) lc_tWindowObject == m_tShapePaletteWindow)   
   {
    m_tShapePaletteWindow.setVisible(false);
   }
   if ((JInternalFrame) lc_tWindowObject == m_tPaletteEditorWindow)   
   {
    m_tPaletteEditorWindow.setVisible(false);
   }           
   if ((JInternalFrame) lc_tWindowObject == m_tColorWindow)   
   {
    m_tColorWindow.setVisible(false);
   }
   updateMainMenu();
  }
 }
}



 /****************************************** GridLifeGuiConnector methods ***********************************************************************/
 /**
  * Handles changes in the combobox for showing help lines
  * 
  */
 public void handleCombobox_PR_ShowHelpLines() {checkButtonStatesPropertiesWindow();}
         
 
 
 /**
  * Handles changes in the combobox for showing cell borders
  * 
  */
 public void handleCombobox_PR_CellBorder()  {checkButtonStatesPropertiesWindow();}
 
 
 
 /**
  * Handles changes in the combobox for inverting the mouse wheel
  * 
  */
 public void handleCombobox_PR_InvertMouseWheel()  
 {
  checkButtonStatesPropertiesWindow();
 }
 
 
  /**
  * Handles changes in the combobox for showing help lines
  * 
  */
 public void handleCombobox_PR_GridLines() {checkButtonStatesPropertiesWindow();}
    
 
 
 /**
  * Handles changes in the combobox for showing cell hilite lines
  * 
  */
 public void  handleCombobox_PR_CellHiliteLines() {checkButtonStatesPropertiesWindow();}
 
 
 
 /**
  * Handles changes in the topology combobox
  * 
  */ 
 public void handleCombobox_PR_Topology() {checkButtonStatesPropertiesWindow();}



 /**
  * Handles changes in the topology combobox
  * 
  */ 
 public void handleCombobox_PR_Neighbourhood() {checkButtonStatesPropertiesWindow();}

 
 
 /**
  * Handles changes to any spinner in the properties window
  * 
  */
 public void handleSpinner_PR_PropertyChanged() {checkButtonStatesPropertiesWindow();}

 
 
 /**
  * Handles clicking any ruleset togglebutton
  * 
  * @param pa_tEvent    item event
  */
 public void handleToggleButton_PR_Ruleset(ItemEvent pa_tEvent) {checkButtonStatesPropertiesWindow();}
 
 

 /**
  * Handles clicking the Apply button in the properties window
  * 
  */
 public void handleButton_PR_Apply()
 {
  GridLifeProperties lc_tRuleset = null ;
  
  boolean lc_bContinue  = true ;
  boolean lc_bRealloc   = false;  
  int     lc_iOldAmount =     0;
  
  if (null != m_tCellMatrix && null != m_tPropertiesWindow)
  {
   lc_tRuleset   = m_tPropertiesWindow.getProperties();
   lc_iOldAmount = m_tCellMatrix.getCellAmount();
    
   //
   // if we're about to shrink the matrix, we ask for permission
   //
   if (0 < lc_iOldAmount)
   {
    if (true == m_tCellMatrix.checkForShrinkingRealloc(lc_tRuleset))
    {
     lc_bContinue = (0 == JOptionPane.showOptionDialog(m_tMainWindow,  
                                                       getLocalizedString(GridLifeConstants.GOL_DIALOG_SHRINKING  ) , 
                                                       getLocalizedString(GridLifeConstants.GOL_TITLE_CONFIRMATION) , 
                                                       JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, 
                                                       getLocalizedYesNoOptions()                                   , 
                                                       getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES  )));
    } 
   }
    
   //
   // okay, do it!
   //
   if (true == lc_bContinue)
   {
    //
    // check if we are resizing the cell array... setNewProperties() also invokes any drawing update needed
    //
    lc_bRealloc = setNewProperties(lc_tRuleset);
    if (null != m_tStatisticsWindow) m_tStatisticsWindow.setMaxCells(m_tCellMatrix.getColumns() * m_tCellMatrix.getRows());
    
    if (true == lc_bRealloc)
    {
     //
     // care for changed flag
     //
     setModifiedFlag();   

     //
     // have some fun
     //
     if (lc_iOldAmount > m_tCellMatrix.getCellAmount())
     {
      JOptionPane.showMessageDialog(m_tMainWindow, 
                                    getLocalizedString(GridLifeConstants.GOL_DIALOG_TINY_VOICES), 
                                    getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION ), 
                                    JOptionPane.INFORMATION_MESSAGE);
     }
    }
   }
  }
 }
   
  
 
 /**
  * Handles clicking the Reset button in the properties window
  * 
  */
 public  void handleButton_PR_Reset()
 {
  if (null != m_tPropertiesWindow)
  {
   m_tPropertiesWindow.setProperties(m_tCellMatrix.getProperties());
   checkButtonStatesPropertiesWindow();
  }
 }

  

 /**
  * Handles clicking the color buttons
  * 
  * @param pa_iWhich button id. o = bg, 1 = grid, 2 = helpline, 3 = cell hilight, 4 = cell borders, 
  *                             5 = selection frame, 6 = selection area, 7 = insertion area, 8 = insertion frame
  */
 public void handleButton_PR_Color(int pa_iWhich, Color pa_tOldColor)
 {
  Color lc_tOldColor = pa_tOldColor; 
  Color lc_tNewColor = JColorChooser.showDialog(m_tMainWindow, getLocalizedString(GridLifeConstants.GOL_TITLE_COLOR), lc_tOldColor);
   
  if (null != lc_tNewColor && null != lc_tOldColor && null != m_tPropertiesWindow)
  {
   if (lc_tNewColor.getRGB() != pa_tOldColor.getRGB()) 
   {
    switch(pa_iWhich)
    {
     case 0: m_tPropertiesWindow.setColorBg                (lc_tNewColor); break;
     case 1: m_tPropertiesWindow.setColorGridLines         (lc_tNewColor); break;
     case 2: m_tPropertiesWindow.setColorHelpLines         (lc_tNewColor); break;
     case 3: m_tPropertiesWindow.setColorCellHilightLines  (lc_tNewColor); break;
     case 4: m_tPropertiesWindow.setColorCellBorders       (lc_tNewColor); break;
     case 5: m_tPropertiesWindow.setColorCellSelectionFrame(lc_tNewColor); break;
     case 6: m_tPropertiesWindow.setColorCellSelectionArea (lc_tNewColor); break;
     case 7: m_tPropertiesWindow.setColorCellInsertionArea (lc_tNewColor); break;
     case 8: m_tPropertiesWindow.setColorCellInsertionFrame(lc_tNewColor); break;
     default:
    }
   }
   checkButtonStatesPropertiesWindow();
  }
 }
 
 

 /**
  * Handles clicking the Save as button in the properties window
  * 
  */
 public void handleButton_PR_SaveAs()
 {
  GridLifeProperties lc_tProperties = null;
  
  boolean lc_bSave      = true;
  String  lc_sFileName  = null ;
  
  String       lc_sSavePath    = (null != m_tWindowSettings ? m_tWindowSettings.getSavePathPropertiesFile() : null              );
  JFileChooser lc_tFileChooser = (null != lc_sSavePath      ? new JFileChooser(lc_sSavePath)                : new JFileChooser());
  
  if (null != lc_tFileChooser)
  {
   lc_tFileChooser.setDialogTitle      (getLocalizedString(GridLifeConstants.GOL_TITLE_PROPERTIES_SAVE));
   lc_tFileChooser.setApproveButtonText(getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_OK    ));

   if (JFileChooser.APPROVE_OPTION == lc_tFileChooser.showOpenDialog(m_tMainWindow))
   {
    if (null != (lc_sFileName = lc_tFileChooser.getSelectedFile().getAbsolutePath()))
    {
     //
     // check if the file exists and confirm or cancel overwriting
     //
     if (true == existsFile(lc_sFileName))
     {
      lc_bSave = (0 == JOptionPane.showOptionDialog(m_tMainWindow, 
                                                    getLocalizedString(GridLifeConstants.GOL_DIALOG_FILE_EXISTS) + lc_sFileName + " ? <br />", 
                                                    getLocalizedString(GridLifeConstants.GOL_TITLE_WARNING), 
                                                    JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, 
                                                    getLocalizedYesNoOptions(), getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES)));
     }      
     if (true == lc_bSave)
     {
      if (null != (lc_tProperties =  m_tPropertiesWindow.getProperties()))
      {
       lc_tProperties.save(lc_sFileName);
       if (null != m_tWindowSettings) m_tWindowSettings.setSavePathPropertiesFile(lc_sFileName);
      }
     }
    }
   }
  }   
 }
 
 
 
 /**
  * Handles clicking the Load button in the properties window
  * 
  */
 public void handleButton_PR_Load()
 {
  GridLifeProperties lc_tProperties = null;
  
  String lc_sLoadPath   = (null != m_tWindowSettings ? m_tWindowSettings.getLoadPathPropertiesFile() : null);
  String lc_sFileName   = null ;
  boolean lc_bRealloc   = false;
  int     lc_iOldAmount = 0    ;
  
  JFileChooser lc_tFileChooser = (null != lc_sLoadPath ? new JFileChooser(lc_sLoadPath) : new JFileChooser());
  
  if (null != lc_tFileChooser && null != m_tCellMatrixWindow && null != m_tStatisticsManager && null != m_tCellMatrix)
  {
   lc_iOldAmount = m_tCellMatrix.getCellAmount();
   
   lc_tFileChooser.setDialogTitle(getLocalizedString(GridLifeConstants.GOL_TITLE_PROPERTIES_LOAD));
   lc_tFileChooser.setApproveButtonText(getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_OK));
   
   if (JFileChooser.APPROVE_OPTION == lc_tFileChooser.showOpenDialog(m_tMainWindow))
   {
    if (null != (lc_sFileName = lc_tFileChooser.getSelectedFile().getAbsolutePath()))
    {
     if (null != (lc_tProperties = loadProperties(lc_sFileName)))
     {
      lc_bRealloc = setNewProperties(lc_tProperties);
      
      if (null != m_tWindowSettings  ) m_tWindowSettings.setLoadPathPropertiesFile(lc_sFileName);
      if (null != m_tStatisticsWindow) m_tStatisticsWindow.setMaxCells(m_tCellMatrix.getColumns() * m_tCellMatrix.getRows());

      if (true == lc_bRealloc)
      {
       if (lc_iOldAmount > m_tCellMatrix.getCellAmount())
       {
        JOptionPane.showMessageDialog(m_tMainWindow, 
                                      getLocalizedString(GridLifeConstants.GOL_DIALOG_TINY_VOICES), 
                                      getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION ), 
                                      JOptionPane.INFORMATION_MESSAGE);
       }
      }
     }
    }
   }
  }
 }


  
 
 /**
  * Handles clicking the zoom into button
  * 
  */
 public void handleButton_CM_ZoomInto()
 {
  int lc_iZoom = 0;
  
  if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   lc_iZoom = m_tCellPaintPanel.getZoomFactorForSelectionZoom(m_tCellMatrixWindow.getMaxZoomValue()); 
   
   m_tCellMatrixWindow.setZoomValue(lc_iZoom);
   m_tCellPaintPanel.scrollToSelectionForZoom();
  }
 }

 
 
 /**
  * Handles clicking the select all button
  * 
  */
 public void handleButton_CM_SelectAll()
 {
  if (null != m_tCellMatrix && null != m_tCellPaintPanel)
  {
   m_tCellPaintPanel.setCellSelectionStartPosition(new Point(0,0)); 
   m_tCellPaintPanel.setCellSelectionEndPosition  (new Point(m_tCellMatrix.getColumns(),m_tCellMatrix.getRows())); 
   m_tCellPaintPanel.repaint();
   checkButtonStates();
  }
 }
 
 
 
 /**
  * Makes a screenshot od the cell matrix
  * 
  */
 public void handleButton_CM_Screenshot()
 {
  String  lc_sFileName  = GridLifeConstants.GOL_FILE_SCREENSHOT_DIR + File.separator + Long.toString(System.nanoTime()) + ".png"; 
  boolean lc_bAnimState = false;
  
  long    lc_lAvailableMemorySize = Runtime.getRuntime().freeMemory();
  long    lc_lImageMemorySize     = 0;
  int     lc_iOldCellSize         = 0;
  
  BufferedImage                lc_tImage      = null;
  GridLifeSaveScreenshotThread lc_tSaveThread = null;
  
  if (null != m_tCellPaintPanel)
  {
   lc_iOldCellSize = m_tCellPaintPanel.getCellSize();
   
   //
   // enable screenshot mode, stop a runnig life animation
   //
   lc_bAnimState = stopLifeAnimation();
   m_tCellPaintPanel.setScreenshotMode(true);
   m_tCellPaintPanel.changeCellSize(m_iSceenshotCellSize); 
   
   lc_lImageMemorySize = m_tCellPaintPanel.getSize().width * m_tCellPaintPanel.getSize().height;    
   if (lc_lImageMemorySize <= lc_lAvailableMemorySize)
   {
    //
    // create image from matrix paint panel and save it. the savethread will call flush() on the created image to free associated bitmap memory.
    //
    try
    {
     if (null != (lc_tImage = new BufferedImage(m_tCellPaintPanel.getSize().width, m_tCellPaintPanel.getSize().height, BufferedImage.TYPE_INT_RGB)))
     {
      m_tCellPaintPanel.paint(lc_tImage.createGraphics());
     
      //
      // saving screenshots uses a thread to keep the GUI's repant mechanism intact while saving. The thread knows about the GridLifeGuiConnector
      // and will use it to indicate that a screenshot thread has started and ended. This enables us to prevent exiting the program as long as
      // there are screenshot threads saving data.
      //
      if (null != (lc_tSaveThread = new GridLifeSaveScreenshotThread(lc_sFileName, lc_tImage, (GridLifeGuiConnector) this)))
      {
       lc_tSaveThread.start();     
      }   
     }
    }
    //
    // catch errors caused by insufficient memory
    //
    catch (java.lang.OutOfMemoryError lc_tException) {showMemoryAlert();}
    catch (Exception                  lc_tException) {}   // catch anything else. shit happens.
   }
   else 
   {
    showMemoryAlert();
   }

   //
   // disable screenshot mode, restart life animation, if it was active before screenshotting
   //
   m_tCellPaintPanel.setScreenshotMode(false); 
   m_tCellPaintPanel.changeCellSize(lc_iOldCellSize); 
   
   if (true == lc_bAnimState)
   {
    startLifeAnimation(); 
   }
  }
 }
 

 
 /**
  * Handles pressing the Fill button in the cell matrix window
  * 
  */
  public void handleButton_CM_Fill()
  {
   int lc_iResult = 0;
   
   if (null != m_tCellMatrix && null != m_tCellMatrixWindow && null != m_tStatisticsManager && null != m_tStatisticsPaintPanel && null != m_tStatisticsWindow && null != m_tHistogramPaintPanel)
   {
    if (0 < m_tCellMatrix.getCellAmount())
    {
     lc_iResult = JOptionPane.showOptionDialog(m_tMainWindow, 
                                               getLocalizedString(GridLifeConstants.GOL_DIALOG_RANDOM      ), 
                                               getLocalizedString(GridLifeConstants.GOL_TITLE_CONFIRMATION ), 
                                               JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, 
                                               getLocalizedYesNoOptions()                                   , 
                                               getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES  ));
    }
    //
    // if we go on, fill the matrix etc pp
    //
    if (0 == lc_iResult)
    {
     SwingUtilities.invokeLater(new Runnable() 
                                {
                                 public void run() 
                                 {     
                                  m_iFillRatio = m_tCellMatrixWindow.getFillRatio(); 
                                  m_tCellMatrix.randomFillCellMatrix(m_iFillRatio, m_iCellSetAge); 

                                  m_tStatisticsManager.clear();
                                  m_tCellPaintPanel.repaint();
                                  m_tStatisticsPaintPanel.repaint();

                                  setModifiedFlag();

                                  //
                                  // update statistics window and histogram
                                  //
                                  m_tStatisticsWindow.updateStatistics(getStatisticsRecord()); 
                                  m_tHistogramPaintPanel.repaint();

                                  //
                                  // update button states according to new data
                                  //
                                  checkButtonStates();
                                  checkButtonStatesQuickSave(false);
                                 }
                                }
                               );
   }
  }
 }

  
  
 /**
  * Handles pressing the Clear button in the cell matrix window
  * 
  */
  public void handleButton_CM_Clear()
  {
   int lc_iResult = 0;
   int lc_iAmount = 0;
   
   if (null != m_tCellMatrix && null != m_tCellMatrixWindow && null != m_tStatisticsManager && null != m_tStatisticsPaintPanel && null != m_tStatisticsWindow && null != m_tHistogramPaintPanel)
   {
    if (0 < (lc_iAmount = m_tCellMatrix.getCellAmount()))
    {     
     lc_iResult = JOptionPane.showOptionDialog(m_tMainWindow, 
                                               getLocalizedString(GridLifeConstants.GOL_DIALOG_CLEAR       ), 
                                               getLocalizedString(GridLifeConstants.GOL_TITLE_CONFIRMATION ), 
                                               JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, 
                                               getLocalizedYesNoOptions()                                   , 
                                               getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES  ));
     //
     // if we go on, clean the matrix etc pp
     //
     if (0 == lc_iResult)
     {
      m_tCellMatrix.resetCellMatrix(); // this also resets histogram manager
       
      m_tStatisticsPaintPanel.repaint();
      m_tHistogramPaintPanel.repaint();
      m_tCellPaintPanel.clearCellSelection();
      m_tCellPaintPanel.repaint();
       
      m_tCellMatrixWindow.setWindowTitle(getLocalizedString(GridLifeConstants.GOL_WINDOW_CELL_MATRIX_TITLE));
      m_tStatisticsManager.clear();
     }
    }
    
    //
    // update statistics window and repaint graph and histogram
    //
    m_tStatisticsWindow.updateStatistics(getStatisticsRecord()); 
    m_tHistogramPaintPanel.repaint();
    
    //
    // update button states according to new data
    //
    checkButtonStates();   
    checkButtonStatesQuickSave(false);

    //
    // have fun
    //
    if (0 < lc_iAmount && 0 == lc_iResult)
    {
     JOptionPane.showMessageDialog(m_tMainWindow, 
                                   getLocalizedString(GridLifeConstants.GOL_DIALOG_AFTER_CLEAR), 
                                   getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION ), 
                                   JOptionPane.INFORMATION_MESSAGE);
    }
   }
  }
  

  
/**
 * Quick-loads a slot
 * 
 */
 public void handleButton_CM_SlotLoad()
 {
  String lc_sFileName = GridLifeConstants.GOL_FILE_QUICKSAVE_DIR + File.separator ;
  String lc_sSlot     = m_tCellMatrixWindow.getSlot();
  String lc_sFilePath = null;
  
  if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   lc_sFileName += lc_sSlot + GridLifeConstants.GOL_FILE_QUICKSAVE_EXTENSION; 
   m_tCellPaintPanel.clearCellSelection();
  
   lockWindows(true);
   lc_sFilePath = m_tCellMatrix.loadMatrix(lc_sFileName, GridLifeConstants.GOL_FILE_QUICKSAVE_HEADER, 
                                                         GridLifeConstants.GOL_FILE_QUICKSAVE_SEPARATOR,
                                                         GridLifeConstants.GOL_FILE_QUICKSAVE_COMMENT  );
   m_tCellMatrixWindow.setWindowTitle(getLocalizedString(GridLifeConstants.GOL_WINDOW_TITLE_QUICKSLOT) + " " + lc_sSlot);
   lockWindows(false); 

   // 
   // check for successfull loading
   //
   if (null != lc_sFilePath)
   {
    handleLoadSuccess();      
   }
   else
   {
    JOptionPane.showMessageDialog(m_tMainWindow, 
                                  getLocalizedString(GridLifeConstants.GOL_DIALOG_QUICKLOAD_FAILED) + lc_sSlot, 
                                  getLocalizedString(GridLifeConstants.GOL_TITLE_WARNING ), 
                                  JOptionPane.WARNING_MESSAGE);     
   }
  }
 }
 

 
/**
 * Quick-saves a slot
 * 
 */
 public void handleButton_CM_SlotSave()
 {
  String lc_sFileName = GridLifeConstants.GOL_FILE_QUICKSAVE_DIR + File.separator ;
  String lc_sSlot     = m_tCellMatrixWindow.getSlot();
  String lc_sFilePath = null;
  
  if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   lc_sFileName += lc_sSlot + GridLifeConstants.GOL_FILE_QUICKSAVE_EXTENSION;  
   lockWindows(true);
   
   if (null != (lc_sFilePath = m_tCellMatrix.saveMatrix(lc_sFileName, GridLifeConstants.GOL_FILE_QUICKSAVE_HEADER, 
                                                                      GridLifeConstants.GOL_FILE_QUICKSAVE_SEPARATOR,
                                                                      m_tCellMatrixWindow.getZoomValue(),
                                                                      m_tCellPaintPanel.getTopLeftX(),
                                                                      m_tCellPaintPanel.getTopLeftY())))
   {
    resetModifiedFlag(); 
    m_tCellMatrixWindow.setWindowTitle(getLocalizedString(GridLifeConstants.GOL_WINDOW_TITLE_QUICKSLOT) + " " + lc_sSlot);
    checkButtonStatesQuickSave(true);
   }
   else
   {
    JOptionPane.showMessageDialog(m_tMainWindow, 
                                  getLocalizedString(GridLifeConstants.GOL_DIALOG_QUICKSAVE_FAILED) + lc_sSlot, 
                                  getLocalizedString(GridLifeConstants.GOL_TITLE_WARNING ), 
                                  JOptionPane.WARNING_MESSAGE);     
   }   
   lockWindows(false);
  }
 }
 
 
 
 
 /**
  * Handles changing the zoom slider
  * 
  * @param pa_tEvent   ChangeEvent
  */
  public void handleSlider_CM_Zoom(ChangeEvent pa_tEvent)
  {
   int lc_iValue = 0;
   
   if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
   {
    lc_iValue = m_tCellMatrixWindow.getZoomValue();
    
    if (lc_iValue != m_iZoomFactor)
    {
     m_iZoomFactor = lc_iValue;
     m_tCellPaintPanel.changeCellSize(lc_iValue + GridLifeConstants.GOL_CELL_MIN_SIZE); 
    }
   }
  }
  
  

/**
 * Handles a selection of the palette mode combobox
 * 
 */
 public void handleCombobox_CM_PaletteMode()
 {
  boolean lc_bPaletteMode        = false;
  boolean lc_bPaletteEditorMode  = false;
  boolean lc_bCellMatrixMode     = false;
  
  if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   lc_bPaletteMode       = m_tCellMatrixWindow.isShapePaletteSelectionActive();
   lc_bCellMatrixMode    = m_tCellMatrixWindow.isCellMatrixSelectionActive()  ;
   lc_bPaletteEditorMode = m_tCellMatrixWindow.isPaletteEditorSelectionActive();
   
   m_tCellPaintPanel.setPaintFrame(!lc_bCellMatrixMode && !lc_bPaletteMode && !lc_bPaletteEditorMode);
   m_tCellPaintPanel.setPaintCellSelectionFrame(true);
   m_tCellPaintPanel.repaint();
   
   checkButtonStates();
   checkButtonStatesPaletteEditorWindow();
  }
 }
 
 
 
/**
 * Handles changing the palette shape paint mode
 * 
 */
 public void handleCombobox_CM_ShapePaintMode()
 {
  if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   m_tCellPaintPanel.setShapeXorPaintMode(m_tCellMatrixWindow.isShapePaintXorModeActive()); 
  }
 }
 
 
 
 /**
  * Handles hitting the step button
  * 
  */
 public void handleButton_CM_Step()
 {
  computeNextGeneration();
  if (null != m_tStatisticsWindow) m_tStatisticsWindow.setComputingThreads(null);
 }
 


 /**
  * Handles clicking the button Fill with dead cells
  * 
  */
 public void handleButton_CM_FillWithDeadCells()
 {
  fillCellMatrixSelection(0); 
 }
 
 
 
 /**
  * Handles clicking the button Fill with licing cells
  * 
  */
 public void handleButton_CM_FillWithLivingCells()
 {
  fillCellMatrixSelection(1); 
 }
 
 
 
 /**
  * Handles clicking the button Fill with licing cells
  * 
  */
 public void handleButton_CM_InvertCells()
 {
  fillCellMatrixSelection(2);    
 }
 
 
 
 /**
  * Handles hitting the stop button
  * 
  */
 public void handleButton_CM_Stop()
 {
  stopLifeAnimation();  
 }
 
 
 
/**
 * Handles hitting the play button
 * 
 */
 public void handleButton_CM_Play()
 {
  startLifeAnimation(); 
  setModifiedFlag();
 } 
 

 
/**
 * Handles saving the cell matrix
 * 
 */
 public void handleButton_CM_SaveMatrix()
 {  
  int     lc_iPos    = 0   ;
  int     lc_iOption = -1  ;
  boolean lc_bSave   = true;
  
  String lc_sFileName  = null;
  String lc_sFilePath  = null;
  String lc_sExtension = null;
  String lc_sSavePath  = (null != m_tWindowSettings ? m_tWindowSettings.getSavePathCellFile() : null);
  
  JFileChooser  lc_tFileChooser = (null != lc_sSavePath ? new JFileChooser(lc_sSavePath) : new JFileChooser());
  
  if (null != lc_tFileChooser && null != m_tCellMatrix && null != m_tCellMatrixWindow)
  {
   lc_tFileChooser.setDialogTitle(getLocalizedString(GridLifeConstants.GOL_TITLE_CELL_FILE_SAVE));
   lc_tFileChooser.setApproveButtonText(getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_OK));
   
   if (JFileChooser.APPROVE_OPTION == lc_tFileChooser.showOpenDialog(m_tMainWindow))
   {
    if (null != (lc_sFileName = lc_tFileChooser.getSelectedFile().getAbsolutePath()))
    {
     if (-1 != (lc_iPos = lc_sFileName.lastIndexOf(".")))
     {
      if (null != (lc_sExtension = lc_sFileName.substring(lc_iPos)))
      {
       if (false == (lc_sExtension.toLowerCase().endsWith(GridLifeConstants.GOL_FILE_CELL_MATRIX_EXTENSION)))
       {
        lc_sFileName += GridLifeConstants.GOL_FILE_CELL_MATRIX_EXTENSION; 
       } 
      }
     }
     else
     {
      lc_sFileName += GridLifeConstants.GOL_FILE_CELL_MATRIX_EXTENSION;  
     }
     
     //
     // check if the file exists and confirm or cancel overwriting
     //
     if (true == existsFile(lc_sFileName))
     {
      lc_iOption = JOptionPane.showOptionDialog(m_tMainWindow, 
                                                getLocalizedString(GridLifeConstants.GOL_DIALOG_FILE_EXISTS) + lc_sFileName + " ? <br />", 
                                                getLocalizedString(GridLifeConstants.GOL_TITLE_WARNING), 
                                                JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, 
                                                getLocalizedYesNoOptions(), getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES));
      lc_bSave = (0 == lc_iOption);
     }
     
     //
     // save it!
     //
     if (true == lc_bSave)
     {
      lockWindows(true);
      if (null != (lc_sFilePath = m_tCellMatrix.saveMatrix(lc_sFileName, 
                                                           GridLifeConstants.GOL_FILE_CELL_MATRIX_HEADER, 
                                                           GridLifeConstants.GOL_FILE_CELL_MATRIX_VALUE_SEPARATOR,
                                                           m_tCellMatrixWindow.getZoomValue(),
                                                           m_tCellPaintPanel.getTopLeftX(),
                                                           m_tCellPaintPanel.getTopLeftY())))
      {
       resetModifiedFlag(); 
       if (null != m_tWindowSettings) m_tWindowSettings.setSavePathCellFile(lc_sFilePath);
       m_tCellMatrixWindow.setWindowTitle(lc_sFilePath);
      }
      lockWindows(false);
     }
     
     //
     // comment confirmation dialog and saving in general :-)
     //
     if (-1 != lc_iOption)
     {
      if (false == lc_bSave)
      {
       JOptionPane.showMessageDialog(m_tMainWindow, 
                                     getLocalizedString(GridLifeConstants.GOL_DIALOG_CONFIRMATION_CANCEL), 
                                     getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION), JOptionPane.INFORMATION_MESSAGE);     
      }
      else
      {
       JOptionPane.showMessageDialog(m_tMainWindow, 
                                     getLocalizedString(GridLifeConstants.GOL_DIALOG_CONFIRMATION_OK), 
                                     getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION), JOptionPane.INFORMATION_MESSAGE);            
      }
     }
     else
     {
      if (true == lc_bSave)
      {
       JOptionPane.showMessageDialog(m_tMainWindow, 
                                     getLocalizedString(GridLifeConstants.GOL_DIALOG_SAVE_OK), 
                                     getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION), JOptionPane.INFORMATION_MESSAGE);                    
      }
     }
    }
   }
  }
 }
 
 
 
/**
 * Handles loading the cell matrix
 *
 */
 public void handleButton_CM_LoadMatrix()
 {
  String lc_sFileName  = null;
  String lc_sFilePath  = null;
  String lc_sLoadPath  = (null != m_tWindowSettings ? m_tWindowSettings.getLoadPathCellFile() : null);
  
  JFileChooser lc_tFileChooser = (null != lc_sLoadPath ? new JFileChooser(lc_sLoadPath) : new JFileChooser());
  
  if (null != m_tCellMatrix && null != m_tCellMatrixWindow)
  {
   if (0 < m_tCellMatrix.getCellAmount())
   {     
    if (1 ==  JOptionPane.showOptionDialog(m_tMainWindow, 
                                           getLocalizedString(GridLifeConstants.GOL_DIALOG_LOAD        ), 
                                           getLocalizedString(GridLifeConstants.GOL_TITLE_CONFIRMATION ), 
                                           JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, 
                                           getLocalizedYesNoOptions()                                   , 
                                           getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES  )))
    {
     return;  
    }
   }
       
   if (null != lc_tFileChooser)
   {
    lc_tFileChooser.setDialogTitle(getLocalizedString(GridLifeConstants.GOL_TITLE_CELL_FILE_LOAD));
    lc_tFileChooser.setApproveButtonText(getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_OK));
    
    if (JFileChooser.APPROVE_OPTION == lc_tFileChooser.showOpenDialog(m_tMainWindow))
    {
     if (null != (lc_sFileName = lc_tFileChooser.getSelectedFile().getAbsolutePath()))
     {
      //
      // load cell matrix
      // 
      lockWindows(true);
      lc_sFilePath = m_tCellMatrix.loadMatrix(lc_sFileName, GridLifeConstants.GOL_FILE_CELL_MATRIX_HEADER, 
                                                            GridLifeConstants.GOL_FILE_CELL_MATRIX_VALUE_SEPARATOR,
                                                            GridLifeConstants.GOL_FILE_CELL_MATRIX_COMMENT        );
      m_tCellMatrixWindow.setWindowTitle(lc_sFilePath);
      lockWindows(false);

      //
      // check for successfull loading
      //
      if (null != lc_sFilePath)
      {
       if (null != m_tWindowSettings) m_tWindowSettings.setLoadPathCellFile(lc_sFilePath);
       handleLoadSuccess();      
      }
     }
    }     
   }
  }   
 }
 
  
/**
 * Handles clicking the paste button. Generates a shape from the selection and copies it to the paint panel.
 * 
 */
 public void handleButton_CM_Paste()
 {
  int[][]       lc_tSelectionArray    = null;
  GridLifeShape lc_tSelectionShape    = null;
  String        lc_sName              = Long.toString(System.nanoTime());

  int lc_iWidth  = 0;
  int lc_iHeight = 0;

  if (null != (lc_tSelectionArray = getArrayFromCellSelection()))
  {
   if (0 < (lc_iHeight = lc_tSelectionArray.length))
   {
    lc_iWidth  = lc_tSelectionArray[0].length; 
     
    //
    // create shape from int array
    //
    if (null != (lc_tSelectionShape = new GridLifeShape(lc_sName, lc_iWidth, lc_iHeight, lc_tSelectionArray, m_tColorManager, false)))
    {
     //
     // set shape to cell paint panel
     //
     if (null !=m_tCellPaintPanel) m_tCellPaintPanel.setSelectionShape(lc_tSelectionShape);
     checkButtonStates();
    }
   }
  }
  else
  {
   //
   // selection was invalid, because it was empty or an error occured -> void id!
   //
   if (null != m_tCellPaintPanel) m_tCellPaintPanel.clearCellSelection();
   checkButtonStates();
  }
 }
 
 
 
/**
 * Unsets the cell selection shape
 * 
 */
 public void handleButton_CM_Unset()
 {
  if (null != m_tCellPaintPanel) m_tCellPaintPanel.clearSelectionShape();
  checkButtonStates();
 }

 
 
/**
 * Copies the currently selected shape from the shape palette list to the palette editor list
 * 
 */
 public void handleButton_PE_InsertShapePalette()
 {
  String        lc_sName  = null;
  GridLifeShape lc_tShape = null;
  
  if (null != m_tShapePaletteWindow && null != m_tPaletteEditorWindow && null != m_tCellPaintPanel)
  {
   if (null != (lc_sName = requestShapeName(m_tShapePaletteWindow.getSelectedShape().getShapeName())))
   {
    if (null != (lc_tShape = new GridLifeShape(m_tShapePaletteWindow.getSelectedShape(), null, true)))
    {
     lc_tShape.setShapeName(lc_sName);
     m_tPaletteEditorWindow.addShapeToList(lc_tShape); 
     checkButtonStatesPaletteEditorWindow();
     
     m_tCellPaintPanel.setPaletteEditorShape(m_tPaletteEditorWindow.getSelectedShape()); 
     checkButtonStatesPaletteEditorWindow();
    }
   }
  }
 }

 

/**
 * Deletes the currently selected entry from the palette editor shape list
 * 
 */
 public void handleButton_PE_Delete()
 {
  if (null != m_tPaletteEditorWindow && null != m_tCellPaintPanel)
  {
   if (true == m_tPaletteEditorWindow.deleteSelectedShape())
   {
    m_tCellPaintPanel.setPaletteEditorShape(m_tPaletteEditorWindow.getSelectedShape());
    checkButtonStatesPaletteEditorWindow();
   }
  }
 }

 
 
/**
 * Deletes the currently selected entry from the palette editor shape list
 * 
 */
 public void handleButton_PE_Clear()
 {
  if (null != m_tPaletteEditorWindow && null != m_tCellPaintPanel)
  {
   m_tPaletteEditorWindow.clearShapes();
   m_tCellPaintPanel.clearPaletteEditorShape();
   checkButtonStatesPaletteEditorWindow();
  }
 }

 
 
/**
 * Inserts a selection shape or selection area as new entry in the palette editor list
 * 
 */
 public void handleButton_PE_InsertCellMatrix()
 {
  int[][]       lc_tSelectionArray    = null;
  GridLifeShape lc_tSelectionShape    = null;
  String        lc_sName              = null;

  int lc_iWidth  = 0;
  int lc_iHeight = 0;

  if (null != m_tPaletteEditorWindow && null != m_tCellPaintPanel)
  {
   if (null != (lc_tSelectionArray = getArrayFromCellSelection()))
   {
    if (0 < (lc_iHeight = lc_tSelectionArray.length))
    {
     if (0 < (lc_iWidth  = lc_tSelectionArray[0].length))
     {
      if (null != (lc_sName = requestShapeName(Long.toString(System.nanoTime()))))
      {
       //
       // create shape from int array
       //
       if (null != (lc_tSelectionShape = new GridLifeShape(lc_sName, lc_iWidth, lc_iHeight, lc_tSelectionArray, null, true)))
       {
        //
        // set shape to cell paint panel
        //
        m_tPaletteEditorWindow.addShapeToList(lc_tSelectionShape);
        m_tCellPaintPanel.setPaletteEditorShape(m_tPaletteEditorWindow.getSelectedShape()); 
        checkButtonStatesPaletteEditorWindow();
       }
      }
     }
    }
   }
  }
 }
 

 
/**
 * Handles clicking the load button in the PaletteEditorWindow, loads a palette into the palette editor window
 * and updates the palette editor shape for the cell matrix window
 * 
 */
 public void handleButton_PE_Load()
 {
  String lc_sFileName  = null;
  String lc_sLoadPath  = (null != m_tWindowSettings ? m_tWindowSettings.getLoadPathPaletteFile() : null);
  
  GridLifeShapeSet lc_tShapeSet    = null;
  JFileChooser     lc_tFileChooser = (null != lc_sLoadPath ? new JFileChooser(lc_sLoadPath) : new JFileChooser());
  File             lc_tFile        = null;
  
  if (null != m_tPaletteEditorWindow && null != m_tCellPaintPanel)
  {  
   if (null != lc_tFileChooser)
   {
    lc_tFileChooser.setDialogTitle(getLocalizedString(GridLifeConstants.GOL_TITLE_PALETTE_LOAD));
    lc_tFileChooser.setApproveButtonText(getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_OK));
   
    if (JFileChooser.APPROVE_OPTION == lc_tFileChooser.showOpenDialog(m_tMainWindow))
    {
     if (null != (lc_sFileName = lc_tFileChooser.getSelectedFile().getAbsolutePath()))
     {
      if (null != (lc_tFile = new File(lc_sFileName)))
      {
       if (null != (lc_tShapeSet = new GridLifeShapeSet(lc_tFile, null)))
       {
        m_tPaletteEditorWindow.initList(lc_tShapeSet); 
           
        if (null != m_tWindowSettings)
        {
         try                             {m_tWindowSettings.setLoadPathPaletteFile(lc_tFile.getCanonicalPath());}
         catch (Exception lc_tException) {m_tWindowSettings.setLoadPathPaletteFile(lc_sFileName);               }
        }
       }
       m_tCellPaintPanel.setPaletteEditorShape(m_tPaletteEditorWindow.getSelectedShape()); 
       checkButtonStatesPaletteEditorWindow();
      }
     }
    }
   }
  }
 }
 
 
 
/**
 * Handles clicking the load button in the PaletteEditorWindow, saves a palette 
 * 
 */ 
 public void handleButton_PE_Save()
 {
  boolean lc_bSave      = true;
  String  lc_sFileName  = null;
  String  lc_sSavePath  = (null != m_tWindowSettings ? m_tWindowSettings.getSavePathPaletteFile() : null);
  int     lc_iOption    = 0;
  
  File             lc_tFile        = null;
  GridLifeShapeSet lc_tShapeSet    = null;
  JFileChooser     lc_tFileChooser = (null != lc_sSavePath ? new JFileChooser(lc_sSavePath) : new JFileChooser());
  
  if (null != m_tPaletteEditorWindow && null != m_tCellPaintPanel)
  {  
   if (null != lc_tFileChooser)
   {
    lc_tFileChooser.setDialogTitle(getLocalizedString(GridLifeConstants.GOL_TITLE_PALETTE_SAVE));
    lc_tFileChooser.setApproveButtonText(getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_OK));
   
    if (null != (lc_tShapeSet = m_tPaletteEditorWindow.getShapes()))
    {
     if (JFileChooser.APPROVE_OPTION == lc_tFileChooser.showOpenDialog(m_tMainWindow))
     { 
      if (null != (lc_sFileName = lc_tFileChooser.getSelectedFile().getAbsolutePath()))
      {
       //
       // check if the file exists and confirm or cancel overwriting
       //
       if (true == existsFile(lc_sFileName))
       {
        lc_iOption = JOptionPane.showOptionDialog(m_tMainWindow, 
                                                  getLocalizedString(GridLifeConstants.GOL_DIALOG_FILE_EXISTS) + lc_sFileName + " ? <br />", 
                                                  getLocalizedString(GridLifeConstants.GOL_TITLE_WARNING), 
                                                  JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, 
                                                  getLocalizedYesNoOptions(), getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES));
        lc_bSave = (0 == lc_iOption);
       }
      
       //
       // Save it
       //
       if (true == lc_bSave)
       {
        if (null != (lc_tFile = new File(lc_sFileName)))
        {
         lc_tShapeSet.save(lc_tFile); 
         
         if (null != m_tWindowSettings)
         {
          try                             {m_tWindowSettings.setSavePathPaletteFile(lc_tFile.getCanonicalPath());}
          catch (Exception lc_tException) {m_tWindowSettings.setSavePathPaletteFile(lc_sFileName);               }
         }
        }
        checkButtonStatesPaletteEditorWindow();
       }
      }
     }
    }
   }   
  }
 }
 
 
 
/**
 * Handles clicking the shape list
 * 
 */
 public void handleList_PE_Shapes(ListSelectionEvent pa_tEvent)
 {
  GridLifeShape lc_tShape  = null;
  
  if (null != pa_tEvent && null != m_tCellPaintPanel && null != m_tPaletteEditorWindow)
  {
   if (false == pa_tEvent.getValueIsAdjusting())
   {
    if (null != (lc_tShape = m_tPaletteEditorWindow.getSelectedShape()))
    {
     m_tCellPaintPanel.setPaletteEditorShape(lc_tShape); 
    }
   }
  }
 }
 
 
 
  
/**
 * Handles clicking the shape list
 * 
 */
 public void handleList_SP_Palette(ListSelectionEvent pa_tEvent)
 {
  GridLifeShape lc_tShape  = null;
  
  if (null != pa_tEvent && null != m_tCellPaintPanel && null != m_tShapePaletteWindow)
  {
   if (false == pa_tEvent.getValueIsAdjusting())
   {
    if (null != (lc_tShape = m_tShapePaletteWindow.getSelectedShape()))
    {
     m_tCellPaintPanel.setPaletteShape(lc_tShape); 
    }
   }
  }
 }
 
   
/**
 * Handles a selection chance in the palette combobox
 * 
 * @param pa_tEvent   action event
 */
 public void handleCombobox_SP_Palette(java.awt.event.ActionEvent pa_tEvent)  
 {
  if (null != m_tShapePaletteWindow && null != m_tShapeSetManager)
  {
   m_tShapePaletteWindow.initList(m_tShapeSetManager.getShapeSetShapes(m_tShapePaletteWindow.getSelectedPaletteName()), null);
   checkButtonStatesPaletteEditorWindow();
  }
 }

  
 
/**
 * Handles a color set selection. The new color set is propagated tp the color manager, which in turn
 * is used by the CellPaintPanel when drawing cells according to their age
 * 
 * @param pa_tEvent   combobox event
 */  
 public void handleCombobox_AC_Sets(java.awt.event.ActionEvent pa_tEvent)
 {
  String lc_sColorSet = "";

  if (null != m_tColorWindow && null != m_tColorManager && null != m_tAgeHistogramManager && null != m_tHistogramPaintPanel && null != m_tStatisticsWindow && null != m_tCellPaintPanel)
  {
   if (null != (lc_sColorSet =  m_tColorWindow.getSelectedColorSetName()))
   {
    //
    //  init new color list
    //
    m_tColorManager.setActiveSetName(lc_sColorSet);
    m_tColorWindow.initList(m_tColorManager.getColorSetRecords(lc_sColorSet), null);

    //
    // prepare histogram for changes in age structure. switch only when we're not animating, since animation will 
    // repaint on its next calculation based on the new settings anyway (always updating the histogram manager with
    // data by the computing threads, while the histogram paint panel performs a repaint based on the currently 
    // selected colormanager's set.
    //
    m_tAgeHistogramManager.switchToNewColorset(lc_sColorSet, m_tCellMatrix);       
    if (0 < m_tHistogramPaintPanel.checkForEnlargement(m_tColorManager.getColorSet(lc_sColorSet))) {m_tStatisticsWindow.revalidate();}
             
    //
    // repaint affected stuff
    //
    m_tCellPaintPanel.repaint();
    m_tHistogramPaintPanel.repaint();
   }
  }
 }
 
 
 
/**
 * Handles a color set list selection
 * 
 * @param pa_tEvent   list event
 */
 public  void handleList_AC_ColorSet(javax.swing.event.ListSelectionEvent pa_tEvent)
 {
  if (null != pa_tEvent && null != m_tColorWindow && null != m_tColorManager && null != m_tCellPaintPanel)
  {
   if (false == pa_tEvent.getValueIsAdjusting())
   {
    m_iCellSetAge       = m_tColorWindow.getColorSelectionAge()        ;
    m_tCellAgeFillColor = m_tColorManager.getActiveColor(m_iCellSetAge);
    m_tCellPaintPanel.setCellFillColor(m_tCellAgeFillColor)            ;    
   }
  }
 }
 
   
 
/**
 * Handles clicking the checkbox for the properties window
 * 
 */
 public void handleMenuWindowPaletteEditor() 
 {
  if (null != m_tPaletteEditorWindow) m_tPaletteEditorWindow.setVisible(!m_tPaletteEditorWindow.isVisible());
 }
 
 
 
/**
 * Handles clicking the checkbox for the properties window
 * 
 */
 public void handleMenuWindowProperties() 
 {
  if (null != m_tPropertiesWindow) m_tPropertiesWindow.setVisible(!m_tPropertiesWindow.isVisible());
 }

 
 
/**
 * Handles clicking the checkbox for the statistics window
 * 
 */
 public void handleMenuWindowStatistics() 
 {
  if (null != m_tStatisticsWindow) m_tStatisticsWindow.setVisible(!m_tStatisticsWindow.isVisible());
 }

 
 
/**
 * Handles clicking the checkbox for the shape palette window
 * 
 */
 public void handleMenuWindowShapePalette() 
 {
  if (null != m_tShapePaletteWindow) m_tShapePaletteWindow.setVisible(!m_tShapePaletteWindow.isVisible());
 }


 
/**
 * Handles clicking the checkbox for the shape palette window
 * 
 */
 public void handleMenuWindowColor() 
 {
 if (null != m_tColorWindow) m_tColorWindow.setVisible(!m_tColorWindow.isVisible());
 }

  
  
/**
 * Handles choosing the arrange menue item
 * 
 */
 public void handleMenuWindowArrange() {createDefaultWindowSetup(false, true);}
 
 
  
/**
 * Handles clicking the about menu in the main window
 * 
 */
 public void handleMenuAbout() 
 {
  JOptionPane.showMessageDialog(m_tMainWindow, 
                                getLocalizedString(GridLifeConstants.GOL_DIALOG_ABOUT     ), 
                                getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION), 
                                JOptionPane.INFORMATION_MESSAGE);
 }
 
 
 
/**
 * Handles clicking the exit menu in the main window
 * 
 */
 public void handleMenuExit()
 {
  handleProgramExit();
 }
 

 
/**
 * Handles clicking the öpad menu in the main window
 * 
 */
 public void handleMenuLoad()
 {
  handleButton_CM_LoadMatrix();  
 }
 

 
/**
 * Handles clicking the öpad menu in the main window
 * 
 */
 public void handleMenuSave()
 {
  handleButton_CM_SaveMatrix();  
 }

 

/**
 * Handles clicking the öpad menu in the main window
 * 
 */
 public void handleMenuClear()
 {
  handleButton_CM_Clear();  
 }

 
 
/**
 * Prepares application shutdown and finally exits via System.exit()
 * 
 */
 public  void handleProgramExit()
 {
  //
  // stop any running animation
  //    
  boolean lc_bWasRunning = stopLifeAnimation();

  //
  // deny if there are still screenshot threads running
  //
  if (0 != m_iIOThreadsRunning)
  {
   JOptionPane.showMessageDialog(m_tMainWindow, 
                                 getLocalizedString(GridLifeConstants.GOL_DIALOG_SCREENSHOT_RUNNING), 
                                 getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION), JOptionPane.INFORMATION_MESSAGE);     
   return;
  }
   
  if (null != m_tCellMatrix)
  {
   if (0 < m_tCellMatrix.getCellAmount())
   {
    if (true == m_bMatrixModified)
    {
     if (1 == JOptionPane.showOptionDialog(m_tMainWindow, 
                                           getLocalizedString(GridLifeConstants.GOL_DIALOG_EXIT        ), 
                                           getLocalizedString(GridLifeConstants.GOL_TITLE_CONFIRMATION ), 
                                           JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, 
                                           getLocalizedYesNoOptions()                                   , 
                                           getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES  )))
     {
      //
      // restart animation if needed
      //
      if (true == lc_bWasRunning)
      {
       startLifeAnimation(); 
      }
      return;
     }
    }
   }
  }
   
  //
  // store window mode and positions in properties
  //
  if (null != m_tWindowSettings) 
  {
   if (null != m_tLFMenu                               ) m_tWindowSettings.setLookAndFeelName(m_tLFMenu.getSelectedLF());
   if (null != m_tColorManager                         ) m_tWindowSettings.setColorPaletteName(m_tColorManager.getActiveSetName());
   if (null != m_tColorWindow                          ) m_tWindowSettings.setColorPaletteEntryName(Integer.toString(m_tColorWindow.getColorSelectionAge()));
   if (null != m_tShapePaletteWindow                   ) m_tWindowSettings.setShapePaletteName(m_tShapePaletteWindow.getSelectedPaletteName());
   if (null != m_tShapePaletteWindow.getSelectedShape()) m_tWindowSettings.setShapePaletteEntryName(m_tShapePaletteWindow.getSelectedShape().getShapeName());
   
   if (null != m_tCellMatrixWindow) m_tWindowSettings.setPaintMode    (m_tCellMatrixWindow.getPaintMode());
   if (null != m_tCellMatrixWindow) m_tWindowSettings.setPaintType    (m_tCellMatrixWindow.getPaintType());
   if (null != m_tCellMatrixWindow) m_tWindowSettings.setFill         (m_tCellMatrixWindow.getFillRatio());
   if (null != m_tCellMatrixWindow) m_tWindowSettings.setZoom         (m_tCellMatrixWindow.getZoomValue());
   if (null != m_tCellMatrixWindow) m_tWindowSettings.setQuickSaveSlot(m_tCellMatrixWindow.getSlot()     );
   
   m_tWindowSettings.saveProperties(getLocalizedString(GridLifeConstants.GOL_FILE_ANNOTATION));
  }

  //
  // save contents of current properties (those which can be affected in the properties window)
  //
  saveProperties(GridLifeConstants.GOL_FILE_PROPERTIES_NAME);
   
  //
  // close internal frames
  //
  if (null != m_tPropertiesWindow   ) {m_tPropertiesWindow.setVisible   (false); m_tPropertiesWindow.dispose()   ; m_tPropertiesWindow    = null;}
  if (null != m_tStatisticsWindow   ) {m_tStatisticsWindow.setVisible   (false); m_tStatisticsWindow.dispose()   ; m_tStatisticsWindow    = null;}
  if (null != m_tShapePaletteWindow ) {m_tShapePaletteWindow.setVisible (false); m_tShapePaletteWindow.dispose() ; m_tShapePaletteWindow  = null;}
  if (null != m_tCellMatrixWindow   ) {m_tCellMatrixWindow.setVisible   (false); m_tCellMatrixWindow.dispose()   ; m_tCellMatrixWindow    = null;}
  if (null != m_tColorWindow        ) {m_tColorWindow.setVisible        (false); m_tColorWindow.dispose()        ; m_tColorWindow         = null;}
  if (null != m_tPaletteEditorWindow) {m_tPaletteEditorWindow.setVisible(false); m_tPaletteEditorWindow.dispose(); m_tPaletteEditorWindow = null;}

  //
  // dispose used stuff
  //
  if (null != m_tWindowSettings ) m_tWindowSettings.dispose() ; m_tWindowSettings  = null;
  if (null != m_tQuickSlotDialog) m_tQuickSlotDialog.dispose(); m_tQuickSlotDialog = null;
   
  m_tCellPaintPanel = null;
  m_tCellMatrix     = null;

  if (null != m_tShapeSetManager  ) {m_tShapeSetManager.dispose()  ; m_tShapeSetManager   = null;}
  if (null != m_tColorManager     ) {m_tColorManager.dispose()     ; m_tColorManager      = null;}
  if (null != m_tStatisticsManager) {m_tStatisticsManager.dispose(); m_tStatisticsManager = null;}

  //
  // close main window and exit with success code
  //
  m_tMainWindow.setVisible(false);
  m_tMainWindow.dispose();
  System.exit(1);    
 }
  
  
 
 
/****************************************** MouseListener methods ***********************************************************************/
/**
 * Mouse was pressed -> start draing shapes or points
 * 
 * @param pa_tEvent    Mouse event
 */
 public void mousePressed(MouseEvent pa_tEvent) 
 {
 Point lc_tPoint = null;
 int   lc_iOldCellAmount = 0;
 
 if (null != pa_tEvent && null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
 {
  //
  // Handle mouse click event in cell paint panel
  //
  if (pa_tEvent.getSource() == m_tCellPaintPanel)
  {
   //
   // left click: place cell or shape
   //
   if (true == SwingUtilities.isLeftMouseButton(pa_tEvent))
   {
    if (null != (lc_tPoint = getMouseGridPosition(pa_tEvent)))
    {
     lc_iOldCellAmount = m_tCellMatrix.getCellAmount(); 
      
     //
     // handle placing single cells
     //
     if (true == m_tCellMatrixWindow.isPointSelectionActive())
     {
      m_tLastMouseCellPosition = lc_tPoint;
       
      m_tCellMatrix.flipCell(lc_tPoint.x, lc_tPoint.y, m_iCellSetAge, m_tCellMatrixWindow.isShapePaintXorModeActive());
      reactOnMousePressed(lc_iOldCellAmount);
     }
     else if (true == m_tCellMatrixWindow.isShapePaletteSelectionActive() && true == m_tCellPaintPanel.hasPaletteShape())
     {
      //
      // handle placing palette shapes
      //
      m_tLastMouseCellPosition = lc_tPoint;
       
      m_tCellMatrix.placeShape(lc_tPoint, m_tCellPaintPanel.getPaletteShape(), m_tCellMatrixWindow.isShapePaintXorModeActive(), false, m_iCellSetAge); 
      reactOnMousePressed(lc_iOldCellAmount);
     }
      else if (true == m_tCellMatrixWindow.isPaletteEditorSelectionActive() && true == m_tCellPaintPanel.hasPaletteEditorShape())
      {
       //
       // handle placing palette editor shapes
       //
       m_tLastMouseCellPosition = lc_tPoint;

       m_tCellMatrix.placeShape(lc_tPoint, m_tCellPaintPanel.getPaletteEditorShape(), m_tCellMatrixWindow.isShapePaintXorModeActive(), false, m_iCellSetAge); 
       reactOnMousePressed(lc_iOldCellAmount);
      }
      else if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() && true == m_tCellPaintPanel.hasSelectionShape())
      {
       //
       // handle placing selection shapes
       //
       m_tLastMouseCellPosition = lc_tPoint;
        
       m_tCellMatrix.placeShape(lc_tPoint, m_tCellPaintPanel.getSelectionShape(), m_tCellMatrixWindow.isShapePaintXorModeActive(), true, m_iCellSetAge); 
       reactOnMousePressed(lc_iOldCellAmount);
      }
      //
      // handle selection frame
      //
      else if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() &&  false == m_tCellPaintPanel.hasSelectionShape())
      {
       m_tCellPaintPanel.setCellSelectionStartPosition(lc_tPoint); 
       m_tCellPaintPanel.setCellSelectionEndPosition  (lc_tPoint); 
       m_tCellPaintPanel.repaint();
      }
     }
    }
    //
    // right click: delete cells in point selection mode
    //
    else if (true == SwingUtilities.isRightMouseButton(pa_tEvent))
    {
     if (null != (lc_tPoint = getMouseGridPosition(pa_tEvent)))
     {
      if (true == m_tCellMatrixWindow.isPointSelectionActive())
      {
       lc_iOldCellAmount = m_tCellMatrix.getCellAmount(); 
       m_tLastMouseCellPosition = lc_tPoint;    
       m_tCellMatrix.setCell(lc_tPoint.x, lc_tPoint.y, 0);
       reactOnMousePressed(lc_iOldCellAmount);
      }
     }
    }    
   }
  
  }
 }

 
 
/**
 * Mouse was released -> finish a cell selection (drawing that selection frame)
 * 
 * @param pa_tEvent    Mouse event
 */
 public void mouseReleased(MouseEvent pa_tEvent) 
 {
  Point lc_tPoint = null;

  if (null != pa_tEvent && null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   if (false == m_tCellPaintPanel.hasSelectionShape())
   {
    //
    // Handle mouse click event in cell paint panel
    //
    if (pa_tEvent.getSource() == m_tCellPaintPanel)
     {
     //
     // left click: check for end of drawing a cell matrix selection
     //
     if (true == SwingUtilities.isLeftMouseButton(pa_tEvent))
     {
      if (null != (lc_tPoint = getMouseGridPosition(pa_tEvent)))
       {
       if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive())
       {
        m_tCellPaintPanel.setCellSelectionEndPosition(lc_tPoint); 
        m_tCellPaintPanel.repaint();
        checkButtonStates();
        checkButtonStatesPaletteEditorWindow();
       }
      }
     }
    }
   }
  }
 }
 
 
 
 /**
  * Mouse has entered an area
  * 
  * @param pa_tEvent    Mouse event
  */
 public void mouseEntered(MouseEvent pa_tEvent) 
 {
  if (null != pa_tEvent && null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive())
   {
    if (pa_tEvent.getSource() == m_tCellPaintPanel)
    {
     if (false == m_tCellPaintPanel.hasSelectionShape()) // indicate selection shape by not changing to crosshair cursor
     {
      m_tCellPaintPanel.setCrossHairCursor();
      m_tCellPaintPanel.repaint();
     }
    }
   }
   handleMouseMovement(pa_tEvent, true);
  }
 }
 
 
  
 /**
  * Mouse has exited an area
  * 
  * @param pa_tEvent    Mouse event
  */
 public void mouseExited(MouseEvent pa_tEvent) 
 {
  if (null != pa_tEvent && null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive())
   {
    if (pa_tEvent.getSource() == m_tCellPaintPanel)
    {
     m_tCellPaintPanel.setDefaultCursor();      
     m_tCellPaintPanel.repaint();
    }
   }
   handleMouseMovement(pa_tEvent, false);
  }
 }
    

 
/**
 * Handles a mouse click -> rotating and mirroring shapes
 * 
 * @param pa_tEvent   Mouse event
 */
public void mouseClicked(MouseEvent pa_tEvent) 
{
 if (null != pa_tEvent)
 {
  if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   //
   // Handle mouse click event in cell paint panel
   //
   if (pa_tEvent.getSource() == m_tCellPaintPanel)
   {
    //
    // right click: rotate shape, calling the rotatePaletteShape() method invokes repaint()
    //
    if (true == SwingUtilities.isRightMouseButton(pa_tEvent))
    {
     //
     // palette shape
     //
     if (true == m_tCellMatrixWindow.isShapePaletteSelectionActive() && true == m_tCellPaintPanel.hasPaletteShape())
     {
      m_tCellPaintPanel.rotatePaletteShape();  
     }
     //
     // palette editor shape
     //
     else if (true == m_tCellMatrixWindow.isPaletteEditorSelectionActive() && true == m_tCellPaintPanel.hasPaletteEditorShape())
     {
      m_tCellPaintPanel.rotatePaletteEditorShape();  
     }
     //
     // cell seleciton shape
     //
     else if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() && true == m_tCellPaintPanel.hasSelectionShape())
     {
      m_tCellPaintPanel.rotateSelectionShape();  
     }
    }
    //
    // middle click: mirror shape
    //
    else if (true == SwingUtilities.isMiddleMouseButton(pa_tEvent))
    {
     //
     // palette shape
     //
     if (true == m_tCellMatrixWindow.isShapePaletteSelectionActive() && true ==m_tCellPaintPanel.hasPaletteShape())
     {
      m_tCellPaintPanel.mirrorPaletteShape();  
     }
     //
     // palette editor shape
     //
     else if (true == m_tCellMatrixWindow.isPaletteEditorSelectionActive() && true == m_tCellPaintPanel.hasPaletteEditorShape())
     {
      m_tCellPaintPanel.mirrorPaletteEditorShape();  
     }
     //
     // cell selection shape
     //
     else if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() && true == m_tCellPaintPanel.hasSelectionShape())
     {
      m_tCellPaintPanel.mirrorSelectionShape();  
     }
    }
   }
  }
 }
}
  
 

 
/****************************************** MouseWheelListener methods ***********************************************************************/
/**
 * Handles the mouse wheel  - affects the zoom factor
 * 
 * @param pa_tEvent   ChangeEvent
 */
 public void mouseWheelMoved(MouseWheelEvent pa_tEvent)
 {
  int lc_iNotches      = 0 ;
  int lc_iNewZoomValue = m_iZoomFactor;
   
  Point lc_tPoint = null;
  
  if (null != pa_tEvent && null != m_tCellMatrixWindow)
  {
   lc_iNotches = pa_tEvent.getWheelRotation();
   if (true == m_bInvertMouseWheel) lc_iNotches *= -1;
   lc_iNewZoomValue += lc_iNotches;
    
   if (0                                     > lc_iNewZoomValue) lc_iNewZoomValue = 0;
   if (m_tCellMatrixWindow.getMaxZoomValue() < lc_iNewZoomValue) lc_iNewZoomValue = m_tCellMatrixWindow.getMaxZoomValue();
   
   if (lc_iNewZoomValue != m_iZoomFactor)
   {
    m_iZoomFactor = lc_iNewZoomValue;
    m_tCellMatrixWindow.setZoomValue(m_iZoomFactor);
    
    if (null != m_tCellPaintPanel)
    {
     m_tCellPaintPanel.changeCellSize(m_iZoomFactor + GridLifeConstants.GOL_CELL_MIN_SIZE); 
     
     //
     // Updating frames, selections etc. 
     //
     if (null != (lc_tPoint = getMouseGridPosition(pa_tEvent)))
     {
      //
      // check for cell selection mode
      //
      if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() && m_tCellPaintPanel.hasSelectionShape())
      {
       m_tCellPaintPanel.setPaintSelectionShape(true); 
       
       if (true == m_tCellPaintPanel.checkSelectionShapePosition(lc_tPoint, true))
       {
        m_tCellPaintPanel.repaint(); 
       }
      }
      //
      // check for palette shape mode
      //
      else if (true == m_tCellMatrixWindow.isShapePaletteSelectionActive() && m_tShapePaletteWindow.hasPaletteSelection())
      {
       m_tCellPaintPanel.setPaintPaletteShape(true); 
       
       if (true == m_tCellPaintPanel.checkPaletteShapePosition(lc_tPoint, true))
       {
        m_tCellPaintPanel.repaint(); 
       }
      }    
      //
      // check for palette editor shape mode
      //
      else if (true == m_tCellMatrixWindow.isPaletteEditorSelectionActive() && m_tPaletteEditorWindow.hasListSelection())
      {
       m_tCellPaintPanel.setPaintPaletteEditorShape(true); 
       
       if (true == m_tCellPaintPanel.checkPaletteEditorShapePosition(lc_tPoint, true))
       {
        m_tCellPaintPanel.repaint(); 
       }
      }
      //
      // check for point mode
      //
      else if (true == m_tCellMatrixWindow.isPointSelectionActive())
      {
       m_tCellPaintPanel.setPaintFrame(true); 

       if (true == m_tCellPaintPanel.checkFramePosition(lc_tPoint, true))
       {
        m_tCellPaintPanel.repaint(); 
       }
      }     
     }
    }
   }  
  }
 }

 
  
/****************************************** MouseMotionListener methods ***********************************************************************/
/**
 * Handles mouse dragging -> painting shapes or points
 * 
 *   Mouse event
 */
public void mouseDragged(MouseEvent pa_tEvent) 
{
 Point lc_tPoint    = null;
 Point lc_tGfxPoint = null;
 int   lc_iOldCellAmount = 0;
 
 if (null != pa_tEvent)
 {
  if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel)
  {
   if (pa_tEvent.getSource() == m_tCellPaintPanel)
   {
   //
   // Handle left (!!) mouse click event in cell paint panel
   //
   if (true == SwingUtilities.isLeftMouseButton(pa_tEvent))
   {
    if (null != (lc_tPoint = getMouseGridPosition(pa_tEvent)))
    {
     //
     // set point
     //
     if (true == m_tCellMatrixWindow.isPointSelectionActive())
     {
      //
      // make sure we moved to another cell position!
      //
      if (lc_tPoint.x != m_tLastMouseCellPosition.x || lc_tPoint.y != m_tLastMouseCellPosition.y) 
      {
       m_tLastMouseCellPosition = lc_tPoint;      
       lc_iOldCellAmount       = m_tCellMatrix.getCellAmount();
      
       //
       // set single cell according to paintmode
       //
       m_tCellMatrix.flipCell(lc_tPoint.x, lc_tPoint.y, m_iCellSetAge, m_tCellMatrixWindow.isShapePaintXorModeActive());
       m_tCellPaintPanel.setFramePosition(m_tLastMouseCellPosition);
       reactOnMouseDragged(lc_iOldCellAmount);
      }
     }
     //
     // palette shape mode
     //
     else if (true == m_tCellMatrixWindow.isShapePaletteSelectionActive() && m_tCellPaintPanel.hasPaletteShape())
     {
      m_tLastMouseCellPosition = lc_tPoint;      
      lc_iOldCellAmount        = m_tCellMatrix.getCellAmount();
 
      m_tCellMatrix.placeShape(lc_tPoint, m_tCellPaintPanel.getPaletteShape(), m_tCellMatrixWindow.isShapePaintXorModeActive(), false, m_iCellSetAge); 
      m_tCellPaintPanel.setPaletteShapePosition(m_tLastMouseCellPosition);
      reactOnMouseDragged(lc_iOldCellAmount);
     }
     //
     // palette editor shape mode
     //
     else if (true == m_tCellMatrixWindow.isPaletteEditorSelectionActive() && m_tCellPaintPanel.hasPaletteEditorShape())
     {
      m_tLastMouseCellPosition = lc_tPoint;      
      lc_iOldCellAmount        = m_tCellMatrix.getCellAmount();
 
      m_tCellMatrix.placeShape(lc_tPoint, m_tCellPaintPanel.getPaletteEditorShape(), m_tCellMatrixWindow.isShapePaintXorModeActive(), false, m_iCellSetAge); 
      m_tCellPaintPanel.setPaletteEditorShapePosition(m_tLastMouseCellPosition);
      reactOnMouseDragged(lc_iOldCellAmount);
     }
     //
     // cell selection mode, but a shape
     //
     else if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() && true == m_tCellPaintPanel.hasSelectionShape())
     {
      m_tLastMouseCellPosition = lc_tPoint;      
      lc_iOldCellAmount        = m_tCellMatrix.getCellAmount();
 
      m_tCellMatrix.placeShape(lc_tPoint, m_tCellPaintPanel.getSelectionShape(), m_tCellMatrixWindow.isShapePaintXorModeActive(), true, m_iCellSetAge); 
      m_tCellPaintPanel.setSelectionShapePosition(m_tLastMouseCellPosition);
      reactOnMouseDragged(lc_iOldCellAmount);
     }
     //
     // cell selection mode, but just a frame. and don't draw a frame if we have a selection shape.
     //
     else if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() && false == m_tCellPaintPanel.hasSelectionShape())
     {
      //
      // make sure we moved to another cell position!
      //
      if (lc_tPoint.x != m_tCellPaintPanel.getCellSelectionEndPosition().x || lc_tPoint.y != m_tCellPaintPanel.getCellSelectionEndPosition().y) 
      {
       m_tCellPaintPanel.setCellSelectionEndPosition(lc_tPoint); 
       m_tCellPaintPanel.repaint();
       
       if (null != (lc_tGfxPoint = new Point(pa_tEvent.getX(), pa_tEvent.getY())))
       {
        //
        // if the point being dragged is out of sight, scroll to it
        //
        m_tCellPaintPanel.checkScrollToPosition(lc_tGfxPoint, false); 
       }
      }
     }      
    }
   }
   //
   // Handle rightmouse click event in cell paint panel -> delet cell, when in point selection mode
   //
   else if (true == SwingUtilities.isRightMouseButton(pa_tEvent))
   {
    if (null != (lc_tPoint = getMouseGridPosition(pa_tEvent)))
    {
     //
     // set point
     //
     if (true == m_tCellMatrixWindow.isPointSelectionActive())
     {
      //
      // make sure we moved to another cell position!
      //
      if (lc_tPoint.x != m_tLastMouseCellPosition.x || lc_tPoint.y != m_tLastMouseCellPosition.y) 
      {
       m_tLastMouseCellPosition = lc_tPoint;      
       lc_iOldCellAmount       = m_tCellMatrix.getCellAmount();
      
       //
       // set single cell according to paintmode
       //
       m_tCellMatrix.setCell(lc_tPoint.x, lc_tPoint.y, 0);
       m_tCellPaintPanel.setFramePosition(m_tLastMouseCellPosition);
       reactOnMouseDragged(lc_iOldCellAmount);
      }
     }
    }
   }
   }
  }
 }
}



/**
 * Handles mouse movement in cell drawing area.
 * 
 *   Mouse event
 */
public void mouseMoved(MouseEvent pa_tEvent) 
{
 handleMouseMovement(pa_tEvent, true);
}



 /**
  * Mouse has move around -> control shape painting
  * 
  * @param pa_tEvent    Mouse event
  */
 public void handleMouseMovement(MouseEvent pa_tEvent, boolean pa_tMouseEnterEvent) 
 {
  Point lc_tPoint = null; 
 
  if (null != m_tCellMatrixWindow && null != m_tCellPaintPanel && null != pa_tEvent)
  {
   //
   // handle movement in cell paint panel
   //
   if (pa_tEvent.getSource() == m_tCellPaintPanel) 
   {
    if (null != (lc_tPoint = getMouseGridPosition(pa_tEvent)))
    {
     //
     // check for palette shape mode
     //
     if (true == m_tCellMatrixWindow.isShapePaletteSelectionActive() && m_tShapePaletteWindow.hasPaletteSelection())
     {
      if (true == pa_tMouseEnterEvent) // mouseenter
      {
       m_tCellPaintPanel.setPaintPaletteShape(true); 
       
       if (true == m_tCellPaintPanel.checkPaletteShapePosition(lc_tPoint, pa_tMouseEnterEvent))
       {
        m_tCellPaintPanel.repaint(); 
       }
      }
      else // mouseexit
      {
       m_tCellPaintPanel.setPaintPaletteShape(false);   
       m_tCellPaintPanel.repaint(); 
      }
     }    
     //
     // check for palette editor shape mode
     //
     else if (true == m_tCellMatrixWindow.isPaletteEditorSelectionActive() && m_tPaletteEditorWindow.hasListSelection())
     {
      if (true == pa_tMouseEnterEvent) // mouseenter
      {
       m_tCellPaintPanel.setPaintPaletteEditorShape(true); 
       
       if (true == m_tCellPaintPanel.checkPaletteEditorShapePosition(lc_tPoint, pa_tMouseEnterEvent))
       {
        m_tCellPaintPanel.repaint(); 
       }
      }
      else // mouseexit
      {
       m_tCellPaintPanel.setPaintPaletteEditorShape(false);   
       m_tCellPaintPanel.repaint(); 
      }
     }    
     //
     // check for selection shape  mode
     //
     //     if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() && m_tCellPaintPanel.hasCellSelection())
     else if (true == m_tCellMatrixWindow.isCellMatrixSelectionActive() && m_tCellPaintPanel.hasSelectionShape())
     {
      if (true == pa_tMouseEnterEvent) // mouseenter
      {
       m_tCellPaintPanel.setPaintSelectionShape(true); 
       
       if (true == m_tCellPaintPanel.checkSelectionShapePosition(lc_tPoint, pa_tMouseEnterEvent))
       {
        m_tCellPaintPanel.repaint(); 
       }
      }
      else // mouseexit
      {
       m_tCellPaintPanel.setPaintSelectionShape(false);   
       m_tCellPaintPanel.repaint(); 
      }
     }  
     //
     // check for point selection mode
     //
     else if (true == m_tCellMatrixWindow.isPointSelectionActive())
     {
      if (true == pa_tMouseEnterEvent) // mouseenter
      {
       //
       // if we don't have a shape, we have a frame!
       //
       m_tCellPaintPanel.setPaintFrame(true); 

       if (true == m_tCellPaintPanel.checkFramePosition(lc_tPoint, pa_tMouseEnterEvent))
       {
        m_tCellPaintPanel.repaint(); 
       }
      }
      else // mouseexit
      {
       m_tCellPaintPanel.setPaintFrame(false); 
       m_tCellPaintPanel.repaint(); 
      }
     } 
    }  
   }
  }
 }
    

/****************************************** DocumentListener methods ***********************************************************************/
/**
 * Handles a modify-event in the slot textfield
  * 
 * @param pa_tEvent 
 */
public void changedUpdate(DocumentEvent pa_tEvent) 
{
 checkButtonStatesQuickSave(true);
}



/**
 * Handles an update-event in the slot textfield
  * 
 * @param pa_tEvent 
 */
public void insertUpdate(DocumentEvent pa_tEvent) 
{
 checkButtonStatesQuickSave(true);
}



/**
 * Handles a remove-event in the slot textfield
  * 
 * @param pa_tEvent 
 */
public void removeUpdate(DocumentEvent pa_tEvent) 
{
 checkButtonStatesQuickSave(true);
}



      
/****************************************** AdjustmentListener methods ***********************************************************************/
/**
 * Handles scrollbar events in the cell matrix window to update the radar panel
 * 
 * @param pa_tEvent   Adjustment event
 */
 public void adjustmentValueChanged(AdjustmentEvent pa_tEvent) 
 {
  if (null != pa_tEvent && null != m_tCellMatrixWindow)
  {
   if (pa_tEvent.getSource() == (JScrollBar) m_tCellMatrixWindow.getCellScrollPane().getVerticalScrollBar() ||
       pa_tEvent.getSource() == (JScrollBar) m_tCellMatrixWindow.getCellScrollPane().getHorizontalScrollBar() )
   {}
  }
 }

 
 


/****************************************** Helper methods ***********************************************************************************/


/**
  * After setting a point or placing a shape we need to update displays
  * 
  * @param pa_iOldCellAmount former cell amount
  */
 private void reactOnMousePressed(int pa_iOldCellAmount)
 {
  if (null != m_tCellMatrix)
  {
   //
   // set modified flag
   //
   setModifiedFlag();

   //
   // repaint in cell radar panel and cell matrix panel
   //
   if (null != m_tCellPaintPanel) m_tCellPaintPanel.repaint();     
      
   //
   // in case of changes to the amount of cells we need to update the statistics window and the buttons
   //
   if (pa_iOldCellAmount != m_tCellMatrix.getCellAmount())
   {
    //
    // update statistics window and histogram
    //
    if (null != m_tStatisticsWindow   ) m_tStatisticsWindow.updateStatistics(getStatisticsRecord()); 
    if (null != m_tHistogramPaintPanel) m_tHistogramPaintPanel.repaint();
   

    //
    //
    // update button states according to new data only when switching to/esacping from 0 cells
    //
    if ((0 == pa_iOldCellAmount && 0 != m_tCellMatrix.getCellAmount() ||
        (0 != pa_iOldCellAmount && 0 == m_tCellMatrix.getCellAmount() )))
    {
     checkButtonStates();
     checkButtonStatesQuickSave(false);
    }
   }
  }
 }
 
 

 /**
  * After setting a point or placing a shape we need to update displays
  * 
  * @param pa_iOldCellAmount former cell amount
  */
 private void reactOnMouseDragged(int pa_iOldCellAmount)
 {
  if (null != m_tCellMatrix)
  {
   //
   // set modified flag
   //
   setModifiedFlag();

   //
   // repaint in cell radar panel and cell matrix panel
   //
   if (null != m_tCellPaintPanel) m_tCellPaintPanel.repaint();     
      
   //
   // in case of changes to the amount of cells we need to update the statistics window and the buttons
   //
   if (pa_iOldCellAmount != m_tCellMatrix.getCellAmount())
   {
    if (null != m_tStatisticsWindow   ) m_tStatisticsWindow.updateStatistics(getStatisticsRecord()); 
    if (null != m_tHistogramPaintPanel) m_tHistogramPaintPanel.repaint();
  
       
    //
    // update button states according to new data only when switching to/esacping from 0 cells
    //
    if ((0 == pa_iOldCellAmount && 0 != m_tCellMatrix.getCellAmount() ||
        (0 != pa_iOldCellAmount && 0 == m_tCellMatrix.getCellAmount() )))
    {
     checkButtonStates();  
     checkButtonStatesQuickSave(false);
    }
   }
  }
 }

 
 
/**
 * Updates the main menu check boxes according to the internal frames' status
 * 
 */
 private void updateMainMenu()
 {
  if (null != m_tMainWindow)
  {
   m_tMainWindow.setWindowMenuCheckBoxes( (null != m_tPropertiesWindow    ? m_tPropertiesWindow.isVisible()    : false),
                                          (null != m_tStatisticsWindow    ? m_tStatisticsWindow.isVisible()    : false),
                                          (null != m_tShapePaletteWindow  ? m_tShapePaletteWindow.isVisible()  : false),
                                          (null != m_tColorWindow         ? m_tColorWindow.isVisible()         : false),
                                          (null != m_tPaletteEditorWindow ? m_tPaletteEditorWindow.isVisible() : false));
   
  }
 }
 


 /**
  * Checks, whether the mouse event took place inside the grid drawing panel and if so, returns cell matrix coordinates column/row
  * 
  * @param pa_tEvent   Mouse event
  * @return            null, if the mouse was not inside the grid or in case of error, Point otherwise with Point.x=column, Point.y=row
  */
 private Point getMouseGridPosition(MouseEvent pa_tEvent)
 {
  Point     lc_tResult    = null; 
  Component lc_tComponent = null;
  
  int   lc_iMouseX = 0;
  int   lc_iMouseY = 0;
  float lc_fColumn = 0;
  float lc_fRow    = 0;
  
  if (null != pa_tEvent && null != m_tCellPaintPanel)
  {
   if (null != (lc_tComponent = pa_tEvent.getComponent()))
   {
    //
    // check for mouse click inside cell grid
    //
    if (lc_tComponent instanceof JPanel)  
    {
     if (m_tCellPaintPanel == (JPanel) lc_tComponent)
     {
      lc_iMouseX = pa_tEvent.getX(); 
      lc_iMouseY = pa_tEvent.getY();   
     
      
      lc_fColumn = (lc_iMouseX / m_tCellPaintPanel.getCellSize());  //lc_fColumn = ((lc_iMouseX + 0.5f) / m_tCellPaintPanel.getCellSize());
      lc_fRow    = (lc_iMouseY / m_tCellPaintPanel.getCellSize());  //lc_fRow    = ((lc_iMouseY + 0.5f) / m_tCellPaintPanel.getCellSize());  
      lc_tResult = new Point((new Float(lc_fColumn)).intValue(), (new Float(lc_fRow)).intValue());
     }
    }
   }
  }
  return lc_tResult;
 }
 
 
 
/**
 * Enables or disables the quick save slot buttons according to the textfield contents
 * 
 */
private void checkButtonStatesQuickSave(boolean pa_bCheckLoad)
{
 String  lc_sText = "";
 String  lc_sFile = "";
 
 boolean lc_bSave     = false;
 boolean lc_bLoad     = false;
 boolean lc_bHasCells = false;
 
 if (null != m_tCellMatrixWindow && null != m_tCellMatrix)
 {
  lc_bHasCells = (m_tCellMatrix.getCellAmount() > 0);   
  
  if (null != (lc_sText = m_tCellMatrixWindow.getSlot())) 
  {
   if (false == lc_sText.isEmpty())
   {     
    lc_bSave = true;
    
    if (false == pa_bCheckLoad)
    {
     lc_bLoad = m_tCellMatrixWindow.getButtonSlotLoad().isEnabled();
    }
    else
    {
     lc_sFile = GridLifeConstants.GOL_FILE_QUICKSAVE_DIR + File.separator + lc_sText + GridLifeConstants.GOL_FILE_QUICKSAVE_EXTENSION;
     lc_bLoad = existsFile(lc_sFile);
    }
   }
  }
 }
 m_tCellMatrixWindow.getButtonSlotLoad().setEnabled(!m_bLifeInProgress & lc_bLoad); 
 m_tCellMatrixWindow.getButtonSlotSave().setEnabled(!m_bLifeInProgress & lc_bSave & lc_bHasCells); 
}



 /**
  * Enables or disables buttons according to data avaiable
  * 
  */
 private void checkButtonStates()
 {
  boolean lc_bHasColorSets           = false;
  boolean lc_bHasShapesets           = false;
  boolean lc_bHasCells               = false;
  boolean lc_bHasCellMatrixSelection = false;
  boolean lc_bHasCellMatrixShape     = false;
  boolean lc_bCellModeActive         = false;
 
  if (null != m_tCellMatrixWindow && null != m_tColorManager && null != m_tShapeSetManager && null != m_tCellMatrix && null != m_tCellPaintPanel)
  {

   lc_bHasColorSets           = (m_tColorManager.getColorSetAmount()    > 0);
   lc_bHasShapesets           = (m_tShapeSetManager.getShapeSetAmount() > 0);
   lc_bHasCells               = (m_tCellMatrix.getCellAmount()          > 0);
   lc_bHasCellMatrixSelection = (true == m_tCellPaintPanel.hasCellSelection());
   lc_bHasCellMatrixShape     = (true == m_tCellPaintPanel.hasSelectionShape());
   lc_bCellModeActive         = (true == m_tCellMatrixWindow.isCellMatrixSelectionActive());

   m_tCellMatrixWindow.getButtonClear().setEnabled     (!m_bLifeInProgress & lc_bHasCells);
   m_tCellMatrixWindow.getButtonSave().setEnabled      (!m_bLifeInProgress & lc_bHasCells);
   m_tCellMatrixWindow.getButtonPlay().setEnabled      (!m_bLifeInProgress & lc_bHasCells);
   m_tCellMatrixWindow.getButtonStep().setEnabled      (!m_bLifeInProgress & lc_bHasCells);
   m_tCellMatrixWindow.getButtonFill().setEnabled      (!m_bLifeInProgress               );
   m_tCellMatrixWindow.getButtonLoad().setEnabled      (!m_bLifeInProgress               );
   m_tCellMatrixWindow.getButtonStop().setEnabled      ( m_bLifeInProgress               );
   m_tCellMatrixWindow.getButtonScreenshot().setEnabled( lc_bHasCells                    );
   m_tCellMatrixWindow.getButtonSelectAll().setEnabled ( lc_bCellModeActive              );
   m_tCellMatrixWindow.getSpinnerPercent().setEnabled  (!m_bLifeInProgress               );
   
   m_tCellMatrixWindow.getButtonZoomInto().setEnabled           (lc_bHasCellMatrixSelection                     );
   m_tCellMatrixWindow.getButtonFillWithDeadCells().setEnabled  (lc_bHasCellMatrixSelection                     );
   m_tCellMatrixWindow.getButtonFillWithLivingCells().setEnabled(lc_bHasCellMatrixSelection                     );
   m_tCellMatrixWindow.getButtonInvert().setEnabled             (lc_bHasCellMatrixSelection                     );
   m_tCellMatrixWindow.getButtonPaste().setEnabled              (lc_bHasCellMatrixSelection & lc_bCellModeActive);
   m_tCellMatrixWindow.getButtonUnset().setEnabled              (lc_bHasCellMatrixShape     & lc_bCellModeActive);
   
   m_tCellMatrixWindow.getTextFieldSlot().setEnabled(!m_bLifeInProgress);
   m_tCellMatrixWindow.getButtonSlot().setEnabled   (!m_bLifeInProgress);
   
   if (null != m_tShapePaletteWindow) {m_tShapePaletteWindow.getComboboxPaletteSet().setEnabled(lc_bHasShapesets);}
   if (null != m_tColorWindow       ) {m_tColorWindow.getComboboxColorSet().setEnabled(lc_bHasColorSets);         }
   
   if (null != m_tMainWindow)
   {
    m_tMainWindow.getMenuItemLoad().setEnabled              (!m_bLifeInProgress                ); 
    m_tMainWindow.getMenuItemSave().setEnabled              (!m_bLifeInProgress & lc_bHasCells ); 
    m_tMainWindow.getMenuItemAbout().setEnabled             (!m_bLifeInProgress                ); 
    m_tMainWindow.getMenuItemClear().setEnabled             (!m_bLifeInProgress & lc_bHasCells ); 
    m_tMainWindow.getMenuItemCleanupScreenshots().setEnabled(!m_bLifeInProgress                ); 
    m_tMainWindow.getMenuItemCleanupQuickSaves().setEnabled (!m_bLifeInProgress                ); 
   }
  }
 }
 
 
 
 /**
  * checkd the buttons in the palette editor window
  * 
  */
 public void checkButtonStatesPaletteEditorWindow()
 {
  boolean lc_bHasPaletteSelection    = false;
  boolean lc_bHasListEntries         = false;
  boolean lc_bHasListSelection       = false;
  boolean lc_bHasCopyShape           = false;
  boolean lc_bHasSelectionShape      = false;
  
  if (null != m_tCellMatrixWindow && null!= m_tShapePaletteWindow && null != m_tPaletteEditorWindow && null != m_tCellPaintPanel)
  {
   lc_bHasPaletteSelection = (true == m_tShapePaletteWindow.hasPaletteSelection());        
   lc_bHasListEntries      = m_tPaletteEditorWindow.hasListEntries()  ;
   lc_bHasListSelection    = m_tPaletteEditorWindow.hasListSelection();
   lc_bHasCopyShape        = m_tCellPaintPanel.hasSelectionShape() || m_tCellPaintPanel.hasCellSelection();
   lc_bHasSelectionShape   = m_tCellPaintPanel.hasSelectionShape();
   
   m_tPaletteEditorWindow.getButtonClear().setEnabled        (lc_bHasListEntries  );
   m_tPaletteEditorWindow.getButtonDelete().setEnabled       (lc_bHasListSelection);           
   m_tPaletteEditorWindow.getButtonInsertMatrix().setEnabled ((lc_bHasCopyShape | lc_bHasSelectionShape)); 
   m_tPaletteEditorWindow.getButtonInsertPalette().setEnabled(lc_bHasPaletteSelection); 
    
   m_tPaletteEditorWindow.getButtonLoad().setEnabled(!m_bLifeInProgress);    
   m_tPaletteEditorWindow.getButtonSave().setEnabled(!m_bLifeInProgress && lc_bHasListEntries);       
  }
 }
 
 
 
 /**
  * Enables or disables buttons in the preferences window according to data available.
  * Makes sure that in case of enabling at least one selection is active in every ruleset row.
  * 
  */
 private void checkButtonStatesPropertiesWindow()
 {
  boolean lc_bHasChangedSettings    = false;
  boolean lc_bHasValidRules         = false;
  GridLifeProperties lc_tProperties = null ;
  
  if (null != m_tCellMatrix && null != m_tPropertiesWindow)
  {
   if (null != (lc_tProperties = m_tCellMatrix .getProperties()))
   {
    lc_bHasChangedSettings = (!(lc_tProperties.isEqual(m_tPropertiesWindow.getProperties())));
    lc_bHasValidRules      = m_tPropertiesWindow.getProperties().hasValidRules(); 
    
    m_tPropertiesWindow.getButtonApply().setEnabled  (!m_bLifeInProgress & lc_bHasChangedSettings & lc_bHasValidRules); 
    m_tPropertiesWindow.getButtonRestore().setEnabled(!m_bLifeInProgress & lc_bHasChangedSettings); 
    m_tPropertiesWindow.checkRuleButtons(!m_bLifeInProgress);
   }
  }   
 }
 
 
 
 /**
  * Enables or disables internal windows according to the given value
  * 
  * @param pa_bValue   true, if windows should be disabled, false if they should be enabled
  */
 private void lockWindows(boolean pa_bValue)
 {
  if (null != m_tPropertiesWindow) m_tPropertiesWindow.setWindowEnabled(!pa_bValue);
 }
 
 

 /**
  * Checks whether a file already exists
  * 
  * @param pa_sFileName    Filename
  * @return  true, if the file exists, false otherwise
  */
 private boolean existsFile(String pa_sFileName)
 {
  File lc_tFile = null ;
  
  if (null != (lc_tFile = new File(pa_sFileName)))
  {
   return (true == lc_tFile.exists() && true == lc_tFile.isFile()); 
  }
  return false;
 }
 
 
 
 /**
  * Adds the current cell values to the statistics manager
  * 
 */
 private void addCurrentValuesToStatistics()
 {
  GridLifeStatisticsRecord lc_tRecord = null;
  
  if (null != m_tStatisticsManager)
  {
   if (null != (lc_tRecord = getStatisticsRecord()))
   {
    m_tStatisticsManager.add(lc_tRecord.getTimestamp()     , 
                             lc_tRecord.getCellAmount()    , 
                             lc_tRecord.getGeneration()    , 
                             lc_tRecord.getTendency()      ,
                             lc_tRecord.getTendencyString());
   }
  }
 }


 
 /**
  * Returns a statistics record from the cell matrix with a localized tendency string
  * @return 
  */
 private GridLifeStatisticsRecord getStatisticsRecord()
 {
  GridLifeStatisticsRecord lc_tRecord      = null;
  String                   lc_sTendencyKey = null;
  
  if (null != m_tCellMatrix)
  {
   if (null != (lc_tRecord =  m_tCellMatrix.getStatisticsRecord()))
   {
    switch(lc_tRecord.getTendency())
    {
     case GridLifeConstants.GOL_STAT_TENTENDY_UNCHANGED: lc_sTendencyKey = GridLifeConstants.GOL_STAT_MSG_TENTENDY_UNCHANGED; break;
     case GridLifeConstants.GOL_STAT_TENTENDY_INCREASE : lc_sTendencyKey = GridLifeConstants.GOL_STAT_MSG_TENTENDY_INCREASE ; break;
     case GridLifeConstants.GOL_STAT_TENTENDY_DECREASE : lc_sTendencyKey = GridLifeConstants.GOL_STAT_MSG_TENTENDY_DECREASE ; break;
    }
    lc_tRecord.setTendencyString(getLocalizedString(lc_sTendencyKey));
   }
  }
  return lc_tRecord;   
 }
 
 
 
 /**
  * Stops the life animation
  * 
  */
 private boolean stopLifeAnimation()
 {
  boolean lc_bWasRunning = m_bLifeInProgress; 
   
  m_bLifeInProgress = false; 
  lockWindows(false);
  
  if (null != m_tAnimationThread)
  {
   m_tAnimationThread.stopAnimation();
   if (null != m_tPropertiesWindow) m_tPropertiesWindow.checkRuleButtons(true);
  }
  checkButtonStates(); 
  checkButtonStatesQuickSave(true);
  checkButtonStatesPropertiesWindow();     
  checkButtonStatesPaletteEditorWindow();
  
  if (null != m_tStatisticsWindow) m_tStatisticsWindow.setComputingThreads(null);
  return lc_bWasRunning;
 }
 
 
 
 /**
  * Starts the life animation
  * 
  */
 private void startLifeAnimation()
 {
  m_bLifeInProgress = true; 
  checkButtonStates();   
  checkButtonStatesQuickSave(false);
  checkButtonStatesPropertiesWindow();   
  checkButtonStatesPaletteEditorWindow();
  
  lockWindows(true);
  
  if (null != m_tCellMatrix && null != m_tCellMatrix.getProperties())
  {
   if (null != (m_tAnimationThread = new GridLifeAnimationThread((GridLifeGuiConnector) this, m_tCellMatrix.getProperties().getAnimationDelay())))
   {
    m_tAnimationThread.start(); 
    if (null != m_tPropertiesWindow) m_tPropertiesWindow.checkRuleButtons(false);
   }
  }
 }


 
 /**
  * Resets the modified flag to false
  * 
  */
 private void resetModifiedFlag()
 {
  if (true == m_bMatrixModified)
  {
   m_bMatrixModified = false; 
  }  
 }
 
 
 
 /**
  * Sets the modified flag and updates the matrix' window title
  * 
  */
 private void setModifiedFlag()
 {
  if (false == m_bMatrixModified)
  {
   m_bMatrixModified = true; 
   if (null != m_tCellMatrixWindow)
   {
    m_tCellMatrixWindow.setWindowTitle(m_tCellMatrixWindow.getWindowTitle() + " " + 
                                       getLocalizedString(GridLifeConstants.GOL_WINDOW_MODIFIED_TITLE   ));
   }
  }
 }
 
 
 
 /**
  * Saves currently valid properties to a file
  * 
  */
 private void saveProperties(String pa_sName)
 {
  if (null != m_tCellMatrix)
  {
   if (null != m_tCellMatrix.getProperties())
   {
    m_tCellMatrix.getProperties().save(pa_sName);
   }
  }
 }
 
 
 
 /**
  * Loads the  properties file
  * 
  * @return  Properties object
  */
 private GridLifeProperties loadProperties(String pa_sName)
 {
  GridLifeProperties lc_tProperties = new GridLifeProperties();
  
  if (null != lc_tProperties)
  {
   if (true == lc_tProperties.load(pa_sName)) 
   {
    return lc_tProperties; 
   }
  }
  return null;                           
 }
 
 

 /**
  * Handles clicking the button Fill area
  * 
  * @param pa_iMode  0 = fill with dead cells, 1 = fill with living cells, 2= invert
  *       
  */
 public void fillCellMatrixSelection(int pa_iMode)
 {
  int lc_iX1 = 0;
  int lc_iX2 = 0;
  int lc_iY1 = 0;
  int lc_iY2 = 0;
  
  int lc_iRow          = 0;
  int lc_iColumn       = 0;
  int lc_iOldCellValue = 0;
  int lc_iNewCellValue = 0;
  int lc_iNewCellAge   = 0;
  
  Point lc_tGridPosition1 = null;
  Point lc_tGridPosition2 = null;
  
  if (null != m_tCellMatrix && null != m_tCellPaintPanel && null != m_tStatisticsWindow && null != m_tHistogramPaintPanel)
  {
   lc_iNewCellAge    = (null != m_tColorWindow ? m_tColorWindow.getColorSelectionAge() : 1);
   lc_tGridPosition1 = m_tCellPaintPanel.getCellSelectionStartPosition();
   lc_tGridPosition2 = m_tCellPaintPanel.getCellSelectionEndPosition()  ;
  
   lc_iX1 = Math.min(lc_tGridPosition1.x, lc_tGridPosition2.x);
   lc_iX2 = Math.max(lc_tGridPosition1.x, lc_tGridPosition2.x);
   lc_iY1 = Math.min(lc_tGridPosition1.y, lc_tGridPosition2.y);
   lc_iY2 = Math.max(lc_tGridPosition1.y, lc_tGridPosition2.y);
  
   for (lc_iRow = lc_iY1; lc_iRow < lc_iY2; lc_iRow++)
   {
    for (lc_iColumn = lc_iX1; lc_iColumn < lc_iX2; lc_iColumn++)
    {
     lc_iOldCellValue = m_tCellMatrix.getCell(lc_iColumn, lc_iRow);
     
     switch(pa_iMode)
     {
      case 1 :
              lc_iNewCellValue = lc_iNewCellAge;
              break;   
      case 2 :
              lc_iNewCellValue = (0 == lc_iOldCellValue ? lc_iNewCellAge : 0);
              break;   
      case 0 :
      default:
              lc_iNewCellValue = 0;
     }
              
     m_tCellMatrix.setCell(lc_iColumn, lc_iRow, lc_iNewCellValue);
    } 
   }
   m_tCellPaintPanel.repaint();
   m_tStatisticsWindow.updateStatistics(getStatisticsRecord()); 
   m_tHistogramPaintPanel.repaint();
 
   checkButtonStates(); 
   checkButtonStatesQuickSave(false);
  }
 }
 

 
 /**
  * Returns the cell matrix selection area as int array
  * 
  * @return   null in case of error or empty selection, otherwise int array
  */
 private int[][] getArrayFromCellSelection()
 {
  int[][] lc_tSelectionArray    = null;
  Point   lc_tSelectionStartPos = null;
  Point   lc_tSelectionEndPos   = null;

  int lc_iWidth  = 0;
  int lc_iHeight = 0;
  
  int lc_iRow    = 0;
  int lc_iColumn = 0;
  
  int lc_iSelectionRow    = 0;
  int lc_iSelectionColumn = 0;
  
  int lc_iX1     = 0;
  int lc_iX2     = 0;
  int lc_iY1     = 0;
  int lc_iY2     = 0;
  
  int lc_iCount = 0;
  
  if (null != m_tCellPaintPanel && null != m_tCellMatrix)
  {
   lc_tSelectionStartPos = m_tCellPaintPanel.getCellSelectionStartPosition(); 
   lc_tSelectionEndPos   = m_tCellPaintPanel.getCellSelectionEndPosition(); 
   
   lc_iX1 = Math.min(lc_tSelectionStartPos.x, lc_tSelectionEndPos.x);
   lc_iX2 = Math.max(lc_tSelectionStartPos.x, lc_tSelectionEndPos.x);
   lc_iY1 = Math.min(lc_tSelectionStartPos.y, lc_tSelectionEndPos.y);
   lc_iY2 = Math.max(lc_tSelectionStartPos.y, lc_tSelectionEndPos.y);

   lc_iWidth  = lc_iX2 - lc_iX1;
   lc_iHeight = lc_iY2 - lc_iY1;
   
   if (null != (lc_tSelectionArray = new int[lc_iHeight][lc_iWidth]))
   {
    //
    // create int data array from selection
    //
    for (lc_iRow = lc_iY1; lc_iRow < lc_iY2; lc_iRow++)
    {
     lc_iSelectionColumn = 0;
     for (lc_iColumn = lc_iX1; lc_iColumn < lc_iX2; lc_iColumn++)
     {
      lc_tSelectionArray[lc_iSelectionRow][lc_iSelectionColumn] = m_tCellMatrix.getCell(lc_iColumn, lc_iRow);
      if (0 != lc_tSelectionArray[lc_iSelectionRow][lc_iSelectionColumn]) lc_iCount++;
      lc_iSelectionColumn++;
     }
     lc_iSelectionRow++;
    }
   }
  }
  return (null != lc_tSelectionArray ? (0 != lc_iCount ? lc_tSelectionArray : null) : null);
 }

 
 
 /**
  * Lets the user enter a shape name
  * 
  * @return   Shape name
  */
 private String requestShapeName(String pa_sInitialValue)
 {
  String lc_sInput = null;
  boolean lc_bDone = false;
  
  if (null != m_tPaletteEditorWindow)
  {
   while (false == lc_bDone)
   {
    lc_sInput = (String) JOptionPane.showInputDialog(m_tMainWindow, 
                                                     getLocalizedString(GridLifeConstants.GOL_DIALOG_ENTER_SHAPE_NAME), 
                                                     getLocalizedString(GridLifeConstants.GOL_TITLE_INPUT), 
                                                     JOptionPane.INFORMATION_MESSAGE, null, null, pa_sInitialValue); 
   
    if (null != lc_sInput)
    {
     if (false == lc_sInput.isEmpty())
     {
      if (false == m_tPaletteEditorWindow.hasEntry(lc_sInput)) 
      {
       lc_bDone = true;
      }
     }
    }
    else
    {
     lc_bDone = true; 
    }
   } // while
  }
  return lc_sInput;
 }

 
  
  
 /**
  * Computes the next gerenation using upto 128 threads.
  * 
  * completely rewritten to take care of non-squared matrices properly
  * 
  */
 public synchronized void computeNextGeneration()
 {
  GridLifeComputingThread[] lc_tThreads   = null;
  GridLifeComputingThread   lc_tNewThread = null;
  
  int  lc_iI          = 0;
  int  lc_iTendency   = 0;
  long lc_lAmount     = 0;
  long lc_lOldAmount  = 0;
  
  int lc_iThreadStartX  = 0;
  int lc_iThreadStartY  = 0;
  int lc_iThreadEndX    = 0;
  int lc_iThreadEndY    = 0;
  
  int  lc_iThreadCount       = 0;
  int  lc_iLessAmountThreads = 0;
  int  lc_iAmountThreads     = 1;
  long lc_lMaxCells          = 0;
  
  int lc_iRows    = 0;
  int lc_iColumns = 0;
  
  int lc_iLastSliceWidth  = 0;
  int lc_iSliceWidth      = 0;
  int lc_iSliceHeight     = 0;
  int lc_iLastSliceHeight = 0;
  
  if (null != m_tCellMatrix && null !=  m_tCellMatrix.getProperties())
  {
   //
   // store current values
   //
   addCurrentValuesToStatistics();  
   lc_lOldAmount  = m_tCellMatrix.getCellAmount();
   lc_iRows       = m_tCellMatrix.getProperties().getRows();
   lc_iColumns    = m_tCellMatrix.getProperties().getColumns();       
   lc_lMaxCells   = lc_iColumns  * lc_iRows;
   
   //
   // determine amount of threads to be used
   //
   if      (lc_lMaxCells <=    50000)  lc_iAmountThreads =   1; 
   else if (lc_lMaxCells <=   250000)  lc_iAmountThreads =   4; 
   else if (lc_lMaxCells <=  1000000)  lc_iAmountThreads =   8; 
   else if (lc_lMaxCells <=  3000000)  lc_iAmountThreads =  16; 
   else if (lc_lMaxCells <=  5000000)  lc_iAmountThreads =  32;
   else if (lc_lMaxCells <=  8000000)  lc_iAmountThreads =  64;
   else if (lc_lMaxCells <= 12000000)  lc_iAmountThreads =  92;
   else                                lc_iAmountThreads = 128;
   
   if (null != m_tStatisticsWindow) m_tStatisticsWindow.setComputingThreads(Integer.toString(lc_iAmountThreads));
   lc_iLessAmountThreads = lc_iAmountThreads - 1;
   
   //
   // create threads
   //
   if (null != (lc_tThreads = new GridLifeComputingThread[lc_iAmountThreads]))
   {
    lc_iSliceWidth  = lc_iColumns / lc_iAmountThreads; 
    lc_iSliceHeight = lc_iRows    / lc_iAmountThreads; 

    lc_iLastSliceWidth   = lc_iSliceWidth  + (lc_iColumns % lc_iAmountThreads);
    lc_iLastSliceHeight  = lc_iSliceHeight + (lc_iRows    % lc_iAmountThreads);
  
    if (lc_iColumns > lc_iRows)
    {
     //
     // create vertical slices
     //
     for (lc_iI = 0; lc_iI < lc_iAmountThreads; lc_iI++) 
     {
      lc_iThreadStartX = lc_iI * lc_iSliceWidth;
      lc_iThreadEndX   = lc_iThreadStartX + ((lc_iLessAmountThreads == lc_iI) ? lc_iLastSliceWidth : lc_iSliceWidth) -1;
      
      lc_iThreadStartY = 0;
      lc_iThreadEndY   = lc_iRows -1;
      
      if (null != (lc_tNewThread = new GridLifeComputingThread(m_tCellMatrix.getCells(), lc_iThreadStartX, lc_iThreadStartY, lc_iThreadEndX, lc_iThreadEndY, m_tCellMatrix.getProperties(), m_tColorManager)))
      {
       lc_tThreads[lc_iThreadCount++] = lc_tNewThread;
      }      
     } // for
    }
    else
    {
     //
     // create horizontal slices
     //
     for (lc_iI = 0; lc_iI < lc_iAmountThreads; lc_iI++) 
     {
      lc_iThreadStartY = lc_iI * lc_iSliceHeight;
      lc_iThreadEndY   = lc_iThreadStartY + ((lc_iLessAmountThreads == lc_iI) ? lc_iLastSliceHeight : lc_iSliceHeight) -1;
      
      lc_iThreadStartX = 0;
      lc_iThreadEndX   = lc_iColumns -1;

      if (null != (lc_tNewThread = new GridLifeComputingThread(m_tCellMatrix.getCells(), lc_iThreadStartX, lc_iThreadStartY, lc_iThreadEndX, lc_iThreadEndY, m_tCellMatrix.getProperties(), m_tColorManager)))
      {
       lc_tThreads[lc_iThreadCount++] = lc_tNewThread;
      }      
     } // for
    }

    //
    // start threads
    //
    for (lc_iI = 0; lc_iI < lc_iAmountThreads; lc_iI++)
    {
     lc_tThreads[lc_iI].start(); 
    }
    //
    // wait for threads
    //
    for (lc_iI = 0; lc_iI < lc_iAmountThreads; lc_iI++)
    {
     try
     {
      lc_tThreads[lc_iI].join(); 
     }
     catch (InterruptedException lc_tException)
     {}
    }

    //
    // get results, store new cell amount and tendency and generation number
    //
    for (lc_iI = 0; lc_iI < lc_iAmountThreads; lc_iI++)
    {
     lc_lAmount += lc_tThreads[lc_iI].writeBackResults();
    }

    //
    // get age histogram
    //
    if (null != m_tAgeHistogramManager)
    {
     m_tAgeHistogramManager.prepareImport();
     for (lc_iI = 0; lc_iI < lc_iAmountThreads; lc_iI++)
     {
      //
      // import data sorts the values according to the age value classes
      //
      m_tAgeHistogramManager.importData(lc_tThreads[lc_iI].getHistogramManager());
     }
    }
    //
    // set tendency string
    //
    if      (lc_lAmount == lc_lOldAmount) lc_iTendency = GridLifeConstants.GOL_STAT_TENTENDY_UNCHANGED;
    else if (lc_lAmount <  lc_lOldAmount) lc_iTendency = GridLifeConstants.GOL_STAT_TENTENDY_DECREASE ;
    else                                  lc_iTendency = GridLifeConstants.GOL_STAT_TENTENDY_INCREASE ;

    //
    // store new amount of cells and tendency, increase generation counter
    // and memorize all that stuff in a statistics record
    //
    m_tCellMatrix.setCellAmount((int)lc_lAmount);
    m_tCellMatrix.setTendency(lc_iTendency);
    m_tCellMatrix.increaseCellGeneration() ;
    addCurrentValuesToStatistics(); 

    //
    // update statistics window
    //
    if (null != m_tStatisticsWindow) m_tStatisticsWindow.updateStatistics(getStatisticsRecord());        
    

    //
    // repaint
    //
    if (null != m_tCellPaintPanel && null != m_tStatisticsPaintPanel && null != m_tHistogramPaintPanel)
    {
     SwingUtilities.invokeLater(new Runnable() 
                                {
                                 public void run() 
                                 {
                                  m_tCellPaintPanel.repaint();
                                  m_tStatisticsPaintPanel.repaint();
                                  m_tHistogramPaintPanel.repaint();
                                 }
                               });
    }   
    setModifiedFlag();

    //
    // auto-stop, if we have 0 cells left
    //
    if (0 == lc_lAmount) 
    {
     SwingUtilities.invokeLater(new Runnable() 
                                {
                                 public void run() 
                                 {
                                  stopLifeAnimation();
                                 } 
                                }); 
    }
   }
  }
 }
 
 
 
 /**
  * A screenshot thread just started
  * 
  */
 public void notifyIOThreadStarted(String pa_sData)
 {
  m_iIOThreadsRunning++; 
  if (null != m_tStatisticsWindow) m_tStatisticsWindow.setIOThreads(Integer.toString(m_iIOThreadsRunning)); 
 }
 

 
 /**
  * A screenshot thread just terminated
  * 
  */
 public void notifyIOThreadTerminated(String pa_sData)
 {
  m_iIOThreadsRunning--; 
  if (null != m_tStatisticsWindow) m_tStatisticsWindow.setIOThreads((0 == m_iIOThreadsRunning ? "" : Integer.toString(m_iIOThreadsRunning)));
 }


 
 /**
  * Modifies window settings according to given screen size and arranges subwindows and leaves gaps between them.
  * Makes a pretty default layout :)
  * 
  * @param  pa_bResizeMainWindow  true, if the main window has to be resized and moved, false otherwise
  * @param pa_bUpdateWindowsMenu  true, if the Windows meun should be updated, false otherwise
  * 
  * Layout: Color window         left         , 12% width, 100% height
  *         Cell matrix window   middle top   , 56.2% width, 80%  height
  *         Statistics window    middle bottom, 29% width, remainder height
  *         Properties window    top right    , remainder width, 30% height
  *         Shape palette window right left   , half remainder width, remainder height
  *         Shape palette window right right  , half remainder width, remainder height
  */ 
 private void createDefaultWindowSetup(boolean pa_bResizeMainWindow, boolean pa_bUpdateWindowsMenu)
 {
  Dimension lc_tScreenSize = Toolkit.getDefaultToolkit().getScreenSize();
  Rectangle lc_tInnerSize  = null;
  
  int lc_iGapX = 10;
  int lc_iGapY = 10;
  
  int lc_iWindowAreaWidth  = 0;
  int lc_iWindowAreaHeight = 0;
  
  int lc_iWindowX = 0;
  int lc_iWindowY = 0;
  int lc_iWindowW = 0;
  int lc_iWindowH = 0;
  
  int lc_iSafetyGapY = (int) (0.04 * lc_tScreenSize.height); // get area of main window which can be used for subwindows, subtract 5% safety space at every border for taskbars 
  
  int lc_iWidthRemain  = 0;
  int lc_iHeightRemain = 0;
  
  //
  // resize main window
  //
  if (true == pa_bResizeMainWindow)
  {
   m_tMainWindow.setBounds(0, 0, lc_tScreenSize.width, lc_tScreenSize.height - lc_iSafetyGapY);
  }
  lc_tInnerSize = m_tMainWindow.getContentPane().getBounds();
  
  lc_iWindowAreaWidth  = lc_tInnerSize.width  - (2 * lc_iGapX);
  lc_iWindowAreaHeight = lc_tInnerSize.height - (2 * lc_iGapY);
  lc_iHeightRemain     = lc_iWindowAreaHeight;
  lc_iWidthRemain      = lc_iWindowAreaWidth ;
  
  //
  // color palette window
  //
  lc_iWindowX      = lc_iGapX;
  lc_iWindowY      = lc_iGapY;
  lc_iWindowW      = (int) (lc_iWindowAreaWidth * 0.12);
  lc_iWindowH      = lc_iWindowAreaHeight              ;
  lc_iWidthRemain -= lc_iGapX + lc_iWindowW;

  m_tColorWindow.setBounds(new Rectangle(lc_iWindowX, lc_iWindowY, lc_iWindowW, lc_iWindowH));
  m_tColorWindow.setVisible(true);
  try {m_tColorWindow.setIcon(false);} catch (PropertyVetoException lc_tException) {}
  
  //
  // cell matrix window
  //
  lc_iWindowX      += lc_iWindowW + lc_iGapX; // use width of previously computed window for new coordinate 
  lc_iWindowW       = (int) (lc_iWindowAreaWidth  * 0.562);
  lc_iWindowH       = (int) (lc_iWindowAreaHeight * 0.82 );
  lc_iWidthRemain  -= lc_iGapX + lc_iWindowW;
  lc_iHeightRemain -= lc_iGapY + lc_iWindowH;
  m_tCellMatrixWindow.setBounds(new Rectangle(lc_iWindowX, lc_iWindowY, lc_iWindowW, lc_iWindowH));
  try {m_tCellMatrixWindow.setIcon(false);} catch (PropertyVetoException lc_tException) {}
  
  //
  // statistics window
  //
  lc_iWindowY      += lc_iWindowH + lc_iGapY;
  lc_iWindowH       = lc_iHeightRemain;
  m_tStatisticsWindow.setBounds(new Rectangle(lc_iWindowX, lc_iWindowY, lc_iWindowW, lc_iWindowH));
  m_tStatisticsWindow.setVisible(true);
  try {m_tStatisticsWindow.setIcon(false);} catch (PropertyVetoException lc_tException) {}
  
  //
  // properties window
  //
  lc_iHeightRemain  = lc_iWindowAreaHeight;
  lc_iWindowX      += lc_iGapX + lc_iWindowW; // use width of previously computed window for new coordinate 
  lc_iWindowY       = lc_iGapY;
  lc_iWindowW       = lc_iWidthRemain;
  lc_iWindowH       = (int) (lc_iWindowAreaHeight * 0.29);
  lc_iHeightRemain -= lc_iGapY + lc_iWindowH;
  m_tPropertiesWindow.setBounds(new Rectangle(lc_iWindowX, lc_iWindowY, lc_iWindowW, lc_iWindowH));
  m_tPropertiesWindow.setVisible(true);
  try {m_tPropertiesWindow.setIcon(false);} catch (PropertyVetoException lc_tException) {}

  //
  // palette window
  //
  lc_iWindowY      += lc_iWindowH + lc_iGapY;
  lc_iWindowW       = (int) ((lc_iWidthRemain - lc_iGapX) * 0.43);
  lc_iWindowH       = lc_iHeightRemain; 
  lc_iWidthRemain  -= lc_iWindowW;
  m_tShapePaletteWindow.setBounds(new Rectangle(lc_iWindowX, lc_iWindowY, lc_iWindowW, lc_iWindowH));
  m_tShapePaletteWindow.setVisible(true);
  try {m_tShapePaletteWindow.setIcon(false);} catch (PropertyVetoException lc_tException) {}
  
  //
  // palette editor window
  //
  lc_iWindowX      += lc_iGapX + lc_iWindowW;
  lc_iWindowW       = lc_iWidthRemain - lc_iGapX;
  m_tPaletteEditorWindow.setBounds(new Rectangle(lc_iWindowX, lc_iWindowY, lc_iWindowW, lc_iWindowH));
  m_tPaletteEditorWindow.setVisible(true);
  try {m_tPaletteEditorWindow.setIcon(false);} catch (PropertyVetoException lc_tException) {}
  
  //
  // set menu chechboxes according to window states, if requested
  //  
  if (true == pa_bUpdateWindowsMenu)
  {
   updateMainMenu();
  }
 }
 

 
 /**
  * Sets the new properties
  * 
  * @param pa_tProperties   new properties
  */
 private boolean setNewProperties(GridLifeProperties pa_tProperties)
 {
  boolean lc_bRealloc = false;
      
  if (null != pa_tProperties && null != m_tCellMatrix)
  {
   lc_bRealloc          = m_tCellMatrix.setProperties(pa_tProperties);
   m_iSceenshotCellSize = pa_tProperties.getScreenshotCellSize();
   m_bInvertMouseWheel  = (1 == pa_tProperties.getInvertMouseWheel()); 
 
   if (true == lc_bRealloc)
   {
    //
    // care for changed flag
    //
    if (null != m_tStatisticsManager) m_tStatisticsManager.clear();

    //
    // after reallocation of the cell matrix we need to propagate the new object's reference
    // to any painting panel and make them readjust their sizes
    //          
    if (null != m_tCellPaintPanel) 
    {
     m_tCellPaintPanel.setCells(m_tCellMatrix.getCells());
     m_tCellPaintPanel.setPanelSize();
    }
    
    //
    // make sure to display new properties (esp. values of columns/rows)
    // and clear any statistics
    //      
    if (null != m_tStatisticsPaintPanel) m_tStatisticsPaintPanel.repaint();
   }
  
   //
   // update animation thread
   //
   if (null != m_tAnimationThread) m_tAnimationThread.setAnimationDelay(pa_tProperties.getAnimationDelay()); 
   
     
   //
   // update properties
   //
   if (null != m_tPropertiesWindow    ) m_tPropertiesWindow.setProperties    (m_tCellMatrix.getProperties());  
   if (null != m_tStatisticsPaintPanel) m_tStatisticsPaintPanel.setProperties(m_tCellMatrix.getProperties());
   if (null != m_tHistogramPaintPanel ) m_tHistogramPaintPanel.setProperties (m_tCellMatrix.getProperties());
  
   if (null != m_tCellPaintPanel) 
   {
    m_tCellPaintPanel.setProperties(m_tCellMatrix.getProperties());     
    m_tCellPaintPanel.repaint();
   }
  
   //
   // update statistics window and histogram
   //
   if (null != m_tStatisticsWindow && null != m_tStatisticsPaintPanel && null != m_tHistogramPaintPanel)
   {
    m_tStatisticsWindow.updateStatistics(getStatisticsRecord()); 
    m_tStatisticsPaintPanel.repaint();
    m_tHistogramPaintPanel.repaint();
   }
      
   //
   // update button states according to new data
   //
   checkButtonStates();    
   checkButtonStatesPropertiesWindow();     
  }
  return lc_bRealloc;
 }

 
 
 /**
  * Installs the look and feel menu
  * 
  */
 private void installLFMenu()
 {
  m_tLFMenu = new GridLifeLookAndFeelMenu(getLocalizedString(GridLifeConstants.GOL_MENU_LOOKANDFEEL), m_tMainWindow); 
  
  if (null != m_tLFMenu)
  {
   if (true == m_tLFMenu.isLFMenuValid())
   {
    m_tMainWindow.getWindowMenuBar().add(m_tLFMenu); 
   }
  }
 }



 /**
  * Switches to the ui defined in the windows settings file
  * 
  */
 private void initLF()
 {
  if (null != m_tWindowSettings && null != m_tLFMenu) 
  {
   m_tLFMenu.selectLF(m_tWindowSettings.getLookAndFeelName());
  }
 }

 

 /**
  * Returns the localized string for the given key
  * 
  * @param pa_sKey
  * @return  localized string
  */
 private String getLocalizedString(String pa_sKey) 
 {
  return (null != m_tLocalizedMessages ? m_tLocalizedMessages.getString(pa_sKey) : ""); 
 }



 /**
  * Returns localized yes/no buttons texts
  * 
  */
 private String[] getLocalizedYesNoOptions()
 {
  String[] lc_tResult = new String[2];
  
  if (null != lc_tResult)
  {
   lc_tResult[0] = getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES); 
   lc_tResult[1] = getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_NO ); 
  }
  return lc_tResult;
 }
 
 

 
 /**
  * Returns localized yes/no buttons texts
  * 
  */
 private String[] getLocalizedOkCancelOptions()
 {
  String[] lc_tResult = new String[2];
  
  if (null != lc_tResult)
  {
   lc_tResult[0] = getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_OK    ); 
   lc_tResult[1] = getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_CANCEL); 
  }
  return lc_tResult;
 } 



 /**
  * Updates the ui and cell display according to the recently loaded file
  * 
  */
 private void handleLoadSuccess()
 {
  int lc_iX = 0;
  int lc_iY = 0;
  
  if (null != m_tCellMatrixWindow && null != m_tStatisticsManager && null != m_tCellMatrix   && null != m_tCellPaintPanel)
  {
   //
   // after reallocation of the cell matrix we need to propagate the new object's reference
   // to any painting panel and make them readjust their sizes
   //
   m_iZoomFactor = m_tCellMatrix.getLoadZoom(); 
   m_tCellMatrixWindow.setZoomValue(m_iZoomFactor);

   lc_iX = m_tCellMatrix.getLoadPosX();
   lc_iY = m_tCellMatrix.getLoadPosY();
   
   m_tCellPaintPanel.setCells(m_tCellMatrix.getCells());
   m_tCellPaintPanel.changeCellSize(GridLifeConstants.GOL_CELL_MIN_SIZE + m_iZoomFactor); 
   m_tCellPaintPanel.checkScrollToPosition(new Point(lc_iX, lc_iY), true);
   m_tCellPaintPanel.repaint();

   //
   // reset modification flag
   //
   resetModifiedFlag();

   //
   // make sure to display new properties (esp. values of columns/rows)
   // and clear any statistics
   //
   m_tStatisticsManager.clear();

   //
   // propagate new properties/property changes
   //
   if (null != m_tPropertiesWindow    ) m_tPropertiesWindow.setProperties(m_tCellMatrix.getProperties());
   if (null != m_tStatisticsPaintPanel) m_tStatisticsPaintPanel.setProperties(m_tCellMatrix.getProperties());
   if (null != m_tHistogramPaintPanel ) m_tHistogramPaintPanel.setProperties(m_tCellMatrix.getProperties());
   if (null != m_tStatisticsWindow    ) m_tStatisticsWindow.setMaxCells(m_tCellMatrix.getColumns() * m_tCellMatrix.getRows());

   //
   // update statistics and histogram window
   //
   if (null != m_tStatisticsWindow && null != m_tStatisticsPaintPanel)
   {
    m_tStatisticsWindow.updateStatistics(getStatisticsRecord()); 
    m_tStatisticsPaintPanel.repaint();
   }
   if (null != m_tHistogramPaintPanel) m_tHistogramPaintPanel.repaint(); 
   
   //
   // update button states according to new data
   //
   checkButtonStates();     
   checkButtonStatesQuickSave(false);
   checkButtonStatesPropertiesWindow();
  }
 }

 
 
 /**
  * Checks whether the main window fits inside the current screen
  * 
  * @return true, if it fits, false otherwise
  * 
  */
 private boolean checkWindowSettings()
 {
  Dimension lc_tScreenSize   = (null != Toolkit.getDefaultToolkit() ? Toolkit.getDefaultToolkit().getScreenSize() : null);
  Dimension lc_tCurrentSize  = (null != m_tMainWindow               ? m_tMainWindow.getSize()                     : null);
  Point     lc_tCurrentPos   = (null != m_tMainWindow               ? m_tMainWindow.getLocation()                 : null);
  
  int lc_iX = 0;
  int lc_iY = 0;
  int lc_iW = 0;
  int lc_iH = 0;
          
  if (null != m_tWindowSettings && null != lc_tScreenSize && null != lc_tCurrentSize && null != lc_tCurrentPos)  
  {
   lc_iX = lc_tCurrentPos.x;
   lc_iY = lc_tCurrentPos.y;
   lc_iW = lc_tCurrentSize.width;
   lc_iH = lc_tCurrentSize.height;
   
   if (lc_iX + lc_iW <= lc_tScreenSize.width        && 
       lc_iY + lc_iH <= lc_tScreenSize.height       &&
       lc_iX >= 0 && lc_iX <= lc_tScreenSize.width  &&    
       lc_iY >= 0 && lc_iY <= lc_tScreenSize.height )
   {
    return true;
   }
  }
  return false;
 }
 
 
 
 /**
  * Handles clicking the Slot button. Opens a modal dialog.
  * 
  */
 public void handleButton_CM_SlotBrowse()
 {
  String[] lc_tQuickSaveSlots      = null;
  String   lc_sQuickSaveName       = null;
  File[]   lc_tQuickSaveSlotsFiles = null;
  File     lc_tQuickSaveDir        = new File(GridLifeConstants.GOL_FILE_QUICKSAVE_DIR);

  GridLifeQuickSaveFileFilter lc_tFilter   = new GridLifeQuickSaveFileFilter();
  Rectangle lc_tInnerSize  = null;

  int lc_iAmount = 0;
  int lc_iI      = 0; 
  int lc_iPos    = 0;
  int lc_iX      = 0;
  int lc_iY      = 0;
  int lc_iW      = 0;
  int lc_iH      = 0;
  
  //
  // read files
  //
  if (null != lc_tQuickSaveDir)
  {
   if (true == lc_tQuickSaveDir.isDirectory() && true == lc_tQuickSaveDir.exists()) 
   {
    if (null != (lc_tQuickSaveSlotsFiles = lc_tQuickSaveDir.listFiles(lc_tFilter)))
    {
     if (0 < (lc_iAmount = lc_tQuickSaveSlotsFiles.length))
     {
      if (1 < lc_iAmount)
      {
       Arrays.sort(lc_tQuickSaveSlotsFiles);        
      }
      //
      // construct name array
      //
      if (null != (lc_tQuickSaveSlots = new String[lc_iAmount]))
      {
       for (lc_iI = 0; lc_iI < lc_iAmount; lc_iI++)
       {
        lc_sQuickSaveName = lc_tQuickSaveSlotsFiles[lc_iI].getName();
        lc_iPos           = lc_sQuickSaveName.indexOf(GridLifeConstants.GOL_FILE_QUICKSAVE_EXTENSION);
        
        if (-1 != lc_iPos)
        {
         lc_sQuickSaveName = lc_sQuickSaveName.substring(0, lc_iPos);
        }
        lc_tQuickSaveSlots[lc_iI] = lc_sQuickSaveName;
       }
   
       //
       // populate dialog list
       //
       if (null != m_tQuickSlotDialog)
       {
        m_tQuickSlotDialog.initList(lc_tQuickSaveSlots, m_tCellMatrixWindow.getSlot());  
       }
      }
     }
     
    }
   }
  }
  //
  // center dialog and bring it up...
  //
  if (null != m_tQuickSlotDialog)
  {
   if (0 < lc_iAmount)
   {
    lc_tInnerSize = m_tMainWindow.getBounds();
    lc_iW         = m_tQuickSlotDialog.getWidth() ;
    lc_iH         = m_tQuickSlotDialog.getHeight();
    lc_iX         = (lc_tInnerSize.width  - lc_iW) / 2;
    lc_iY         = (lc_tInnerSize.height - lc_iH) / 2;
    
    m_tQuickSlotDialog.setLocation(new Point(lc_iX, lc_iY)); 
    m_tQuickSlotDialog.setVisible(true);
   }
   else
   {
    JOptionPane.showMessageDialog(m_tMainWindow, 
                                  getLocalizedString(GridLifeConstants.GOL_DIALOG_QUICKSAVE_NO_FILES), 
                                  getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION ), 
                                  JOptionPane.INFORMATION_MESSAGE);
   }
  }
 }



 /**
  * Handles clicking the Select button in the quick slot dialog
  * 
  */
 public void handleButton_QS_Close()
 {
  String lc_sSlot = null;
  
  if (null != m_tQuickSlotDialog)
  {
   lc_sSlot = m_tQuickSlotDialog.getSelectedSlot();
   if (null != m_tCellMatrixWindow)
   {
    m_tCellMatrixWindow.setSlot(lc_sSlot);
    checkButtonStatesQuickSave(true);
   }
   m_tQuickSlotDialog.setVisible(false);
  }
 } 
 
 
 
 /**
  * Handles selecting the cleanup screenshots menu
  * 
  */
 public void handleMenuCleanupScreenshots()
 {
  if (0 == m_iIOThreadsRunning)
  { 
   cleanupFiles(GridLifeConstants.GOL_DIALOG_CLEANUP_SCREENSHOTS, GridLifeConstants.GOL_FILE_SCREENSHOT_DIR, new GridLifeScreenshotFileFilter());
  }
  else
  {
   JOptionPane.showMessageDialog(m_tMainWindow, 
                                 getLocalizedString(GridLifeConstants.GOL_DIALOG_CLEANUP_WAIT), 
                                 getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION ), 
                                 JOptionPane.INFORMATION_MESSAGE);     
  }
 }
 
 
 
 /**
  * Handles selecting the cleanup quicksaves menu
  * 
  */
 public void handleMenuCleanupQuickSaves()
 {
  cleanupFiles(GridLifeConstants.GOL_DIALOG_CLEANUP_QUICKSAVE, GridLifeConstants.GOL_FILE_QUICKSAVE_DIR, new GridLifeQuickSaveFileFilter());   
  checkButtonStatesQuickSave(true);
 }

 
 
 
 /**
  * Removes files matching the given filename filter from the given directory.
  * 
  */
 private void cleanupFiles(String pa_sDialogText, String pa_sDirectory, FilenameFilter pa_tFileNameFilter)
 {
  int     lc_iAmount  = 0;
  int     lc_iErased  = 0;
  int     lc_iI       = 0;
  boolean lc_bNoFiles = true;
  
  File   lc_tDir    = new File(pa_sDirectory);
  File[] lc_tFiles  = null;
  
  if (0 == JOptionPane.showOptionDialog(m_tMainWindow,  
                                        getLocalizedString(pa_sDialogText), 
                                        getLocalizedString(GridLifeConstants.GOL_TITLE_CONFIRMATION) , 
                                        JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, 
                                        getLocalizedYesNoOptions()                                   , 
                                        getLocalizedString(GridLifeConstants.GOL_DIALOG_OPTIONS_YES  ))) 
  {
   if (null != lc_tDir)
   {
    if (true == lc_tDir.isDirectory() && true == lc_tDir.exists()) 
    {
     if (null != (lc_tFiles = lc_tDir.listFiles(pa_tFileNameFilter)))
     {
      //
      // do we have files to be deleted?
      //
      if (0 < (lc_iAmount = lc_tFiles.length))
      {
       lc_bNoFiles = false;
       
       for (lc_iI = 0; lc_iI < lc_iAmount; lc_iI++)
       {
        try
        {
         if (true == lc_tFiles[lc_iI].delete())
         {
          lc_iErased++; 
         }
        }
        catch (Exception lc_tException)
        {}
       }
       
       if (lc_iAmount == lc_iErased)
       {
        JOptionPane.showMessageDialog(m_tMainWindow, 
                                      getLocalizedString(GridLifeConstants.GOL_DIALOG_CLEANUP_SUCCESS), 
                                      getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION ), 
                                      JOptionPane.INFORMATION_MESSAGE);     
       }
       else
       {
        JOptionPane.showMessageDialog(m_tMainWindow, 
                                      getLocalizedString(GridLifeConstants.GOL_DIALOG_CLEANUP_FAIL), 
                                      getLocalizedString(GridLifeConstants.GOL_TITLE_WARNING ), 
                                      JOptionPane.WARNING_MESSAGE);              
       }
      }
     }
    }
   }
   if (true == lc_bNoFiles)
   {
    JOptionPane.showMessageDialog(m_tMainWindow, 
                                  getLocalizedString(GridLifeConstants.GOL_DIALOG_CLEANUP_EMPTY), 
                                  getLocalizedString(GridLifeConstants.GOL_TITLE_INFORMATION ), 
                                  JOptionPane.INFORMATION_MESSAGE);             
   }
  }
 }
 

 
 /**
  * OutOfMemory. 
  * 
  */
 private void showMemoryAlert()
 {
  JOptionPane.showMessageDialog(m_tMainWindow, 
                                getLocalizedString(GridLifeConstants.GOL_DIALOG_OOM), 
                                getLocalizedString(GridLifeConstants.GOL_TITLE_WARNING ), 
                                JOptionPane.WARNING_MESSAGE);                 
 }

  
} // eoc
