/**
 * Game of Life
 * 
 */
package gridlife.core;



/**
 * Imports
 * 
 */
import gridlife.shape.GridLifeShape;
import gridlife.color.GridLifeColorManager;

import java.util.Random         ;
import java.util.StringTokenizer;
import java.awt.Point           ;

import java.io.File          ;
import java.io.FileReader    ;
import java.io.FileWriter    ;
import java.io.BufferedReader;
import java.io.BufferedWriter;



/**
 * This class stores information about a cell matrix
 * 
 * @author Ulf Wagemann
 */
public class GridLifeCellMatrix 
{
 private int[][] m_tCellArray;
 
 private GridLifeProperties           m_tProperties;
 private GridLifeColorManager         m_tColorManager;
 private GridLifeAgeHistogramManager  m_tHistogramManager;
 
 private long m_lGeneration;
 private int  m_iCellAmount;
 private int  m_iTendency  ;
 
 private int m_iLoadZoom;
 private int m_iLoadPosX;
 private int m_iLoadPosY;
 
 
 /**
  * Constructor
  * 
  */
 public GridLifeCellMatrix(GridLifeProperties pa_tMatrixProperties, GridLifeColorManager pa_tColorManager, GridLifeAgeHistogramManager pa_tHistogramManager)
 {
  m_lGeneration = 0;
  m_iCellAmount = 0;
  m_iTendency   = 0;
  
  m_tColorManager     = pa_tColorManager    ;
  m_tHistogramManager = pa_tHistogramManager;
  
  m_iLoadZoom = 0;
  m_iLoadPosX = 0;
  m_iLoadPosY = 0;
  
  if (null != pa_tMatrixProperties) 
  {
   m_tProperties = new GridLifeProperties(pa_tMatrixProperties);
   m_tCellArray = new int[m_tProperties.getRows()][m_tProperties.getColumns()];
  }
 }
 
 
 
 /**
  * "Copy"-constructor, creates an own cell array from the one contained in the fiven matrix
  * 
  * @param pa_tMatrix   cell matrix object top copy
  */
 public GridLifeCellMatrix(GridLifeCellMatrix pa_tMatrix)
 {
  if (null != pa_tMatrix && this != pa_tMatrix)  
  {
   m_lGeneration = 0;
   m_iCellAmount = 0;
   m_iTendency   = 0;
  
   if (null != pa_tMatrix.getProperties()) 
   {
    m_tProperties = new GridLifeProperties(pa_tMatrix.getProperties());
    m_tCellArray = new int[m_tProperties.getRows()][m_tProperties.getColumns()];
    
    copyCells(pa_tMatrix.getCells(), m_tCellArray, m_tProperties.getColumns(), m_tProperties.getRows());    
   }
  }
 }
 
 
 
 /**
  * Returns the cell array
  * 
  * @return   cell arry, null in case of error
  */
 public final int[][] getCells() {return m_tCellArray;}
 
 
 
 /**
  * Returns the amount of columns
  * 
  * @return   Amount of columns
  */
 public final int getColumns() {return (null != m_tProperties ? m_tProperties.getColumns() : 0);}
 

 
 /**
  * Returns the amount of rows
  * 
  * @return   Amount of rows
  */
 public final int getRows() {return (null != m_tProperties ? m_tProperties.getRows() : 0);}
 
 
 
 /**
  * Returns the cell at the given location
  * 
  * @parma pa_iColumn   column
  * @param pa_iRow      row
  * @return   -1 in case of error, otherwise cell age (0 = dead .. )
  */
 public final int getCell(int pa_iColumn, int pa_iRow) 
 {
  if (0 <= pa_iColumn && m_tProperties.getColumns() > pa_iColumn)
  {
   if (0 <= pa_iRow && m_tProperties.getRows() > pa_iRow)
   {
    return m_tCellArray[pa_iRow][pa_iColumn];
   }
  }
  return -1;
 }
 
 
 /**
  * Sets the cell amount to the given value
  * 
  * @param  pa_iValue new value
  * */
 public final void setCellAmount(int pa_iValue) 
 {
  if (0 <= pa_iValue && pa_iValue <= (m_tProperties.getColumns() * m_tProperties.getRows()))
  {
   m_iCellAmount = pa_iValue;
  }
 }
 
 
 
 /**
  * Returns the amount of cells
  * 
  * @return   amount of cells
  */
 public final int getCellAmount() {return m_iCellAmount;}
 
 
 
 /**
  * Returns the amount of cells
  * 
  * @return   amount of cells
  */
 public final long getCellGeneration() {return m_lGeneration;}
 
 
 
 /**
  * Returns the cell generation after increasing the value by 1
  * 
  * @return   Generation
  */
 public final long increaseCellGeneration() {return ++m_lGeneration;}
 
 
 /**
  * Returns the amount of cells
  * 
  * @return   amount of cells
  */
 public final int getTendency() {return m_iTendency;}

 
 /**
  * Sets the tendeny to the given value
  * 
  * @param pa_iTendency 
  */
 public final void setTendency(int pa_iTendency)
 {
  switch(pa_iTendency)
  {
   case GridLifeConstants.GOL_STAT_TENTENDY_DECREASE: 
   case GridLifeConstants.GOL_STAT_TENTENDY_INCREASE:  
   case GridLifeConstants.GOL_STAT_TENTENDY_UNCHANGED: 
     m_iTendency = pa_iTendency; 
     break;  
   default:
     m_iTendency = GridLifeConstants.GOL_STAT_TENTENDY_UNCHANGED; 
  }   
 }
 
 
 /**
  * Fills the matrix randomly with cells according to the given percentual fill ratio. Creates a new
  * statistics record after clearing the statistics.
  * 
  * @param pa_iRatio   Fill ratio
  * @param pa_iAge     Age of a new cell
  */
 public final void randomFillCellMatrix(int pa_iRatio, int pa_iAge)
 {
  float lc_fMaxCells  = 0;  
  float lc_fFillCells = 0;
  
  int lc_iX        = 0;
  int lc_iY        = 0;
  int lc_iCount    = 0;
  int lc_iNewValue = (0 < pa_iAge && Integer.MAX_VALUE > pa_iAge ? pa_iAge : 1);
          
  Random  lc_tRandom = new Random(System.nanoTime());
  Integer lc_tCellAgeColorClass = null;
  
  if (null != m_tProperties && null != m_tCellArray)
  {   
   if (1 <= pa_iRatio && 100 >= pa_iRatio)
   {
    resetCellMatrix();
    
    lc_fMaxCells  = m_tProperties.getColumns() * m_tProperties.getRows(); 
    lc_fFillCells = (lc_fMaxCells * pa_iRatio) / 100;
    
    m_lGeneration = 1;
    m_iCellAmount = 0;
    
    if (100 == pa_iRatio)
    {
     for (lc_iY = 0; lc_iY < getRows(); lc_iY++)
     {
      for (lc_iX = 0; lc_iX < getColumns(); lc_iX++)
      {
       m_tCellArray[lc_iY][lc_iX] = lc_iNewValue;
       m_iCellAmount++;
       
       if (null != m_tHistogramManager && null != m_tColorManager) 
       {
        if (null != (lc_tCellAgeColorClass = m_tColorManager.getAgeClass(m_tColorManager.getActiveSetName(), lc_iNewValue, null)))
        {
         m_tHistogramManager.add(lc_tCellAgeColorClass); 
        }
       }

      }    
     }
    }
    else
    {
     while (lc_iCount < lc_fFillCells)
     {
      lc_iX   = lc_tRandom.nextInt(m_tProperties.getColumns()); 
      lc_iY   = lc_tRandom.nextInt(m_tProperties.getRows()   );
      
      if (m_tCellArray[lc_iY][lc_iX] == 0)
      {
       m_tCellArray[lc_iY][lc_iX] = lc_iNewValue;
       m_iCellAmount++;
       lc_iCount++;
       
       if (null != m_tHistogramManager && null != m_tColorManager) 
       {
        if (null != (lc_tCellAgeColorClass = m_tColorManager.getAgeClass(m_tColorManager.getActiveSetName(), lc_iNewValue, null)))
        {
         m_tHistogramManager.add(lc_tCellAgeColorClass); 
        }
       }
       
      }
     } // while
    }
   }
  }
 }
 
 
 
 /**
  * Resets the complete cell matrix, turning any cell into a dead cell
  * 
  */
 public final void resetCellMatrix()
 {
  int lc_iI   = 0;
  int lc_iJ   = 0;
  
  if (null != m_tCellArray && null != m_tProperties)
  {
   for (lc_iI = 0; lc_iI  < m_tProperties.getRows(); lc_iI++)
   {
    for (lc_iJ = 0; lc_iJ < m_tProperties.getColumns(); lc_iJ++)
    {
     m_tCellArray[lc_iI][lc_iJ] = 0;
    }
   } 
  }
  m_lGeneration = 0;
  m_iCellAmount = 0;
  m_iTendency   = 0;
  
  if (null != m_tHistogramManager) 
  {
   m_tHistogramManager.reset();
  }
  
 }
  
 
 
 /**
  * Returns the most current statistics record
  * 
  * @return   a mock record with current values is returned.
  */
 public final GridLifeStatisticsRecord getStatisticsRecord()
 {
  return (new GridLifeStatisticsRecord(System.nanoTime(), m_iCellAmount, m_lGeneration, m_iTendency, null));
 }
 
 
 
 /**
  * Flips the grid content at the give coordinates
  * 
  * @param pa_iColumn   column
  * @param pa_iRow      row
  * @param pa_iAge      age [0..[
  */
 public final void flipCell(int pa_iColumn, int pa_iRow, int pa_iAge, boolean pa_bXorMode)
 {
  int lc_iOldValue  = 0;
  Integer lc_tCellAgeColorClass = null;
  
  if (null != m_tCellArray && null != m_tProperties)
  {
   if (0 <= pa_iColumn && m_tProperties.getColumns() > pa_iColumn)
   {
    if (0 <= pa_iRow && m_tProperties.getRows() > pa_iRow)
    {
     lc_iOldValue = m_tCellArray[pa_iRow][pa_iColumn];
     
     if (true == pa_bXorMode)
     {
      if (0 < lc_iOldValue)
      {
       m_tCellArray[pa_iRow][pa_iColumn] = 0; 
       m_iCellAmount--;

       //
       // update histogram manager
       //
       if (null != m_tHistogramManager && null != m_tColorManager) 
       {
        if (null != (lc_tCellAgeColorClass = m_tColorManager.getAgeClass(m_tColorManager.getActiveSetName(), lc_iOldValue, null)))
        {
         m_tHistogramManager.decrease(lc_tCellAgeColorClass); 
        }
       }
       
      }
      else
      {
       m_tCellArray[pa_iRow][pa_iColumn] = (0 < pa_iAge && Integer.MAX_VALUE > pa_iAge ? pa_iAge : 1);
       m_iCellAmount++;
       
       //
       // update histogram manager
       //
       if (null != m_tHistogramManager && null != m_tColorManager) 
       {
        if (null != (lc_tCellAgeColorClass = m_tColorManager.getAgeClass(m_tColorManager.getActiveSetName(), m_tCellArray[pa_iRow][pa_iColumn], null)))
        {
         m_tHistogramManager.add(lc_tCellAgeColorClass); 
        }
       }       
      }
     }
     else
     {
      m_tCellArray[pa_iRow][pa_iColumn] = (0 < pa_iAge  && Integer.MAX_VALUE > pa_iAge ? pa_iAge : 1);  
      if (0 == lc_iOldValue)
      {
       m_iCellAmount++;
       
       //
       // update histogram manager
       //
       if (null != m_tHistogramManager && null != m_tColorManager) 
       {
        if (null != (lc_tCellAgeColorClass = m_tColorManager.getAgeClass(m_tColorManager.getActiveSetName(), m_tCellArray[pa_iRow][pa_iColumn], null)))
        {
         m_tHistogramManager.add(lc_tCellAgeColorClass); 
        }
       }       
      }
     }
    }
   }
  }
  if (0 == m_iCellAmount)
  {
   m_lGeneration = 1;
  }
 }

 
 
 /**
  * Sets the cell at the given position to the given age
  * 
  * @param pa_iColumn  Column
  * @param pa_iRow     Row
  * @param pa_iValue   Age [0..[
  */
 public final void setCell(int pa_iColumn, int pa_iRow, int pa_iValue)
 {
  int lc_iOldValue = 0;
  Integer lc_tCellAgeColorClass= null;
 
  if (null != m_tCellArray && null != m_tProperties)
  {
   if (0 <= pa_iColumn && m_tProperties.getColumns() > pa_iColumn)
   {
    if (0 <= pa_iRow && m_tProperties.getRows() > pa_iRow)
    {
     lc_iOldValue = m_tCellArray[pa_iRow][pa_iColumn];
     m_tCellArray[pa_iRow][pa_iColumn] = ( 0 <= pa_iValue ? pa_iValue : 1);
     
     //
     // check if we created a new cell
     //
     if (0 == lc_iOldValue && 0 < m_tCellArray[pa_iRow][pa_iColumn])
     {
      m_iCellAmount++;

      if (null != m_tHistogramManager && null != m_tColorManager) 
      {
       if (null != (lc_tCellAgeColorClass = m_tColorManager.getAgeClass(m_tColorManager.getActiveSetName(), m_tCellArray[pa_iRow][pa_iColumn], null)))
       {
        m_tHistogramManager.add(lc_tCellAgeColorClass); 
       }
      }
     }
 
     //
     // check if we deleted a cell
     //
     if (0 < lc_iOldValue && 0 == m_tCellArray[pa_iRow][pa_iColumn])
     {
      if (null != (lc_tCellAgeColorClass = m_tColorManager.getAgeClass(m_tColorManager.getActiveSetName(), lc_iOldValue, null)))
      {
       m_tHistogramManager.decrease(lc_tCellAgeColorClass); 
      }
      m_iCellAmount--; 
     }
     
     //
     // check for correct population number
     //
     if (0 == m_iCellAmount)
     {
      m_lGeneration = 1; 
     }
    }
   }
  }
 }
 
 
 
 /**
  * Places a shape at the given grid position
  * 
  * @param pa_tPoint   Position   position to place the shape at
  * @param pa_tShape   Shape      shape to place
  * @param pa_bXorMode true if cell placement should be xor style, false otherwise
  * @param pa_iAge     Age of new cells
  */
 public final void placeShape(Point pa_tPoint, GridLifeShape pa_tShape, boolean pa_bXorMode, boolean pa_bUseCellAge, int pa_iAge)
 {
  int lc_iColumns = 0;
  int lc_iRows    = 0;
  int lc_iI       = 0;
  int lc_iJ       = 0;
  int lc_iX       = 0;
  int lc_iY       = 0;
  int lc_iAge     = 0;
  
  int[][] lc_tShapeData = null;
  
  if (null != pa_tPoint && null != pa_tShape)
  {
   if (null != m_tCellArray && null != m_tProperties)
   {
    lc_iColumns = pa_tShape.getShapeColumns();
    lc_iRows    = pa_tShape.getShapeRows()   ;
    
    if (0 < lc_iColumns && 0 < lc_iRows)
    {
     if (null != (lc_tShapeData = pa_tShape.getShapeDataIntArray()))
     {
      for (lc_iI = 0; lc_iI  < lc_iRows; lc_iI++)
      {
       lc_iY = pa_tPoint.y + lc_iI;
       
       if (0 <= lc_iY && m_tProperties.getRows() > lc_iY)
       {
        for (lc_iJ = 0; lc_iJ < lc_iColumns; lc_iJ++)
        {
         // 
         // only continue, if the shape has a block at the given position
         //
         if (0 != lc_tShapeData[lc_iI][lc_iJ]) 
         {
          lc_iAge = (true == pa_bUseCellAge ? lc_tShapeData[lc_iI][lc_iJ] : pa_iAge) ;
          lc_iX   = pa_tPoint.x + lc_iJ;
          
          if (0 <= lc_iX && m_tProperties.getColumns() > lc_iX)
          {
           //
           // normal set mode
           //
           if (false == pa_bXorMode)
           {
            setCell(lc_iX, lc_iY, lc_iAge);  
           }
           //
           // XOR mode: An existing cell is erased, an empty cell is called to life with the given age
           //
           else
           {
            setCell(lc_iX, lc_iY, (0 < m_tCellArray[lc_iY][lc_iX] ? 0 : lc_iAge));
           }
           // done 
          }
         }
        } // for lc_iJ
       }
      } // for lc_iI
     }
    }
   }  
  }
 }
 
 
 /**
  * Sets a new ruleset, reallocation the array if needed
  * 
  * @param pa_tProperties
  * @return true if a resizing happened, otherwise false
  */
 public final boolean setProperties(GridLifeProperties pa_tProperties)
 {
  boolean lc_bRealloc = false;
  
  int lc_iOldColumns = 0;
  int lc_iOldRows    = 0;
  
  int[][] lc_tDummyArray = null;
  
  if (null != pa_tProperties && null != m_tProperties) 
  {
   lc_iOldColumns = m_tProperties.getColumns();
   lc_iOldRows    = m_tProperties.getRows()   ;
      
   lc_bRealloc = (lc_iOldColumns !=  pa_tProperties.getColumns() || lc_iOldRows !=  pa_tProperties.getRows());
   m_tProperties.copyRules(pa_tProperties);
   
   //
   // before reallocating, we store the current array intermediately
   //
   if (true == lc_bRealloc)
   {
    if (null != (lc_tDummyArray = new int[lc_iOldRows][lc_iOldColumns]))
    {
     copyCells(m_tCellArray, lc_tDummyArray, lc_iOldColumns, lc_iOldRows);
    }
      
    m_tCellArray = new int[m_tProperties.getRows()][m_tProperties.getColumns()];

    //
    // after reallocating, copy back as much as we can
    //
    if (null != lc_tDummyArray)
    {   
     if (lc_iOldColumns > m_tProperties.getColumns()) lc_iOldColumns = m_tProperties.getColumns();
     if (lc_iOldRows    > m_tProperties.getRows()   ) lc_iOldRows    = m_tProperties.getRows()   ;

     copyCells(lc_tDummyArray, m_tCellArray, lc_iOldColumns, lc_iOldRows);
     lc_tDummyArray = null;
     countCells();
    }
   }
  }
  return lc_bRealloc;
 }
 
 
 
 /**
  * Copies the cells from the source array into the destination cell array. Both arrays MUST have the same size!
  * 
  * @param pa_tCells   cells to copy
  */
 protected final void copyCells(int[][] pa_tSource, int[][] pa_tDestination, int pa_iMaxColumns, int pa_iMaxRows)
 {
  int lc_iI = 0;  
  
  if (null != pa_tDestination && null != pa_tSource)
  {
   for (lc_iI = 0; lc_iI  < pa_iMaxRows; lc_iI++)
   {
    System.arraycopy(pa_tSource[lc_iI], 0, pa_tDestination[lc_iI], 0, pa_iMaxColumns);
   }
  }
 }

 
 
 /**
  * Counts the cells
  * 
  * @return   Amount of cells
  */
 public final int countCells()
 {
  int lc_iI     = 0;
  int lc_iJ     = 0;
  int lc_iCount = 0;

  if (null != m_tCellArray && null != m_tProperties)
  {
   for (lc_iI = 0; lc_iI < m_tProperties.getRows(); lc_iI++)
   {
    for (lc_iJ = 0; lc_iJ < m_tProperties.getColumns(); lc_iJ++)
    {
     if (0 != m_tCellArray[lc_iI][lc_iJ])
     {
      lc_iCount++; 
     }
    }  
   }
   m_iCellAmount = lc_iCount;
  }
  return lc_iCount;
 }

 
 
 /**
  * Checks whether the given ruleset would cause a reallocation shrinking the cell matrix
  * 
  * @param pa_tProperties
  * @return 
  */
 public final boolean checkForShrinkingRealloc(GridLifeProperties pa_tProperties)
 {
  if (null == pa_tProperties || null == m_tProperties) return false;
  return (pa_tProperties.getColumns() <  m_tProperties.getColumns() || 
          pa_tProperties.getRows()    <  m_tProperties.getRows()); 
 }
 
 
 
 /**
  * Saves the cell matrix
  * 
  * @param pa_sFileName    file name
  * @param pa_sHeader      file header
  * @param pa_sSeparator   value separator
  * @return               full path, if saving was successfull, otherwise null
  */
 public final String saveMatrix(String pa_sFileName, String pa_sHeader, String pa_sSeparator,
                                int pa_iZoom, int pa_iX, int pa_iY)
 {
  StringBuffer lc_sMatrixLine = null;
  String       lc_sFilePath   = null;
  
  int lc_iI     = 0;
  int lc_iJ     = 0;
  int lc_iCount = 0;
  
  
  File           lc_tFile               = null;
  FileWriter     lc_tFileWriter         = null;
  BufferedWriter lc_tBuffererFileWriter = null;
   
  if (null != pa_sHeader && null != pa_sSeparator)
  {
   if (false == pa_sHeader.isEmpty() && false == pa_sSeparator.isEmpty())
   {
    if (null != (lc_tFile = new File(pa_sFileName)))
    {
     try
     {
      if (null != (lc_tFileWriter = new FileWriter(lc_tFile)))   
      {
       if (null != (lc_tBuffererFileWriter = new BufferedWriter(lc_tFileWriter)))   
       {
        //
        // write header
        //
        lc_tBuffererFileWriter.write(pa_sHeader + m_tProperties.getColumns() + pa_sSeparator + m_tProperties.getRows() + pa_sSeparator +
                                     Integer.toString(pa_iZoom) + pa_sSeparator + 
                                     Integer.toString(pa_iX   ) + pa_sSeparator + 
                                     Integer.toString(pa_iY   ) + "\n");
           
        //
        // write matrix
        //
        for (lc_iI = 0; lc_iI < m_tProperties.getRows(); lc_iI++)
        {
         if (null != (lc_sMatrixLine = new StringBuffer()))
         {
          for (lc_iJ = 0; lc_iJ < m_tProperties.getColumns(); lc_iJ++)
          {
           lc_sMatrixLine.append(m_tCellArray[lc_iI][lc_iJ]);
           if (lc_iJ != m_tProperties.getColumns() - 1)
           {
            lc_sMatrixLine.append(GridLifeConstants.GOL_FILE_CELL_MATRIX_VALUE_SEPARATOR); 
           }
          } 
          lc_tBuffererFileWriter.write(lc_sMatrixLine.toString() + "\n");
          lc_iCount++;
         }
        }
        if (0 < lc_iCount)
        {
         lc_sFilePath = lc_tFile.getCanonicalPath(); 
        }
       }
      }
     }
     catch (Exception lc_tException)
     {}
    }
   }
  }
  
  //
  // Closing file stuff at this position to make sure that in any case we close them!
  //
  try {if (null != lc_tBuffererFileWriter) lc_tBuffererFileWriter.close(); } catch (Exception lc_tException) {}   
  try {if (null != lc_tFileWriter        ) lc_tFileWriter.close();         } catch (Exception lc_tException) {}   

  return lc_sFilePath;
 }
 
 
 
 /**
  * Saves the cell matrix
  * 
  * param pa_sFileName    file name
  * @return               full path, if saving was successfull, otherwise null
  */
 public final String loadMatrix(String pa_sFileName , 
                                String pa_sHeader   , 
                                String pa_sSeparator, 
                                String pa_sComment  )
 {
  String lc_sInputLine  = null;
  String lc_sCheckLine  = null;
  String lc_sFilePath   = null;
  String lc_sSizeString = null;
  
  int lc_iRow    = 0;
  int lc_iColumn = 0;
  int lc_iPos    = 0;
  int lc_iValue  = 0;
  
  int lc_iLoadColumns = 0;
  int lc_iLoadRows    = 0;
  
  Integer lc_tCellAgeColorClass = null;
  
  File            lc_tFile               = null;
  FileReader      lc_tFileReader         = null;
  BufferedReader  lc_tBuffererFileReader = null;
  StringTokenizer lc_tTokenizer          = null;
  
  m_iLoadZoom = 0;
  m_iLoadPosX = 0;
  m_iLoadPosY = 0;
  
  if (null != (lc_tFile = new File(pa_sFileName)))
  {
   try
   {
    if (null != (lc_tFileReader = new FileReader(lc_tFile)))   
    {
     if (null != (lc_tBuffererFileReader = new BufferedReader(lc_tFileReader)))   
     {
      lc_sFilePath = lc_tFile.getCanonicalPath(); 
      
      if (null != m_tHistogramManager)
      {
       m_tHistogramManager.reset();  
      }
      
      while (null != (lc_sInputLine = lc_tBuffererFileReader.readLine()))    
      {
       //
       // remove white spaces, check for commentary lines
       //
       if (null != (lc_sCheckLine = lc_sInputLine.trim()))
       {
        if (false == lc_sCheckLine.isEmpty())
        {
         if (false == lc_sCheckLine.startsWith(pa_sComment))
         {
          //
          // check for header line
          //
          if (true == lc_sCheckLine.startsWith(pa_sHeader)) 
          {
           lc_iPos = lc_sCheckLine.indexOf(pa_sHeader) + pa_sHeader.length();
           if (null != (lc_sSizeString = lc_sCheckLine.substring(lc_iPos)))
           {
            if (-1 != (lc_iPos = lc_sSizeString.indexOf(pa_sSeparator)))
            {
             //
             // reset everything and reallocate
             //
             if (null != (lc_tTokenizer = new StringTokenizer(lc_sSizeString, pa_sSeparator)))
             {
              try
              {
               lc_iLoadColumns = Integer.valueOf(lc_tTokenizer.nextToken());
               lc_iLoadRows    = Integer.valueOf(lc_tTokenizer.nextToken());
               m_iLoadZoom     = Integer.valueOf(lc_tTokenizer.nextToken());
               m_iLoadPosX     = Integer.valueOf(lc_tTokenizer.nextToken());      
               m_iLoadPosY     = Integer.valueOf(lc_tTokenizer.nextToken());      
              }
              catch (Exception lc_tException)
              {
               m_iLoadZoom = 0;
               m_iLoadPosX = 0;
               m_iLoadPosY = 0;
              }
              
              m_tProperties.setColumns(lc_iLoadColumns);
              m_tProperties.setRows   (lc_iLoadRows   );
              m_tCellArray = new int[lc_iLoadRows][lc_iLoadColumns];
              m_lGeneration = 0;
              m_iCellAmount = 0;
              m_iTendency   = 0;
             }
            }
           }
          }
          //
          // data line
          //
          else
          {
           if (null != m_tCellArray)
           {
            if (null != (lc_tTokenizer = new StringTokenizer(lc_sCheckLine, pa_sSeparator)))
            {
             lc_iColumn = 0;
             while (lc_tTokenizer.hasMoreTokens())
             {
              lc_iValue = Integer.valueOf(lc_tTokenizer.nextToken());
              if (0 != lc_iValue) m_iCellAmount++;
              
              m_tCellArray[lc_iRow][lc_iColumn++] = lc_iValue;
              
              //
              // update histogram manager
              //
              if (0 != lc_iValue)
              {
               if (null != m_tHistogramManager && null != m_tColorManager) 
               {
                if (null != (lc_tCellAgeColorClass = m_tColorManager.getAgeClass(m_tColorManager.getActiveSetName(), lc_iValue, null)))
                {
                 m_tHistogramManager.add(lc_tCellAgeColorClass); 
                }
               }
              }
             }
            }
            lc_iRow++;
           }
          }
         }
        }
       }
      }// while
      if (0 < m_iCellAmount) m_lGeneration = 1;
     }
    } 
   }  
   catch (Exception lc_tException)
   {}   
  }   

  //
  // Closing file stuff at this position to make sure that in any case we close them!
  //
  try {if (null != lc_tBuffererFileReader) lc_tBuffererFileReader.close(); } catch (Exception lc_tException) {}   
  try {if (null != lc_tFileReader        ) lc_tFileReader.close();         } catch (Exception lc_tException) {}   

  return (0 < lc_iRow ? lc_sFilePath : null);
 }

 

 /**
  * Returns the properties
  * 
  * @return   Properties
  */
  public final GridLifeProperties getProperties() {return m_tProperties;}


  
 /**
  * Returns the zoom factor of the last loaded file
  * 
  * @return   zoom factor, 0 if undefined
  */
 public final int getLoadZoom() {return m_iLoadZoom;}
 
 
 
 /**
  * Returns the column of the top left cell position of the last loaded file
  * 
  * @return   column, 0 if undefined
  */
 public final int getLoadPosX() {return m_iLoadPosX;}
 
  
 /**
  * Returns the row of the top left cell position of the last loaded file
  * 
  * @return   row, 0 if undefined
  */
 public final int getLoadPosY() {return m_iLoadPosY;}
 
  
}// eoc
