/*
 * Game of Life
 * 
 */
package gridlife.shape;



/**
 * Imports
 * 
 */
import gridlife.core.GridLifeConstants    ;
import gridlife.color.GridLifeColorManager;

import java.util.NoSuchElementException   ;
import java.util.StringTokenizer          ;
import java.awt.Graphics                  ;
import java.awt.Color                     ;
import java.awt.image.BufferedImage       ;
import javax.swing.ImageIcon              ;



/**
 * Basis datastructure for a custom shape
 *  
 * @author Ulf Wagemann
 */
public class GridLifeShape 
{
 private String        m_sShapeName        ;
 private int           m_iColumns          ;
 private int           m_iRows             ;
 private int[][]       m_tShapeDataIntArray;
 private BufferedImage m_tShapeDataImage   ;
 private ImageIcon     m_tImageIcon        ;

 private GridLifeColorManager m_tColorManager;
 
 
 
 /**
  * Constructor
  * 
  * @param pa_sShapeString    String with data in comma-separated form
  */
 public GridLifeShape(String pa_sShapeString, GridLifeColorManager pa_tColorManager)
 {
  StringTokenizer lc_tTokenizer = new StringTokenizer(pa_sShapeString, GridLifeConstants.GOL_FILE_SHAPE_VALUE_SEPARATOR); 
 
  init(pa_tColorManager); 

  if (null != lc_tTokenizer)
  {
   m_sShapeName = lc_tTokenizer.nextToken();
     
   if (null != m_sShapeName)
   {
    if (false == m_sShapeName.isEmpty())
    { 
     m_sShapeName = m_sShapeName.substring(GridLifeConstants.GOL_FILE_CELL_MATRIX_SHAPE_HEADER.length()); 
     m_iColumns   = getIntValue(lc_tTokenizer.nextToken());
     m_iRows      = getIntValue(lc_tTokenizer.nextToken());
     
     if (0 < m_iColumns && Integer.MAX_VALUE > m_iColumns)
     {
      if (0 < m_iRows && Integer.MAX_VALUE > m_iRows)
      {
       m_tShapeDataIntArray = new int[m_iRows][m_iColumns];
      }
     }
    }
   }
  }
 }
 
 
 /**
  * Constructor 
  * 
  * @param pa_sName     shape name
  * @param pa_iWidth    shape width in cells
  * @param pa_iHeight   shape height in cells
  * @param pa_tCells    cells to use 
  */
 public GridLifeShape(String pa_sName, int pa_iWidth, int pa_iHeight, int[][] pa_tCells, GridLifeColorManager pa_tColorManager, boolean pa_bFlattenValues)
 {
  init(pa_tColorManager);
  m_sShapeName = pa_sName;
  
  if (0 < pa_iWidth && Integer.MAX_VALUE > pa_iWidth)
  {
   if (0 < pa_iHeight && Integer.MAX_VALUE > pa_iHeight)
   {
    m_iColumns = pa_iWidth ;
    m_iRows    = pa_iHeight;
    
    if (null != (m_tShapeDataIntArray = new int[m_iRows][m_iColumns]))
    {
     copyIntArray(pa_tCells, pa_bFlattenValues);
     createImage();      
    }
   }
  }
 }
 
 
 
 /**
  * Constructor 
  * 
  * @param pa_sName     shape name
  * @param pa_iWidth    shape width in cells
  * @param pa_iHeight   shape height in cells 
  */
 public GridLifeShape(String pa_sName, int pa_iWidth, int pa_iHeight, GridLifeColorManager pa_tColorManager)
 {
  init(pa_tColorManager);
  m_sShapeName = pa_sName;
  
  if (0 < pa_iWidth && Integer.MAX_VALUE > pa_iWidth)
  {
   if (0 < pa_iHeight && Integer.MAX_VALUE > pa_iHeight)
   {
    m_iColumns = pa_iWidth ;
    m_iRows    = pa_iHeight;    
    m_tShapeDataIntArray = new int[m_iRows][m_iColumns];
   }
  }
 }
 
 
 
 /**
  * Constructor
  * 
  * @param pa_tShape    shape to copy
  */
 public GridLifeShape(GridLifeShape pa_tShape, GridLifeColorManager pa_tColorManager, boolean pa_bFlattenValues)
 {
  init(pa_tColorManager);
  
  if (null != pa_tShape)
  {
   m_sShapeName         = pa_tShape.getShapeName();
   m_iColumns           = pa_tShape.getShapeColumns();
   m_iRows              = pa_tShape.getShapeRows   ();
  
   copyIntArray(pa_tShape.getShapeDataIntArray(), pa_bFlattenValues);
   createImage();
  }
 }
 

 
 /**
  * Returns the number of cell rows
  * 
  * @return   amount of cell rows
  */
 public final int getRows() {return m_iRows;}
 
 
 
 /**
  * Returns the number of cell columns
  * 
  * @return   amount of cell columns
  */
 public final int getColumns() {return m_iColumns;}
 
 
 
 /**
  * Sets shape data and creathes the image
  * 
  * @param pa_tShapeData    Shape data
  */
 public final void setShapeData(String[] pa_tShapeData, Color pa_tColor)
 {
  createIntArray(pa_tShapeData);
  createImage(); 
 }
 
 
 
 /**
  * Retrieves an int value from a string
  * 
  * @param pa_sString   String
  * @return   -1 in case of error, otherwise int value from 0 to ....
  * 
  */ 
 private int getIntValue(String pa_sString)
 {
  if (null != pa_sString)  
  {
   if (false == pa_sString.isEmpty()) 
   {
    if      (pa_sString.equals(GridLifeConstants.GOL_FILE_CELL_MATRIX_DEAD_CELL_1  )) return 0;
    else if (pa_sString.equals(GridLifeConstants.GOL_FILE_CELL_MATRIX_DEAD_CELL_2  )) return 0;
    else if (pa_sString.equals(GridLifeConstants.GOL_FILE_CELL_MATRIX_LIVING_CELL_1)) return 1;
    else if (pa_sString.equals(GridLifeConstants.GOL_FILE_CELL_MATRIX_LIVING_CELL_2)) return 1;
    else return (Integer.valueOf(pa_sString)).intValue(); 
   }
  }
  return -1;
 }
 
 
 
  /**
  * Creates an int array with 0/1 values from a comma-separated shape string.
  * First token : name
  * Second token: width
  * Third token : height
  * Following   : data for rows 0..n-1
  * 
  * @param pa_sShapeString   Shape string
  */
 private void createIntArray(String[] pa_sShapeString)
 {
  StringTokenizer lc_tTokenizer = null;

  int lc_iRows    = 0;
  int lc_iI       = 0;
  int lc_iJ       = 0;

  if (null != pa_sShapeString && null != m_tShapeDataIntArray)
  {
   if (0 < (lc_iRows = pa_sShapeString.length))
   {     
    if (lc_iRows == m_iRows)
    {
     for (lc_iI = 0; lc_iI < lc_iRows; lc_iI++)
     {
      if (null != (lc_tTokenizer = new StringTokenizer(pa_sShapeString[lc_iI], GridLifeConstants.GOL_FILE_SHAPE_VALUE_SEPARATOR)))
      {
       for (lc_iJ = 0; lc_iJ < m_iColumns; lc_iJ++)
       {
        try
        {
         m_tShapeDataIntArray[lc_iI][lc_iJ] = getIntValue(lc_tTokenizer.nextToken());
        }
        catch (NoSuchElementException lc_tException)
        {
         m_tShapeDataIntArray[lc_iI][lc_iJ] = 0; // we interpret missing columns as zero values 
        }
       }
      }
     }
    }
   }
  }
 }
 
 
 
 /**
  * Creates a BufferedImage from the int array
  * 
  */
 private void createImage()
 {
  Graphics  lc_tGraphics     = null;
  Color     lc_tBgColor      = new Color(GridLifeConstants.GOL_SHAPE_BACKGROUND_COLOR);
  Color     lc_tFgColor      = new Color(GridLifeConstants.GOL_SHAPE_BORDER_COLOR    );
  Color     lc_tFillColor    = new Color(GridLifeConstants.GOL_SHAPE_FILL_COLOR      );
  Color     lc_tHiliteColor  = new Color(GridLifeConstants.GOL_SHAPE_HILITE_COLOR    );
  Color     lc_tShadow1Color = new Color(GridLifeConstants.GOL_SHAPE_SHADOW1_COLOR   );
  
  String lc_sActiveColor = (null != m_tColorManager ? m_tColorManager.getActiveSetName() : null);
  
  int lc_iI       = 0;
  int lc_iJ       = 0;
  int lc_iCell    = 0;
  int lc_iX       = 0;
  int lc_iY       = 0;
  
  int lc_iWidth  = (m_iColumns * GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH ) + 2; // +2 to include the shadow
  int lc_iHeight = (m_iRows    * GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT) + 2; // +2 to include the shadow
  
  if (null != m_tShapeDataIntArray)
  {
   if (null != (m_tShapeDataImage = new BufferedImage(lc_iWidth, lc_iHeight, BufferedImage.TYPE_INT_RGB)))
   {
    if (null != (lc_tGraphics = m_tShapeDataImage.getGraphics()))
    {
     //
     // draw background
     //
     lc_tGraphics.setColor(lc_tBgColor);
     lc_tGraphics.fillRect(0, 0, lc_iWidth, lc_iHeight);
     
     //
     // draw cell pattern
     //
     for (lc_iI = 0; lc_iI < m_iRows; lc_iI++)
     {
      for (lc_iJ = 0; lc_iJ < m_iColumns; lc_iJ++)
      {
       lc_iCell = m_tShapeDataIntArray[lc_iI][lc_iJ];
       lc_iX    = lc_iJ * GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH ;
       lc_iY    = lc_iI * GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT;
       
       if (0 != lc_iCell)
       {
        lc_tGraphics.setColor(lc_tFgColor);
        lc_tGraphics.drawRect(lc_iX, lc_iY, GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH, GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT);
        
        if (null != lc_sActiveColor)
        {
         //lc_tGraphics.setColor(m_tColorManager.getColorForSet(lc_sActiveColor, lc_iCell));
         lc_tGraphics.setColor(m_tColorManager.getActiveColor(lc_iCell));
        }
        else
        {
         lc_tGraphics.setColor(lc_tFillColor);
        }
        lc_tGraphics.fillRect(lc_iX+1, lc_iY+1, GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH-2, GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT-2);

        lc_tGraphics.setColor(lc_tHiliteColor);
        lc_tGraphics.drawLine(lc_iX + GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH -1, lc_iY + GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT -1, lc_iX + 1, lc_iY + GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT -1);
        lc_tGraphics.drawLine(lc_iX + GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH -1, lc_iY + 1, lc_iX + GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH -1, lc_iY + GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT -1);  

        //
        // make a tiny shadow for each cell
        //
        lc_tGraphics.setColor(lc_tShadow1Color);
        lc_tGraphics.drawLine(lc_iX + 1, lc_iY + GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT + 1, lc_iX + GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH + 1, lc_iY + GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT + 1);
        lc_tGraphics.drawLine(lc_iX + GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH + 1, lc_iY + GridLifeConstants.GOL_SHAPE_BLOCK_HEIGHT + 1, lc_iX + GridLifeConstants.GOL_SHAPE_BLOCK_WIDTH + 1, lc_iY  + 2);
       }
      }
     }
    }
    m_tImageIcon = new ImageIcon(m_tShapeDataImage);
   }
  }
 }
 
 
 
 /**
  * Returns the shape pattern as int array
  * 
  * @return   Shape pattern, null in case of error
  */
 public final int[][] getShapeDataIntArray() {return m_tShapeDataIntArray;}
 
 
 
 /**
  * Returns the BufferedImage representing the shape
  * 
  * @return   Shape image, null in case of error
  */
 public final BufferedImage getShapeImage() {return m_tShapeDataImage;}
 
 
 /**
  * Returns the ImageIcon representing the shape
  * 
  * @return   Shape image, null in case of error
  */
 public final ImageIcon getShapeImageIcon() {return m_tImageIcon;}

 
 
 /**
  * Retrieves the shape name
  * 
  * @return   Shape name, null in case of error
  */
 public final String getShapeName() {return m_sShapeName;}
 

  /**
  * Sets the shape name
  * 
  * @param pa_sName  new name
  */
 public final void setShapeName(String pa_sName) 
 {
  if (null != pa_sName) m_sShapeName = pa_sName;
 }
 
 
 
 /**
  * Returns the amount of columns of that shape
  * 
  * @return  columns
  */
 public final int getShapeColumns() {return m_iColumns;}

 
 
 /**
  * Returns the amount of rows of that shape
  * 
  * @return  rows
  */
 public final int getShapeRows() {return m_iRows;}
 
 
 
 /**
  * Returns the value of the shape pattern at the given coordinates
  * 
  * @param pa_iColumn  Column
  * @param pa_iRow     Row
  * @return   0 or 1, -1 in case of error
  */
 public final int getShapePattern(int pa_iColumn, int pa_iRow)
 {
  if (null != m_tShapeDataIntArray)
  {
   if (0 <= pa_iColumn && m_iColumns > pa_iColumn)
   {
    if (0 <= pa_iRow && m_iRows > pa_iRow)
    {
     return m_tShapeDataIntArray[pa_iRow][pa_iColumn]; 
    } 
   }
  }
  return -1;
 }
 
 
 
 /**
  * Rotates the shape 90 degrees clock-wise. In case of non-squared shapes the
  * outer form may change. The graphic image is recalculated, too.
  * 
  */
 public final void rotate()
 {
  int lc_iPrevColumns = 0;
  int lc_iPrevRows    = 0;
  int lc_iNewColumns  = 0;
  int lc_iNewRows     = 0;
  
  int lc_iI = 0;
  int lc_iJ = 0;
  
  int[][] lc_tNewShape = null;
  
  //
  // we switch column and row indexes due to the 90 degree shift. this works for both,
  // squared as well as non-squared shapes
  //
  if (null != m_tShapeDataIntArray)
  {
   lc_iPrevColumns = m_iColumns;
   lc_iPrevRows    = m_iRows   ;
   lc_iNewColumns  = m_iRows   ;
   lc_iNewRows     = m_iColumns;
   
   if (null != (lc_tNewShape = new int[lc_iNewRows][lc_iNewColumns]))
   {
    //
    // step 1: rotate shape in helper array
    //
    for (lc_iI = 0; lc_iI  < lc_iPrevRows; lc_iI++)
    {
     for (lc_iJ = 0; lc_iJ  < lc_iPrevColumns; lc_iJ++)
     {
      lc_tNewShape[lc_iJ][lc_iPrevRows - 1 - lc_iI] = m_tShapeDataIntArray[lc_iI][lc_iJ];
     }   
    }
    //
    // step 2: reallocate main array and copy helper array into it
    //
    m_iColumns = lc_iNewColumns;
    m_iRows    = lc_iNewRows   ;
    
    if (null != (m_tShapeDataIntArray = new int[m_iRows][m_iColumns]))
    {
     for (lc_iI = 0; lc_iI  < m_iRows; lc_iI++)
     {
      System.arraycopy(lc_tNewShape[lc_iI], 0, m_tShapeDataIntArray[lc_iI], 0, m_iColumns);
     }
    }
    //
    // step 3: recalculate graphics
    //
    createImage();
   }
  }  
 }
 


 /**
  * Mirrors the shape horizontally. The graphic image is recalculated, too.
  * 
  */
 public final void mirror()
 {
  int lc_iI = 0;
  int lc_iJ = 0;
  
  int[][] lc_tNewShape = null;
  
  //
  // we switch column and row indexes due to the 90 degree shift. this works for both,
  // squared as well as non-squared shapes
  //
  if (null != m_tShapeDataIntArray)
  {
   if (null != (lc_tNewShape = new int[m_iRows][m_iColumns]))
   {
    //
    // step 1: mirror shape in helper array
    //
    for (lc_iI = 0; lc_iI  < m_iRows; lc_iI++)
    {
     for (lc_iJ = 0; lc_iJ  < m_iColumns; lc_iJ++)
     {
      lc_tNewShape[lc_iI][m_iColumns - lc_iJ - 1] = m_tShapeDataIntArray[lc_iI][lc_iJ];
     }   
    }
    //
    // step 2: reallocate main array and copy helper array into it
    //
    for (lc_iI = 0; lc_iI  < m_iRows; lc_iI++)
    {
     System.arraycopy(lc_tNewShape[lc_iI], 0, m_tShapeDataIntArray[lc_iI], 0, m_iColumns);
    }
    //
    // step 3: recalculate graphics
    //
    createImage();
   }
  }  
 }

 
 
 /**
  * Creates an int array with 0/1 values from a comma-separated shape string.
  * First token : name
  * Second token: width
  * Third token : height
  * Following   : data for rows 0..n-1
  * 
  * @param pa_sShapeString   Shape string
  */
 private void copyIntArray(int[][] pa_tIntArray, boolean pa_bFlattenValues)
 {
  int lc_iColumns  = 0;
  int lc_iRows     = 0;
  int lc_iI        = 0;
  int lc_iJ        = 0;
  int lc_iNewValue = 0;
  
  if (null != pa_tIntArray)
  {
   if (0 < (lc_iRows = pa_tIntArray.length)) 
   {
    if (0 < (lc_iColumns = pa_tIntArray[0].length)) 
    {
     if (0 <= lc_iColumns && Integer.MAX_VALUE > lc_iColumns)
     {
      if (0 <= lc_iRows && Integer.MAX_VALUE > lc_iRows)
      {
       //
       // first, create shape as an int array representation
       //
       if (null != (m_tShapeDataIntArray = new int[lc_iRows][lc_iColumns]))
       {
        m_iColumns = lc_iColumns;
        m_iRows    = lc_iRows   ;

        for (lc_iI = 0; lc_iI < m_iRows; lc_iI++)
        {
         for (lc_iJ = 0; lc_iJ < m_iColumns; lc_iJ++)
         {
          lc_iNewValue = pa_tIntArray[lc_iI][lc_iJ];
          
          if (true == pa_bFlattenValues && 0 != lc_iNewValue)
          {
           lc_iNewValue = 1; 
          }
          m_tShapeDataIntArray[lc_iI][lc_iJ] = lc_iNewValue;
         }    
        }
       }
      }
     }
    }
   }
  }
 }

 
 /**
  * Checks whether the shape is valid (e.g. has a value array and valid rows/column values)
  * 
  * @return  true, if the shapie is valid, false otherwise
  */
 public final boolean isValid()
 {
  if (0 < m_iColumns && Integer.MAX_VALUE > m_iColumns)
  {
   if (0 < m_iRows && Integer.MAX_VALUE > m_iRows)
   {
    if (null != m_tShapeDataIntArray && null != m_tShapeDataImage && null != m_tImageIcon)
    {
     return true; 
    }
   }
  }
  return false;
 }
 
 
 
 /**
  * Private init method
  * 
  */
 private void init(GridLifeColorManager pa_tColorManager)
 {
  m_sShapeName         = ""  ;
  m_tShapeDataImage    = null;
  m_tImageIcon         = null;
  m_tShapeDataImage    = null;
  m_tShapeDataIntArray = null;
  m_iColumns           = 0   ;
  m_iRows              = 0   ;
  m_tColorManager = pa_tColorManager;
 }
         
 
 
 /**
  * toString() method
  * 
  * @return shape name
  */
 @Override
 public String toString()
 {
  return m_sShapeName;
 }
 
 
} // eoc
