/*
 * Game of Life
 * 
 */
package gridlife.shape;



/**
 * Imports
 * 
 */
import gridlife.core.GridLifeConstants;
import gridlife.color.GridLifeColorManager;

import java.util.HashMap     ;
import java.util.Set         ;
import java.util.Iterator    ;
import java.util.LinkedList  ;
import java.util.Comparator  ;
import java.util.Collections ;
import java.util.ListIterator;

import java.io.File          ;
import java.io.FileReader    ;
import java.io.BufferedReader;
import java.io.FileWriter    ;
import java.io.BufferedWriter;



/**
 * This class assembles shapes to a set
 * 
 * @author Ulf Wagemann
 */
public class GridLifeShapeSet 
{
 private String m_sShapeSetName ;
 private int    m_iShapeAmount  ;
 
 private HashMap<String, GridLifeShape> m_tShapeTable  ;
 private GridLifeColorManager           m_tColorManager;

 
 /**
  * Constructor. Loads a shape set file
  * 
  * @param pa_sFileName 
  */
 public GridLifeShapeSet(File pa_tFile, GridLifeColorManager pa_tColorManager)
 {
  m_tColorManager  = pa_tColorManager;
  m_sShapeSetName  = "";
  m_iShapeAmount   =  0;

  if (null != (m_tShapeTable = new HashMap<String, GridLifeShape>()))
  {
   load(pa_tFile);
  }
 }
 
 
 
 /**
  * Constructor. Loads a shape set file
  * 
  * @param pa_sFileName 
  */
 public GridLifeShapeSet()
 {
  m_tColorManager  = null;
  m_sShapeSetName  = "";
  m_iShapeAmount   =  0;

  m_tShapeTable = new HashMap<String, GridLifeShape>();
 }
 
 
 
 /**
  * Clears internal memory.
  * We do not want to eat up memory, since we store own datastructures we need to remove them from the
  * hashmap and mark then as gone for the GC.
  * 
  * dispose() should be called when this shapeset is not needed anymore.
  * 
  */
 public final void dispose()
 {
  if (null != m_tShapeTable) {m_tShapeTable.clear();}
 }
 
 
  /**
  * Retrieves all of the shapes belonging to the current set
  * 
  * @return array of shapes, null in case of error
  */ 
 public final GridLifeShape[] getShapes()
 {
  GridLifeShape[]       lc_tShapes   = null;
  Set<String>           lc_tNames    = null;
  Iterator<String>      lc_tIterator = null;
  GridLifeShape         lc_tShape    = null;
  
  int lc_iCount = 0;
  
  if (null != m_tShapeTable)
  {
   if (0 < m_iShapeAmount) 
   {
    if (null != (lc_tNames = m_tShapeTable.keySet()))
    {
     if (null != (lc_tIterator = lc_tNames.iterator()))
     {
      if (null != (lc_tShapes = new GridLifeShape[m_tShapeTable.size()])) 
      {
       while (lc_tIterator.hasNext()) 
       {
        if (null != (lc_tShape = m_tShapeTable.get(lc_tIterator.next())))
        {
         lc_tShapes[lc_iCount++] = lc_tShape; 
        }
       }// while
      }
     }
    }
   }
  }
  return lc_tShapes;
 }
 
 
 
 /**
  * Retrieves the names of all of the shapes belonging to the current set
  * 
  * @return   Names of all shapes as a set, null in case of error
  */
 public final Set<String> getShapeNames()
 {
  if (null != m_tShapeTable)
  {
   if (0 < m_iShapeAmount)
   {
    return m_tShapeTable.keySet();   
   }
  }
  return null;
 }
 
 
 
 /**
  * Returns a linked list containing shapes sorted by name
  * 
  * @return  null in case of error, list of strings otherwise
  */
 public final LinkedList<String> getShapeNamesSorted()
 {
  LinkedList<String> lc_tList     = null;
  Set<String>        lc_tNames    = null;
  Iterator<String>   lc_tIterator = null;
  
  if (null != (lc_tNames = getShapeNames()))
  {
   if (0 < lc_tNames.size())
   {
    if (null != (lc_tIterator = lc_tNames.iterator()))
    {
     if (null != (lc_tList = new LinkedList<String>()))
     {
      //
      // build list
      // 
      while (true == lc_tIterator.hasNext()) 
      {
       lc_tList.add(lc_tIterator.next()); 
      }
      
      //
      // sort it
      //
      Collections.sort(lc_tList, new Comparator<String>() 
                                 {
                                  public int compare(String pa_sString1, String pa_sString2)
                                  { 
                                   return pa_sString1.compareTo(pa_sString2);
                                  }
                                 });
      
     }
    } 
   }
  }
  return lc_tList;
 }
 
 
 
 /**
  * Retrieves the shape with the given name
  * 
  * @param pa_sName   Name of shape
  * @return           Shape, null in case of error
  */
 public final GridLifeShape getShape(String pa_sName)
 {
  GridLifeShape lc_tShape = null;
  
  if (null != m_tShapeTable)
  {
   if (null != pa_sName)
   {
    if (false == pa_sName.isEmpty())
    {
     lc_tShape = m_tShapeTable.get(pa_sName); 
    }
   }
  }
  return lc_tShape;
 }
 
 
 
 /**
  * Retrieves the shape set's name
  * 
  * @return   Name of shape set
  */
 public final String getSetName() {return m_sShapeSetName;};
 
 
 
 /**
  * Retrieves the amount of shapes belonging to the current set
  * 
  * @return   Amount of shapes, 0 in case of error
  */
 public final int getShapeAmount() {return m_iShapeAmount;};
 
 
 
 /**
  * Loads a shape set file. Format:
  * 
  * Line 1    : @GOL_SHAPESET@<name>
  * Line 2    : ShapeString | CommentSring | ShapeLineString
  * 
  * ShapeString    : Shapename,width,height
  * CommentString  : # + any string
  * ShapeLineString: comma-separated list of cell age values for a single line
  * 
  * @param pa_sFileName   Filename 
  */
 public final void load(File pa_tFile)
 {
  FileReader     lc_tInputReader         = null;
  BufferedReader lc_tBufferedInputReader = null;
  GridLifeShape  lc_tShape               = null;

  String[]       lc_sShapeLines          = null;
  String         lc_sInputLine           = null;
  String         lc_sCheckLine           = null;
  
  int  lc_iPosition       = 0;
  int  lc_iShapeLinesRead = 0;
  
  m_iShapeAmount = 0;
  if (null != pa_tFile && null != m_tShapeTable) 
  {
   if (true == pa_tFile.isFile())   
   {
    try
    {
     if (null != (lc_tInputReader = new FileReader(pa_tFile)))   
     {
      if (null != (lc_tBufferedInputReader = new BufferedReader(lc_tInputReader)))   
      {
       while (null != (lc_sInputLine = lc_tBufferedInputReader.readLine()))    
       {
        //
        // remove white spaces, check for commentary lines
        //
        if (null != (lc_sCheckLine = lc_sInputLine.trim()))
        {
         if (false == lc_sCheckLine.isEmpty())
         {
          if (false == lc_sCheckLine.startsWith(GridLifeConstants.GOL_FILE_SHAPE_COMMENT))
          {
           //
           // check for shapeset header line
           //
           if (true == lc_sCheckLine.startsWith(GridLifeConstants.GOL_FILE_SHAPE_HEADER)) 
           {
            lc_iPosition    =  lc_sCheckLine.indexOf(GridLifeConstants.GOL_FILE_SHAPE_HEADER) + GridLifeConstants.GOL_FILE_SHAPE_HEADER.length();
            m_sShapeSetName = lc_sCheckLine.substring(lc_iPosition);
           }
           //
           // check for shape name and size
           //
           else
           {
            if (true == lc_sCheckLine.startsWith(GridLifeConstants.GOL_FILE_SHAPE_NAME_HEADER)) 
            {
             if (null != (lc_tShape = new GridLifeShape(lc_sCheckLine, m_tColorManager)))
             {
              if (0 < lc_tShape.getRows() && 0 < lc_tShape.getColumns())
              {
               lc_sShapeLines     = new String[lc_tShape.getRows()];
               lc_iShapeLinesRead = 0;
              }
              else
              {
               //
               // invalid shape size, delete shape
               //
               lc_tShape = null;
              }
             }
            }
            //
            // must be a single shapeline
            //
            else
            {
             if (null != lc_tShape && null != lc_sShapeLines)
             {
              lc_sShapeLines[lc_iShapeLinesRead++] = lc_sCheckLine;

              //
              // check if we have read the last of the needed data lines. if so, store the shape in the shape list
              //
              if (lc_tShape.getRows() == lc_iShapeLinesRead)
              {
               lc_tShape.setShapeData(lc_sShapeLines, null); // null = use default cell fill color
               if (true == lc_tShape.isValid())
               {
                m_tShapeTable.put(lc_tShape.getShapeName(), lc_tShape);
                m_iShapeAmount++;                        
               }
              }
             }
            }
           }
          }
         }
        }
       } // while
      }
     }
    }
    catch (Exception lc_tException)
    {}

    //
    // Closing file stuff at this position to make sure that in any case we close them!
    //
    try {if (null != lc_tBufferedInputReader) lc_tBufferedInputReader.close(); } catch (Exception lc_tException) {}   
    try {if (null != lc_tInputReader        ) lc_tInputReader.close();         } catch (Exception lc_tException) {}   
   }   
  }
 }
 
 
 
 /**
  * Saves the current shapeset to a file
  * 
  * @param pa_tFile     file structure, must be initialized
  * @param pa_sSetName  set name
  */
 public final void save(File pa_tFile)
 {
  FileWriter     lc_tOutputWriter         = null;
  BufferedWriter lc_tBufferedOutputWriter = null;
  GridLifeShape  lc_tShape                = null;

  LinkedList<String>   lc_tShapeNames   = getShapeNamesSorted();
  ListIterator<String> lc_tListIterator = lc_tShapeNames.listIterator(0);
  
  int    lc_iI       = 0;
  int    lc_iJ       = 0;
  int    lc_iColumns = 0;
  int    lc_iRows    = 0;
  int    lc_iPos     = 0;
  int    lc_iCellValue = 0;
  
  String lc_sSetName = null;
  
  if (null != pa_tFile && null != m_tShapeTable && null != lc_tShapeNames && null != lc_tListIterator) 
  {
   try
   {
    if (null != (lc_tOutputWriter = new FileWriter(pa_tFile)))   
    {
     if (null != (lc_tBufferedOutputWriter = new BufferedWriter(lc_tOutputWriter)))   
     {
      lc_sSetName = pa_tFile.getName();
      if (-1 != (lc_iPos = lc_sSetName.lastIndexOf(".")))
      {
       lc_sSetName =  lc_sSetName.substring(0, lc_iPos); 
      }
       
      //
      // file header
      //
      lc_tBufferedOutputWriter.write(GridLifeConstants.GOL_FILE_SHAPE_HEADER + lc_sSetName + "\n");
       
      //
      // shapes
      //
      while (true == lc_tListIterator.hasNext())
      {
       if (null != (lc_tShape = getShape(lc_tListIterator.next()))) 
       {
        lc_iColumns = lc_tShape.getColumns(); 
        lc_iRows    = lc_tShape.getRows()   ; 

        lc_tBufferedOutputWriter.write(GridLifeConstants.GOL_FILE_SHAPE_NAME_HEADER + lc_tShape.getShapeName() + GridLifeConstants.GOL_FILE_SHAPE_VALUE_SEPARATOR);        
        lc_tBufferedOutputWriter.write(lc_iColumns + GridLifeConstants.GOL_FILE_SHAPE_VALUE_SEPARATOR + lc_iRows + "\n");        
         
        for (lc_iI = 0; lc_iI < lc_iRows; lc_iI++)
        {
         for (lc_iJ = 0; lc_iJ < lc_iColumns; lc_iJ++)
         {
          lc_iCellValue = lc_tShape.getShapePattern(lc_iJ, lc_iI);
          
          lc_tBufferedOutputWriter.write(Integer.toString(lc_iCellValue) + ""); // strangely, without + "" it does not write values..
          if (lc_iJ != (lc_iColumns -1))
          {
           lc_tBufferedOutputWriter.write(GridLifeConstants.GOL_FILE_SHAPE_VALUE_SEPARATOR); 
          }
         } 
         lc_tBufferedOutputWriter.write("\n"); 
        }
        lc_tBufferedOutputWriter.write("\n\n");  
       }
      } // while
    
     }
    }
   }
   catch (Exception lc_tException)
   {}

   //
   // Closing file stuff at this position to make sure that in any case we close them!
   //
   try {if (null != lc_tBufferedOutputWriter) lc_tBufferedOutputWriter.close(); } catch (Exception lc_tException) {}   
   try {if (null != lc_tOutputWriter        ) lc_tOutputWriter.close();         } catch (Exception lc_tException) {}   
  }   
 }

 
 
 /**
  * Adds a shape. Any existing shape with the same name is overwritten.
  * 
  * @param pa_tShape   Shape
  */
 public final void addShape(GridLifeShape pa_tShape)
 {
  if (null != m_tShapeTable)
  {
   m_tShapeTable.put(pa_tShape.getShapeName(), pa_tShape);  
   m_iShapeAmount++;
  }
 }
} // eoc
