/**
 * Game of Life
 * 
 */
package gridlife.shape;



/**
 * Imports
 * 
 */
import gridlife.core.GridLifeConstants;
import java.util.HashMap     ;
import java.util.Collection  ;
import java.util.Iterator    ;
import java.util.Comparator  ;
import java.util.Arrays      ;
import java.io.File          ;



/**
 * This class manages a collection of shapesets
 * 
 * @author Ulf Wagemann
 */
public class GridLifeShapeSetManager 
{
 private int    m_iAmountOfSets ; 
 private HashMap<String, GridLifeShapeSet> m_tShapeSetMap ;
 
 
 /**
  * Constructor
  * 
  */
 public GridLifeShapeSetManager()
 {
  m_tShapeSetMap   = new HashMap<String, GridLifeShapeSet>();
  m_iAmountOfSets  = 0;
  
  loadShapeSets();
 }
 
 
 
 /**
  * We do not want to eat up memory, since we store own datastructures we need to remove them from the
  * hashmap and mark then as gone for the GC.
  * 
  * dispose() should be called when this shapeset manager is not needed anymore.
  * 
  */
 public final void dispose()
 {
  if (null != m_tShapeSetMap) {m_tShapeSetMap.clear();}
 }
 
 

 /**
  * Retrieves all of the shapes belonging to the current set
  * 
  * @return Collection of shapes, null in case of error
  */
 public final Collection<GridLifeShapeSet> getShapeSets()
 {
  if (null != m_tShapeSetMap)
  {
   if (0 < m_iAmountOfSets)
   {
    return m_tShapeSetMap.values();   
   }
  }
  return null;
 }
 
 
 
 /**
  * Retrieves the sorted array of names of all valid shapesets 
  * 
  * @return   Names of all shapes as a set, null in case of error
  */
 public final String[] getShapeSetNames()
 {
  Collection<GridLifeShapeSet>  lc_tShapeSets        = null;
  Iterator<GridLifeShapeSet>    lc_tShapeSetIterator = null;
  GridLifeShapeSet              lc_tShapeSet         = null;
  String[]                      lc_tResult           = null;
  
  int lc_iAmount = 0;
  int lc_iCount  = 0;
  
  if (null != m_tShapeSetMap)
  {
   if (0 < (lc_iAmount = m_iAmountOfSets))
   {
    if (null != (lc_tShapeSets = m_tShapeSetMap.values())) 
    {
     if (null != (lc_tShapeSetIterator = lc_tShapeSets.iterator())) 
     {
      if (null != (lc_tResult = new String[lc_iAmount]))  
      {
       while (true == lc_tShapeSetIterator.hasNext())
       {
        if (null != (lc_tShapeSet = lc_tShapeSetIterator.next()))
        {
         if (0 < lc_tShapeSet.getShapeAmount())
         {
          lc_tResult[lc_iCount++] = lc_tShapeSet.getSetName();
         }
        }
       } // while
       
       if (0 < lc_iCount)
       {
        java.util.Arrays.sort(lc_tResult);   
       }
       return lc_tResult;
      }
     }
    } 
   }
  }
  return null;
 }
 
 
 
 /**
  * Retrieves the shapeset with the given name
  * 
  * @param pa_sName   Name of shapeset
  * @return           Shapeset, null in case of error
  */
 public final GridLifeShapeSet getShapeSet(String pa_sName)
 {
  GridLifeShapeSet lc_tShapeSet = null;
  
  if (null != m_tShapeSetMap)
  {
   if (null != pa_sName)
   {
    if (false == pa_sName.isEmpty())
    {
     lc_tShapeSet = m_tShapeSetMap.get(pa_sName); 
    }
   }
  }
  return lc_tShapeSet;
 }
 
 
 
 /**
  * Retrieves the shapeset with the given name in form of an array
  * 
  * @param pa_sName   Name of shapeset
  * @return           Shapeset, null in case of error
  */
 public final GridLifeShape[] getShapeSetShapes(String pa_sName)
 {
  GridLifeShapeSet lc_tShapeSet = null;
  GridLifeShape[]  lc_tShapes   = null;
  
  if (null != (lc_tShapeSet = getShapeSet(pa_sName)))
  {
   if (null != (lc_tShapes = lc_tShapeSet.getShapes()))
   {
    Arrays.sort(lc_tShapes, new Comparator<GridLifeShape>() 
                                 {
                                  public int compare(GridLifeShape o1, GridLifeShape o2)
                                  { 
                                   return o1.getShapeName().compareTo(o2.getShapeName());
                                  }
                                 });
    
    
    return lc_tShapes; 
   }
  }
  return null;
 }
 
 
 
 /**
  * Retrieves the amount of shapes belonging to the current set
  * 
  * @return   Amount of shapes, 0 in case of error
  */
 public final int getShapeSetAmount() {return m_iAmountOfSets;};

 
 
 /**
  * Adds a shapeset to the manager
  * 
  * @param pa_tSet        Shapeset
  * @param pa_bOverwrite  true, if a present set with the same name should be overwritten
  * @return   true, if the shapeset was added successfully, false otherwise
  * 
  */
 public final boolean addShapeSet(GridLifeShapeSet pa_tSet, boolean pa_bOverwrite)
 {
  boolean lc_bContinue = true ;
  boolean lc_bResult   = false;
  
  if (null != m_tShapeSetMap) 
  {
   if (null != pa_tSet) 
   {
    if (null != m_tShapeSetMap.get(pa_tSet.getSetName()))
    {
     lc_bContinue =  pa_bOverwrite;
    }
    
    if (true == lc_bContinue)
    {
     m_tShapeSetMap.put(pa_tSet.getSetName(), pa_tSet);
     m_iAmountOfSets++; 
     lc_bResult = true;
    }
   }
  }
  return lc_bResult;
 }
 

 
 /**
  * Scans the palette directory for shapesets and loads them
  * 
  */
 private void loadShapeSets()
 {
  File   lc_tFile             = null;
  File   lc_tPaletteDirectory = null;
  File[] lc_tDirectoryEntries = null;
  
  int lc_iAmount        = 0;
  int lc_iI             = 0;

  GridLifeShapeSet           lc_tShapeSet = null;
  GridLifeShapeSetFileFilter lc_tFilter   = new GridLifeShapeSetFileFilter();
  
  if (null != (lc_tPaletteDirectory = new File(GridLifeConstants.GOL_FILE_PALETTE_DIR)))
  {
   if (true == lc_tPaletteDirectory.exists() && true == lc_tPaletteDirectory.isDirectory())
   {
    if (null != (lc_tDirectoryEntries = lc_tPaletteDirectory.listFiles(lc_tFilter)))
    {
     if (0 < (lc_iAmount = lc_tDirectoryEntries.length))
     {
      if (1 < lc_iAmount)
      {
       java.util.Arrays.sort(lc_tDirectoryEntries);        
      }
      for (lc_iI = 0; lc_iI < lc_iAmount; lc_iI++)
      {
       if (null != (lc_tFile = lc_tDirectoryEntries[lc_iI])) 
       {
        if (null != (lc_tShapeSet = new GridLifeShapeSet(lc_tFile, null)))
        {
         addShapeSet(lc_tShapeSet, true);
        }
       }
      }
     }
    }
   }    
  }
 }
 
 
 
} // eoc
