/**
 * Game of Life
 * 
 */
package gridlife.ui;



/**
 * Imports
 * 
 */
import gridlife.core.GridLifeProperties;
import java.awt.Dimension     ;
import java.awt.Rectangle     ;
import java.awt.Color         ;
import javax.swing.JScrollPane;
import java.awt.Cursor        ;



/**
 * Basic class for a panel able of painting cells.
 * 
 * @author Ulf Wagemann
 */
public abstract class GridLifeCellPaintPanel extends GridLifePaintPanel
{
 protected int[][] m_tCells     ;
 protected int     m_iCellSize  ;
 
 protected static Cursor m_tDefaultCursor;
 protected static Cursor m_tCrossHairCursor;
 
 
 
 /**
  * Constructor
  * 
  * @param pa_tScrollPane    own scroll pane
  * @param pa_tRuleset       properties
  * @param pa_iCellSize      cell size in px
  * @param pa_bIsRadarPanel  if true, we calculate for a radar panel, if false, we calculate for a cell matrix panel
  */
 public GridLifeCellPaintPanel(int[][] pa_tCells, JScrollPane pa_tScrollPane, GridLifeProperties pa_tProperties, int pa_iCellSize)
 {
  super(pa_tScrollPane, pa_tProperties);
    
  m_iCellSize = pa_iCellSize;
  m_tCells    = pa_tCells   ;

  m_tDefaultCursor   = Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR  ); 
  m_tCrossHairCursor = Cursor.getPredefinedCursor(Cursor.CROSSHAIR_CURSOR); 

  setBackground(Color.white);
  setDoubleBuffered(true);
  setPanelSize();    
 }
  
 
 
 /**
  * Returns the default cursor
  * 
  * @return   default cursor
  */
 public final Cursor getDefaultCursor() {return m_tDefaultCursor;}
   

 
 
 /**
  * Returns the crosshair cursor
  * 
  * @return   crosshair cursor
  */
 public final Cursor getCrossHairCursor() {return m_tCrossHairCursor;}
 
 
 /**
  * Sets the crosshair cursor
  * 
  */
 public final void setCrossHairCursor() {this.setCursor(getCrossHairCursor());}

 
 
/**
 * Sets the default cursor
 * 
 */
 public final void setDefaultCursor() {this.setCursor(getDefaultCursor());}
 
 

 
 /**
  * Tests whether one of the frame's bounding points lies within the visible x area
  * 
  * @param pa_iX      cell's x coordinate of the top left corner
  * @param pa_tRect   visible rect of scrollpane
  * @return           true, if the frame lies inside x-wise visible inside the rect
  */
 public final boolean isFramePointVisible(int pa_iX, int pa_iY, Rectangle pa_tRect)
 {
  if (null != pa_tRect)
  {
   return ( (pa_iX >= pa_tRect.x && pa_iX < pa_tRect.x + pa_tRect.width) &&
            (pa_iY >= pa_tRect.y && pa_iY < pa_tRect.y + pa_tRect.height) ); 
  }
  return false;
 }

 
 
 /**
  * Tests whether one of the cell's bounding points lies within the visible x area
  * 
  * @param pa_iX      cell's x coordinate of the top left corner
  * @param pa_tRect   visible rect of scrollpane
  * @return           true, if the cell lies inside x-wise visible inside the rect
  */
 public final boolean isCellVisibleHorizontally(int pa_iX, Rectangle pa_tRect)
 {
  boolean lc_bResult = false;
  
  int lc_iX1 = pa_iX;
  int lc_iX2 = pa_iX + m_iCellSize;
  
  lc_bResult |= (lc_iX1 >= pa_tRect.x && lc_iX1 < pa_tRect.x + pa_tRect.width); 
  lc_bResult |= (lc_iX2 >= pa_tRect.x && lc_iX2 < pa_tRect.x + pa_tRect.width);  
  
  return lc_bResult;
 }

 
 
 /**
  * Tests whether one of the cell's bounding points lies within the visible y area
  * 
  * @param pa_iY      cell's y coordinate of the top left corner
  * @param pa_tRect   visible rect of scrollpane
  * @return           true, if the cell lies inside y-wise visible inside the rect
  */
 public final boolean isCellVisibleVertically(int pa_iY, Rectangle pa_tRect)
 {
  boolean lc_bResult = false;
  
  int lc_iY1 = pa_iY;
  int lc_iY2 = pa_iY + m_iCellSize;
  
  lc_bResult |= (lc_iY1 >= pa_tRect.y && lc_iY1 < pa_tRect.y + pa_tRect.height);
  lc_bResult |= (lc_iY2 >= pa_tRect.y && lc_iY2 < pa_tRect.y + pa_tRect.height);     
  
  return lc_bResult;
 }
 

 /**
  * Returns the cell size in px
  * 
  * @return   Cell size in px
  */
 public final int getCellSize() {return m_iCellSize;}

 
 
 /**
  * Changes the cell size, causing reallocation of some buffers
  * 
  * @param pa_iNewSize   new cell size in px
  */
 public final void changeCellSize(int pa_iNewSize)
 {
  m_iCellSize = pa_iNewSize;
  setPanelSize();
 }
 
 
 
 /**
  * Sets the panel's size
  * 
  * todo: bessere auflösungsnuztunh
  */ 
 public final void setPanelSize()
 {
  Dimension lc_tSize  = null;

  int lc_iColumns = (null != getProperties() ? getProperties().getColumns() : 0);
  int lc_iRows    = (null != getProperties() ? getProperties().getRows()    : 0);

  if (0 != lc_iColumns && 0 != lc_iRows)
  {
   if (null != (lc_tSize = new Dimension(1 + (lc_iColumns * m_iCellSize), 1 + (lc_iRows * m_iCellSize))))
   {
    setPreferredSize(lc_tSize);
    setSize(lc_tSize);
   }    
  }
 }
 
 
 /**
  * Sets reference to cell array to given value
  * 
  * @param pa_tCells    cell array reference
  */
 public final void setCells(int[][] pa_tCells)
 {
  m_tCells = pa_tCells; 
 }
 
 
 
 /**
  * Makes sure that a cell location will be inside the matrix' bounds
  * 
  * @param pa_iValue  position to check
  * @param pa_iMax    max valie
  * @return  position, might have been set to 0 or pa_iMax
  */
 protected final int correctLocation(int pa_iValue, int pa_iMax)
 {
  int lc_iResult = pa_iValue;
  
  if (0       > lc_iResult) lc_iResult = 0;
  if (pa_iMax < lc_iResult) lc_iResult = pa_iMax;
  
  return lc_iResult;
 }
  
} // eoc
